# Ebisu Cpp

[![Build Status](https://drone.io/github.com/patefacio/ebisu_cpp/status.png)](https://drone.io/github.com/patefacio/ebisu_cpp/latest)

Code generation library/toolkit that focuses on generating the
structure of c++ code and as much of the boilerplate coding that goes
with that in the creation of c++ systems. Much like its required
companion library [ebisu](https://github.com/patefacio/ebisu) this
library does what it can to leverage a convenient declarative approach
encouraging heavy use of *cascades*.

## General C++ Code Generation Approach

As stated the focus is on c++ *structure* as opposed to attempting to
support all aspects of the language. On any large system there is some
combination of boilerplate and hand-crafted custom business logic. The
larger the amount of boilerplate the larger the benefit of code
generation. In order to make the use of code generation palatable in
the face of large amounts of custom code it is important to be able to
have the two live nicely side-by-side. In *ebisu* and *ebisu_cpp* this
is accomplished with *Protect Blocks*, also called *Custom Blocks*. A
*custom block* is code that exists in generated output that is
protected because it has been hand coded. Custom blocks are designated
with special comments, which are labeled with *tags*. Here are some
examples:

For C++, a custom block with a *tag* of *<ClsPublic Change_tracker>*
designating the class public section of class *Change_tracker*:

    template <typename T>
    class Change_tracker {
     public:
      // custom <ClsPublic Change_tracker>
           **** CUSTOM CODE WRITTEN HERE ****
      // end <ClsPublic Change_tracker>

      //! getter for current_ (access is Ro)
      T current() const { return current_; }

      //! getter for previous_ (access is Ro)
      T previous() const { return previous_; }

     private:
      T current_{};
      T previous_{};
    };

All code around the *custom block* is generated. Any code within a
*Custom Block* is preserved from one run to the next of the code
generation.

For *cmake* a custom block with *tag* of *<date_time_converter libs>*
designating additional libraries required for the
*date_time_converter* application:

    target_link_libraries(date_time_converter
    # custom <date_time_converter libs>
      ${Boost_DATE_TIME_LIBRARY}
      ${Boost_REGEX_LIBRARY}
    # end <date_time_converter libs>
      ${Boost_PROGRAM_OPTIONS_LIBRARY}
      ${Boost_SYSTEM_LIBRARY}
      ${Boost_THREAD_LIBRARY}
    )


### Approach to Naming

These code generation utilities ensure consistent naming by requiring
all potential identifier names to be derived from an identifier
provided by the user in *snake case*. A utility library
[id](https://pub.dartlang.org/packages/id) with class *Id* is used to
ensure names are converted from the input *snake case* to whatever
case is deemed appropriate. The naming conventions are guaranteed
consistent and can be modified to taste by assigning a [Namer]
instance to the [Installation] prior to generation. Two namers are
included, [EbisuCppNamer] and [GoogleNamer]. 

### Structural Targets

Specifically, the following are some of the current structural targets
for covered by *ebisu_cpp*:

* [Enum]: C++ enums, with some support for serialization and different
  flavors - such as *masks*.

          Example:

          enum_('submit_result')
          ..values = [
            'submit_succeeded',
            'submit_invalid_market',
            'submit_invalid_order_details' ],

          Generates Content:

          enum Submit_result {
            Submit_succeeded_e,
            Submit_invalid_market_e,
            Submit_invalid_order_details_e
          };
  In addition to the basic enum, support is provided for converting
  enum values to and from strings, streaming and mask enums.
  
* [Class]: The heart of the language. With simple declarations,
  classes can be generated:

                  class_('change_tracker')
                  ..descr = '''
            Tracks current/previous values of the given type of data. For some
            algorithms it is useful to be able to examine/perform logic on
            current value and compare or evalutate how it has changed since
            previous value.'''
                  ..template = [ 'typename T' ]
                  ..customBlocks = [clsPublic]
                  ..members = [
                    member('current')..type = 'T'..access = ro,
                    member('previous')..type = 'T'..access = ro,
                  ],
  Which generates content:

            /**
             Tracks current/previous values of the given type of data. For some
             algorithms it is useful to be able to examine/perform logic on
             current value and compare or evalutate how it has changed since
             previous value.
            */
            template <typename T>
            class Change_tracker {
             public:
              // custom <ClsPublic Change_tracker>
                 ...                
              // end <ClsPublic Change_tracker>

              //! getter for current_ (access is Ro)
              T current() const { return current_; }

              //! getter for previous_ (access is Ro)
              T previous() const { return previous_; }

             private:
              T current_{};
              T previous_{};
            };


* [Header]: A single header file, typically with a collection of C++
  type things like [includes], enums, classes, forward declarations,
  using statements

            Example:
            
            final raii = lib('raii')
              ..namespace = namespace([ 'fcs', 'raii' ])
              ..headers = [
                header('change_tracker')
                ..includes = [ 'boost/call_traits.hpp' ]
                ..classes = [
                  class_('change_tracker')
                  ...
                ]

            Generates File:

                $TOP/fcs/cpp/fcs/raii/change_tracker.hpp

* [Impl]: A single cpp implementation, typically with a collection of
  C++ type things like includes, enums, classes, forward declarations,
  using statements

* [Lib]: C++ libraries. C++ Libarary sometimes has a newer connotation
  in that there is a concept of *header only* library which implies no
  library at all. In *ebisu_cpp* [Lib] does not necessarily mean a C++
  archive is created. Rather it is a collection of code generated
  within a *namespace* in a consistent directory structure that may or
  may not entail the creation of a C++ archive.

        final utils = lib('utils')
          ..namespace = namespace([ 'fcs', 'utils' ])
          ..headers = [
            header('block_indenter')
            ...,
            header('fixed_size_char_array')
            ...,
            header('utils')
            ...,
            header('version_control_commit')
            ...,
            header('histogram')
            ...,
          ];

        Generates Files:
        No change: $TOP/fcs/cpp/fcs/utils/block_indenter.hpp
        No change: $TOP/fcs/cpp/fcs/utils/fixed_size_char_array.hpp
        No change: $TOP/fcs/cpp/fcs/utils/utils.hpp
        No change: $TOP/fcs/cpp/fcs/utils/version_control_commit.hpp
        No change: $TOP/fcs/cpp/fcs/utils/histogram.hpp


* [App]: A single application executable.

            final date_time_converter = app('date_time_converter')
              ..namespace = namespace(['fcs'])
              ..customBlocks = [ fcbEndNamespace ]
              ..includes = [
                'fcs/timestamp/conversion.hpp',
                'fcs/utils/streamers/table.hpp',
                'fcs/timestamp/conversion.hpp',
                'stdexcept',
              ]
              ..args = [
                arg('timestamp')
                ..shortName = 't'
                ..descr = 'Some form of timestamp'
                ..isMultiple = true
                ..type = ArgType.STRING,
                arg('date')
                ..shortName = 'd'
                ..descr = 'Some form of date'
                ..isMultiple = true
                ..type = ArgType.STRING,
              ];

        Generates: $TOP/fcs/cpp/app/date_time_converter/date_time_converter.cpp
        If run in context of installation updates build scripts with targets content:

        ######################################################################
        # Application build directives
        ######################################################################
        add_executable(date_time_converter
          app/date_time_converter/date_time_converter.cpp
        )

        target_link_libraries(date_time_converter
        # custom <date_time_converter libs>
          ${Boost_DATE_TIME_LIBRARY}
          ${Boost_REGEX_LIBRARY}
        # end <date_time_converter libs>
          ${Boost_PROGRAM_OPTIONS_LIBRARY}
          ${Boost_SYSTEM_LIBRARY}
          ${Boost_THREAD_LIBRARY}
        )
        add_executable(display_csv
          app/display_csv/display_csv.cpp
        )




* [AppArg]: An application argument that gets translated into
  *boost::program_option* code. Even when using helpful libraries like
  *boost::program_options*, there is typically large amounts of
  boilerplate code. The arguments associated with an application are
  localized into a single class containing the program options as
  fields. So for the example application above:

          struct Program_options {
            Program_options(int argc, char** argv) {...}
            static boost::program_options::options_description const& description() {...}
            static void show_help(std::ostream& out) {...}
            //! getter for help_ (access is Ro)
            bool help() const { return help_; }
            //! getter for timestamp_ (access is Ro)
            std::vector<std::string> const& timestamp() const { return timestamp_; }
            //! getter for date_ (access is Ro)
            std::vector<std::string> const& date() const { return date_; }
            friend inline std::ostream& operator<<(std::ostream& out,
                                                   Program_options const& item) {

           private:
            bool help_{};
            std::vector<std::string> timestamp_{};
            std::vector<std::string> date_{};
          };
          ...
          int main(int argc, char** argv) {
            using namespace fcs;
            try {
              Program_options options = {argc, argv};
              if (options.help()) {
                Program_options::show_help(std::cout);
                return 0;
              }
          ...
          }

* [Test]: A generated test. 

* [Script]: Any large C++ installation will typically require or
  benefit from some set of scripts.

* [Installation]: The *kit and caboodle*. An installation is a
  collection of *libs*, *apps*, *tests* and *scripts*. When an
  installation is generated it generates all artifacts, taking care to
  leave all protect blocks in tact.


### Functional Targets

In most large scale projects there are often those mundane tasks that
scream for code generation. Here are some that are covered by
*ebisu_cpp*

* *Common Operators* Often support for various operators can (and should) be code generated. Assume you have a class that looks something like:

            class_('code_packages_value')
            ..opEqual
            ..opLess
            ..isStreamable = true
            ..members = [
              member('name')..type = 'fcs::utils::Fixed_size_char_array<64>',
              member('descr')..type = 'fcs::utils::Fixed_size_char_array<256>',
            ];

    The following common operators are generated.    

    * *operator==*: Equality operator

            bool operator==(Code_packages_value const& rhs) const {
              return this == &rhs || (name == rhs.name && descr == rhs.descr);
            }

    * *operator<*: Less than operator

            bool operator<(Code_packages_value const& rhs) const {
              return name != rhs.name
                         ? name < rhs.name
                         : (descr != rhs.descr ? descr < rhs.descr : (false));
            }

    * *operator<<*: Streaming operator

            friend inline std::ostream& operator<<(std::ostream& out,
                                                   Code_packages_value const& item) {
              out << '\n' << "name:" << item.name;
              out << '\n' << "descr:" << item.descr;
              return out;
            }

* Other types of methods are easily generated.

    * *Constructors initializing members*. For example to add a member
       initializing constructor that initializes the member *name*,
       the following addition will work:

            class_('code_packages_value')
            ..opEqual
            ..opLess
            ..memberCtors = [ memberCtor(['name']) ]
            ...
       Which will cause the addition of:
       
            Code_packages_value(
              fcs::utils::Fixed_size_char_array<64> name) :
              name_ { name } {
            }
       Note, that it is accepting the name by value. Depending on your code this may not be desirable, and in this case it is not. A simple modification to the member will address it:
       
            member('name')
            ..isByRef = true
            ..type = 'fcs::utils::Fixed_size_char_array<64>',
       Now we have pass by reference, much better:

            Code_packages_value(
              fcs::utils::Fixed_size_char_array<64> const & name) :
              name_ { name } {
            }


* *Serialization* Classes can be given support for serialization and the

## Large Scale Code Generation

Perhaps you want to generate all code, business logic as well as
structure. This is a common task when the functionality is very
pattern oriented.

### More on Custom Blocks

Typically, the set of *custom blocks* associated with some type of
entity are determined up front. For instance, for C++ classes you can
specify inclusion of any of the following custom blocks:

    class_('foo')
    ..customBlocks = [clsPreDecl, clsPublic, clsProtected, clsPrivate, clsPostDecl]

Similarly for any C++ file, *Header*, *Impl*, or *App* which is an
*Impl*, you can choose from:

    app('date_time_converter')
    ..namespace = namespace(['fcs'])
    ..customBlocks = [
      fcbCustomIncludes, fcbPreNamespace, fcbPostNamespace,
      fcbBeginNamespace, fcbEndNamespace ]

Essentially the predefined set of *custom blocks* for an entity
comprise the set matching the best guess of the code generation author
as to where additional custom code might be required. The name
indicates the location, so *fcbCustomIncludes* is a protection block
near the top of the file, just below any pre-registered include
statements, allowing the C++ developer to throw in new includes as
they are developing without regenerating. Similarly, *fcbPreNamespace*
is the location just above the namespace declaration containing the
meat of the structure and logic. The interesting thing about these
locations is that they represent not only where a user might like
enter hand written code, they are also the location where a large
scale code generation effort might like to programatically include
more generated content.

For example, suppose you were generating support for a class whose
members are determined programatically.

    _makeClass(String id, Iterable<Column> columns) {
      final result = class_(id)
        ..isStruct = true
        ..opEqual
        ..opLess
        ..isStreamable = true
        ..members = columns.map((c) => _makeMember(c)).toList();
      result.getCodeBlock(clsPublic).snippets
          .add(_stringListSupport(result.members));
      finishClass(result);
      return result;
    }

In this example, like the prior examples a class is created with
*operator==* and *operator<* as well as streaming support,
*operator<<*. However, here the members are not hand coded but rather
pulled in from other data, in this case column names from the schema
of a database of table. Also, note that the *clsPublic* code block
which before was used for a *Custom Block* is used here to inject
*snipets* of code.

## Generated Code Requirements

* *boost* Used for *program_options* to support application
   arguments. Used for *date_time*, and *thread_local_storage* support
   when using database orm facilities.

* *cereal* Used for serialization.

## Philosophies/Goals

* Always support generating code around custom code. Alternatively
  always provide hooks for custom code to be inserted into generated
  code.

* Make usage of *ebisu_cpp* simple enough to be used on tasks where
  hand-coding is still the bulk of the effort, but the option to
  incorporate code generation is still desired

* Where possible steer generation towards good practices

    * Encourage consistent naming
    * Encourage good member encapsulation
    * Encourage namespace usage
    * Encourage consistent directory structure/layout
    * Encourage consistent file structure/layout

## Structural Diagram

<svg version="1.1" viewBox="0.0 0.0 960.0 720.0" fill="none" stroke="none" stroke-linecap="square" stroke-miterlimit="10" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"><clipPath id="p.0"><path d="m0 0l960.0 0l0 720.0l-960.0 0l0 -720.0z" clip-rule="nonzero"></path></clipPath><g clip-path="url(#p.0)"><path fill="#000000" fill-opacity="0.0" d="m0 0l960.0 0l0 720.0l-960.0 0z" fill-rule="nonzero"></path><path fill="#cfe2f3" d="m295.37534 317.72177l88.8819 0l0 24.472443l-88.8819 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="4.0" stroke-linejoin="round" stroke-linecap="butt" d="m295.37534 317.72177l88.8819 0l0 24.472443l-88.8819 0z" fill-rule="nonzero"></path><path fill="#000000" d="m334.57898 334.478l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm2.6972656 -7.375l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm3.6347656 0l-0.984375 0l0 -8.59375l1.0625 0l0 3.0625q0.671875 -0.828125 1.703125 -0.828125q0.578125 0 1.078125 0.234375q0.515625 0.21875 0.84375 0.640625q0.34375 0.421875 0.53125 1.015625q0.1875 0.59375 0.1875 1.265625q0 1.59375 -0.796875 2.46875q-0.796875 0.875 -1.890625 0.875q-1.109375 0 -1.734375 -0.921875l0 0.78125zm-0.015625 -3.15625q0 1.109375 0.3125 1.609375q0.5 0.8125 1.34375 0.8125q0.6875 0 1.1875 -0.59375q0.515625 -0.59375 0.515625 -1.796875q0 -1.21875 -0.484375 -1.796875q-0.484375 -0.578125 -1.171875 -0.578125q-0.6875 0 -1.203125 0.609375q-0.5 0.59375 -0.5 1.734375z" fill-rule="nonzero"></path><path fill="#cfe2f3" d="m340.39108 359.5643l88.8819 0l0 24.472443l-88.8819 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="4.0" stroke-linejoin="round" stroke-linecap="butt" d="m340.39108 359.5643l88.8819 0l0 24.472443l-88.8819 0z" fill-rule="nonzero"></path><path fill="#000000" d="m366.9463 376.32053l0 -8.59375l1.046875 0l0 3.078125q0.734375 -0.84375 1.859375 -0.84375q0.703125 0 1.203125 0.28125q0.515625 0.265625 0.734375 0.75q0.21875 0.46875 0.21875 1.390625l0 3.9375l-1.046875 0l0 -3.9375q0 -0.796875 -0.34375 -1.15625q-0.34375 -0.359375 -0.96875 -0.359375q-0.46875 0 -0.890625 0.25q-0.40625 0.234375 -0.59375 0.65625q-0.171875 0.40625 -0.171875 1.140625l0 3.40625l-1.046875 0zm10.923828 -2.0l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm9.939453 2.9375q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm6.736328 3.125l0 -0.78125q-0.59375 0.921875 -1.734375 0.921875q-0.75 0 -1.375 -0.40625q-0.625 -0.421875 -0.96875 -1.15625q-0.34375 -0.734375 -0.34375 -1.6875q0 -0.921875 0.3125 -1.6875q0.3125 -0.765625 0.9375 -1.15625q0.625 -0.40625 1.390625 -0.40625q0.5625 0 1.0 0.234375q0.4375 0.234375 0.71875 0.609375l0 -3.078125l1.046875 0l0 8.59375l-0.984375 0zm-3.328125 -3.109375q0 1.203125 0.5 1.796875q0.5 0.578125 1.1875 0.578125q0.6875 0 1.171875 -0.5625q0.484375 -0.5625 0.484375 -1.71875q0 -1.28125 -0.5 -1.875q-0.484375 -0.59375 -1.203125 -0.59375q-0.703125 0 -1.171875 0.578125q-0.46875 0.5625 -0.46875 1.796875zm10.220703 1.109375l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.861328 3.703125l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#cfe2f3" d="m384.25723 401.40945l88.88187 0l0 24.472443l-88.88187 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="4.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.25723 401.40945l88.88187 0l0 24.472443l-88.88187 0z" fill-rule="nonzero"></path><path fill="#000000" d="m419.8876 415.88443l1.03125 0.140625q-0.171875 1.0625 -0.875 1.671875q-0.703125 0.609375 -1.71875 0.609375q-1.28125 0 -2.0625 -0.828125q-0.765625 -0.84375 -0.765625 -2.40625q0 -1.0 0.328125 -1.75q0.34375 -0.765625 1.015625 -1.140625q0.6875 -0.375 1.5 -0.375q1.0 0 1.640625 0.515625q0.65625 0.5 0.84375 1.453125l-1.03125 0.15625q-0.140625 -0.625 -0.515625 -0.9375q-0.375 -0.328125 -0.90625 -0.328125q-0.796875 0 -1.296875 0.578125q-0.5 0.5625 -0.5 1.796875q0 1.265625 0.484375 1.828125q0.484375 0.5625 1.25 0.5625q0.625 0 1.03125 -0.375q0.421875 -0.375 0.546875 -1.171875zm1.90625 2.28125l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm6.7597656 -0.765625q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm2.2832031 1.265625l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5zm6.0 0l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5z" fill-rule="nonzero"></path><path fill="#cfe2f3" d="m429.27298 443.2546l88.88187 0l0 24.472412l-88.88187 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="4.0" stroke-linejoin="round" stroke-linecap="butt" d="m429.27298 443.2546l88.88187 0l0 24.472412l-88.88187 0z" fill-rule="nonzero"></path><path fill="#000000" d="m452.5059 460.0108l0 -6.21875l0.9375 0l0 0.875q0.296875 -0.46875 0.78125 -0.734375q0.484375 -0.28125 1.109375 -0.28125q0.6875 0 1.125 0.28125q0.453125 0.28125 0.625 0.796875q0.75 -1.078125 1.921875 -1.078125q0.9375 0 1.421875 0.515625q0.5 0.5 0.5 1.578125l0 4.265625l-1.046875 0l0 -3.921875q0 -0.625 -0.109375 -0.90625q-0.09375 -0.28125 -0.359375 -0.453125q-0.265625 -0.171875 -0.640625 -0.171875q-0.65625 0 -1.09375 0.4375q-0.421875 0.4375 -0.421875 1.40625l0 3.609375l-1.0625 0l0 -4.046875q0 -0.703125 -0.265625 -1.046875q-0.25 -0.359375 -0.828125 -0.359375q-0.453125 0 -0.828125 0.234375q-0.375 0.234375 -0.546875 0.6875q-0.171875 0.453125 -0.171875 1.296875l0 3.234375l-1.046875 0zm14.246094 -2.0l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.876953 3.703125l0 -6.21875l0.9375 0l0 0.875q0.296875 -0.46875 0.78125 -0.734375q0.484375 -0.28125 1.109375 -0.28125q0.6875 0 1.125 0.28125q0.453125 0.28125 0.625 0.796875q0.75 -1.078125 1.921875 -1.078125q0.9375 0 1.421875 0.515625q0.5 0.5 0.5 1.578125l0 4.265625l-1.046875 0l0 -3.921875q0 -0.625 -0.109375 -0.90625q-0.09375 -0.28125 -0.359375 -0.453125q-0.265625 -0.171875 -0.640625 -0.171875q-0.65625 0 -1.09375 0.4375q-0.421875 0.4375 -0.421875 1.40625l0 3.609375l-1.0625 0l0 -4.046875q0 -0.703125 -0.265625 -1.046875q-0.25 -0.359375 -0.828125 -0.359375q-0.453125 0 -0.828125 0.234375q-0.375 0.234375 -0.546875 0.6875q-0.171875 0.453125 -0.171875 1.296875l0 3.234375l-1.046875 0zm10.964844 0l-0.984375 0l0 -8.59375l1.0625 0l0 3.0625q0.671875 -0.828125 1.703125 -0.828125q0.578125 0 1.078125 0.234375q0.515625 0.21875 0.84375 0.640625q0.34375 0.421875 0.53125 1.015625q0.1875 0.59375 0.1875 1.265625q0 1.59375 -0.796875 2.46875q-0.796875 0.875 -1.890625 0.875q-1.109375 0 -1.734375 -0.921875l0 0.78125zm-0.015625 -3.15625q0 1.109375 0.3125 1.609375q0.5 0.8125 1.34375 0.8125q0.6875 0 1.1875 -0.59375q0.515625 -0.59375 0.515625 -1.796875q0 -1.21875 -0.484375 -1.796875q-0.484375 -0.578125 -1.171875 -0.578125q-0.6875 0 -1.203125 0.609375q-0.5 0.59375 -0.5 1.734375zm9.970703 1.15625l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.861328 3.703125l0 -6.21875l0.9531555 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.0469055 0z" fill-rule="nonzero"></path><path fill="#b6d7a8" d="m531.9344 373.04443l0 0c0 -0.6869812 0.55688477 -1.2438965 1.2438965 -1.2438965l86.394104 0c0.32989502 0 0.64630127 0.13104248 0.87957764 0.36431885c0.23327637 0.23327637 0.36431885 0.5496826 0.36431885 0.87957764l0 17.575165c0 0.6870117 -0.5569458 1.243927 -1.2438965 1.243927l-86.394104 0c-0.6870117 0 -1.2438965 -0.5569153 -1.2438965 -1.243927z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m531.9344 373.04443l0 0c0 -0.6869812 0.55688477 -1.2438965 1.2438965 -1.2438965l86.394104 0c0.32989502 0 0.64630127 0.13104248 0.87957764 0.36431885c0.23327637 0.23327637 0.36431885 0.5496826 0.36431885 0.87957764l0 17.575165c0 0.6870117 -0.5569458 1.243927 -1.2438965 1.243927l-86.394104 0c-0.6870117 0 -1.2438965 -0.5569153 -1.2438965 -1.243927z" fill-rule="nonzero"></path><path fill="#000000" d="m556.472 386.35202l0 -7.578125l-2.828125 0l0 -1.015625l6.8125 0l0 1.015625l-2.84375 0l0 7.578125l-1.140625 0zm9.267578 -2.0l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.455078 1.84375l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5zm8.71875 0.921875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm5.0996094 0.171875q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm3.6738281 3.125l-0.984375 0l0 -8.59375l1.0625 0l0 3.0625q0.671875 -0.828125 1.703125 -0.828125q0.578125 0 1.078125 0.234375q0.515625 0.21875 0.84375 0.640625q0.34375 0.421875 0.53125 1.015625q0.1875 0.59375 0.1875 1.265625q0 1.59375 -0.796875 2.46875q-0.796875 0.875 -1.890625 0.875q-1.109375 0 -1.734375 -0.921875l0 0.78125zm-0.015625 -3.15625q0 1.109375 0.3125 1.609375q0.5 0.8125 1.34375 0.8125q0.6875 0 1.1875 -0.59375q0.515625 -0.59375 0.515625 -1.796875q0 -1.21875 -0.484375 -1.796875q-0.484375 -0.578125 -1.171875 -0.578125q-0.6875 0 -1.203125 0.609375q-0.5 0.59375 -0.5 1.734375zm5.689453 3.15625l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm6.9472656 -2.0l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125z" fill-rule="nonzero"></path><path fill="#b6d7a8" d="m635.32544 335.56174l0 0c0 -1.8467712 1.4971313 -3.3439026 3.343933 -3.3439026l96.11536 0c0.8868408 0 1.7373657 0.35232544 2.364502 0.97943115c0.6270752 0.6270752 0.9793701 1.4776306 0.9793701 2.3644714l0 13.375214c0 1.8467712 -1.4971313 3.3439026 -3.343872 3.3439026l-96.11536 0l0 0c-1.8468018 0 -3.343933 -1.4971313 -3.343933 -3.3439026z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m635.32544 335.56174l0 0c0 -1.8467712 1.4971313 -3.3439026 3.343933 -3.3439026l96.11536 0c0.8868408 0 1.7373657 0.35232544 2.364502 0.97943115c0.6270752 0.6270752 0.9793701 1.4776306 0.9793701 2.3644714l0 13.375214c0 1.8467712 -1.4971313 3.3439026 -3.343872 3.3439026l-96.11536 0l0 0c-1.8468018 0 -3.343933 -1.4971313 -3.343933 -3.3439026z" fill-rule="nonzero"></path><path fill="#000000" d="m652.8198 346.76935l0 -7.578125l-2.828125 0l0 -1.015625l6.8125 0l0 1.015625l-2.84375 0l0 7.578125l-1.140625 0zm9.267578 -2.0l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.455078 1.84375l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5zm8.71875 0.921875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm4.1210938 -1.828125l1.0625 -0.09375q0.078125 0.65625 0.359375 1.0625q0.28125 0.40625 0.859375 0.671875q0.59375 0.25 1.328125 0.25q0.640625 0 1.140625 -0.1875q0.5 -0.203125 0.734375 -0.53125q0.25 -0.34375 0.25 -0.734375q0 -0.40625 -0.234375 -0.703125q-0.234375 -0.3125 -0.765625 -0.515625q-0.359375 -0.140625 -1.546875 -0.421875q-1.171875 -0.28125 -1.640625 -0.53125q-0.625 -0.328125 -0.921875 -0.796875q-0.296875 -0.484375 -0.296875 -1.078125q0 -0.640625 0.359375 -1.203125q0.375 -0.578125 1.078125 -0.859375q0.71875 -0.296875 1.578125 -0.296875q0.953125 0 1.6875 0.3125q0.734375 0.296875 1.125 0.90625q0.390625 0.59375 0.421875 1.34375l-1.09375 0.078125q-0.09375 -0.8125 -0.609375 -1.21875q-0.5 -0.421875 -1.484375 -0.421875q-1.03125 0 -1.5 0.375q-0.46875 0.375 -0.46875 0.90625q0 0.46875 0.328125 0.765625q0.328125 0.296875 1.703125 0.609375q1.390625 0.3125 1.90625 0.546875q0.75 0.359375 1.109375 0.890625q0.359375 0.515625 0.359375 1.21875q0 0.6875 -0.390625 1.296875q-0.390625 0.59375 -1.125 0.9375q-0.734375 0.328125 -1.65625 0.328125q-1.171875 0 -1.96875 -0.328125q-0.78125 -0.34375 -1.234375 -1.03125q-0.4375 -0.6875 -0.453125 -1.546875zm12.316406 0.484375l1.03125 0.140625q-0.171875 1.0625 -0.875 1.671875q-0.703125 0.609375 -1.71875 0.609375q-1.28125 0 -2.0625 -0.828125q-0.765625 -0.84375 -0.765625 -2.40625q0 -1.0 0.328125 -1.75q0.34375 -0.765625 1.015625 -1.140625q0.6875 -0.375 1.5 -0.375q1.0 0 1.640625 0.515625q0.65625 0.5 0.84375 1.453125l-1.03125 0.15625q-0.140625 -0.625 -0.515625 -0.9375q-0.375 -0.328125 -0.90625 -0.328125q-0.796875 0 -1.296875 0.578125q-0.5 0.5625 -0.5 1.796875q0 1.265625 0.484375 1.828125q0.484375 0.5625 1.25 0.5625q0.625 0 1.03125 -0.375q0.421875 -0.375 0.546875 -1.171875zm6.1875 0.28125l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.876953 3.703125l0 -6.21875l0.9375 0l0 0.875q0.6875 -1.015625 1.984375 -1.015625q0.5625 0 1.03125 0.203125q0.484375 0.203125 0.71875 0.53125q0.234375 0.328125 0.328125 0.765625q0.046875 0.296875 0.046875 1.03125l0 3.828125l-1.046875 0l0 -3.78125q0 -0.65625 -0.125 -0.96875q-0.125 -0.3125 -0.4375 -0.5q-0.3125 -0.203125 -0.734375 -0.203125q-0.671875 0 -1.171875 0.4375q-0.484375 0.421875 -0.484375 1.609375l0 3.40625l-1.046875 0zm10.736328 -0.765625q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm2.6894531 3.125l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0zm4.0117188 -7.375l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm2.2753906 -3.109375q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m473.1391 413.64566l29.39746 0l0 -31.811005l29.4057 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m473.1391 413.64566l29.39746 0l0 -31.811005l29.4057 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m429.27298 371.80054l28.640747 0l0 10.047241l74.03641 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m429.27298 371.80054l28.640747 0l0 10.047241l74.03641 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m384.83203 384.03674l0 8.68631l43.873993 0l0 8.699524" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.83203 384.03674l0 8.68631l43.873993 0l0 8.699524" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m428.69815 425.8819l0 8.68631l45.007874 0l0 8.699524" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m428.69815 425.8819l0 8.68631l45.007874 0l0 8.699524" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m473.7139 467.72702l0 25.120636l-45.007874 0l0 25.1156" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m473.7139 467.72702l0 25.120636l-45.007874 0l0 25.1156" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m473.7139 467.72702l0 25.120636l42.897675 0l0 25.1156" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m473.7139 467.72702l0 25.120636l42.897675 0l0 25.1156" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m576.3753 371.80054l0 -29.543304l58.960632 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m576.3753 371.80054l0 -29.543304l58.960632 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m339.81628 342.1942l0 8.685089l45.007874 0l0 8.700745" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m339.81628 342.1942l0 8.685089l45.007874 0l0 8.700745" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m572.1942 315.23096l54.078735 0l0 24.472443l-54.078735 0z" fill-rule="nonzero"></path><path fill="#000000" d="m584.8192 333.19098l0 -0.59375q-0.453125 0.703125 -1.3125 0.703125q-0.546875 0 -1.015625 -0.3125q-0.46875 -0.3125 -0.734375 -0.859375q-0.25 -0.546875 -0.25 -1.265625q0 -0.703125 0.234375 -1.265625q0.234375 -0.578125 0.6875 -0.875q0.46875 -0.296875 1.046875 -0.296875q0.421875 0 0.75 0.171875q0.328125 0.171875 0.53125 0.46875l0 -2.3125l0.796875 0l0 6.4375l-0.734375 0zm-2.5 -2.328125q0 0.890625 0.375 1.34375q0.375 0.4375 0.890625 0.4375q0.515625 0 0.875 -0.421875q0.375 -0.421875 0.375 -1.296875q0 -0.953125 -0.375 -1.40625q-0.375 -0.453125 -0.90625 -0.453125q-0.53125 0 -0.890625 0.4375q-0.34375 0.4375 -0.34375 1.359375zm7.844116 0.828125l0.828125 0.09375q-0.203125 0.71875 -0.734375 1.125q-0.515625 0.390625 -1.328125 0.390625q-1.015625 0 -1.625 -0.625q-0.59375 -0.640625 -0.59375 -1.78125q0 -1.171875 0.609375 -1.8125q0.609375 -0.65625 1.578125 -0.65625q0.9375 0 1.515625 0.640625q0.59375 0.625 0.59375 1.78125q0 0.078125 0 0.21875l-3.484375 0q0.046875 0.765625 0.4375 1.171875q0.390625 0.40625 0.984375 0.40625q0.4375 0 0.734375 -0.21875q0.3125 -0.234375 0.484375 -0.734375zm-2.59375 -1.28125l2.609375 0q-0.046875 -0.59375 -0.296875 -0.890625q-0.375 -0.453125 -0.984375 -0.453125q-0.546875 0 -0.921875 0.375q-0.359375 0.359375 -0.40625 0.96875zm4.781616 2.78125l0 -4.046875l-0.703125 0l0 -0.625l0.703125 0l0 -0.484375q0 -0.484375 0.078125 -0.703125q0.125 -0.3125 0.40625 -0.5q0.296875 -0.1875 0.8125 -0.1875q0.328125 0 0.734375 0.078125l-0.125 0.6875q-0.234375 -0.046875 -0.453125 -0.046875q-0.359375 0 -0.515625 0.15625q-0.15625 0.15625 -0.15625 0.578125l0 0.421875l0.921875 0l0 0.625l-0.921875 0l0 4.046875l-0.78125 0zm2.4041748 -5.53125l0 -0.90625l0.796875 0l0 0.90625l-0.796875 0zm0 5.53125l0 -4.671875l0.796875 0l0 4.671875l-0.796875 0zm2.0723877 0l0 -4.671875l0.71875 0l0 0.671875q0.5 -0.765625 1.484375 -0.765625q0.421875 0 0.765625 0.15625q0.359375 0.140625 0.53125 0.390625q0.171875 0.234375 0.25 0.578125q0.046875 0.21875 0.046875 0.765625l0 2.875l-0.796875 0l0 -2.84375q0 -0.484375 -0.09375 -0.71875q-0.09375 -0.234375 -0.328125 -0.375q-0.234375 -0.15625 -0.546875 -0.15625q-0.5 0 -0.875 0.328125q-0.359375 0.3125 -0.359375 1.21875l0 2.546875l-0.796875 0zm8.375366 -1.5l0.828125 0.09375q-0.203125 0.71875 -0.734375 1.125q-0.515625 0.390625 -1.328125 0.390625q-1.015625 0 -1.625 -0.625q-0.59375 -0.640625 -0.59375 -1.78125q0 -1.171875 0.609375 -1.8125q0.609375 -0.65625 1.578125 -0.65625q0.9375 0 1.515625 0.640625q0.59375 0.625 0.59375 1.78125q0 0.078125 0 0.21875l-3.484375 0q0.046875 0.765625 0.4375 1.171875q0.390625 0.40625 0.984375 0.40625q0.4375 0 0.734375 -0.21875q0.3125 -0.234375 0.484375 -0.734375zm-2.59375 -1.28125l2.609375 0q-0.046875 -0.59375 -0.296875 -0.890625q-0.375 -0.453125 -0.984375 -0.453125q-0.546875 0 -0.921875 0.375q-0.359375 0.359375 -0.40625 0.96875zm4.281616 1.390625l0.78125 -0.125q0.0625 0.46875 0.359375 0.71875q0.3125 0.25 0.84375 0.25q0.546875 0 0.8125 -0.21875q0.265625 -0.21875 0.265625 -0.515625q0 -0.28125 -0.234375 -0.421875q-0.171875 -0.109375 -0.8125 -0.28125q-0.875 -0.21875 -1.21875 -0.375q-0.328125 -0.15625 -0.5 -0.4375q-0.171875 -0.28125 -0.171875 -0.625q0 -0.3125 0.140625 -0.578125q0.140625 -0.265625 0.390625 -0.453125q0.1875 -0.125 0.5 -0.21875q0.328125 -0.09375 0.6875 -0.09375q0.546875 0 0.953125 0.15625q0.421875 0.15625 0.625 0.421875q0.203125 0.265625 0.28125 0.71875l-0.78125 0.109375q-0.046875 -0.359375 -0.3125 -0.5625q-0.25 -0.203125 -0.703125 -0.203125q-0.546875 0 -0.78125 0.1875q-0.234375 0.171875 -0.234375 0.421875q0 0.15625 0.09375 0.265625q0.09375 0.140625 0.3125 0.21875q0.109375 0.046875 0.6875 0.203125q0.84375 0.21875 1.171875 0.359375q0.34375 0.140625 0.53125 0.421875q0.1875 0.265625 0.1875 0.671875q0 0.40625 -0.234375 0.75q-0.234375 0.34375 -0.671875 0.546875q-0.421875 0.1875 -0.984375 0.1875q-0.90625 0 -1.390625 -0.375q-0.46875 -0.390625 -0.59375 -1.125z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m491.4672 356.62466l34.771637 0l0 24.472443l-34.771637 0z" fill-rule="nonzero"></path><path fill="#000000" d="m501.06094 369.0534l0 -0.90625l0.796875 0l0 0.90625l-0.796875 0zm0 5.53125l0 -4.671875l0.796875 0l0 4.671875l-0.796875 0zm1.7599182 -1.390625l0.78125 -0.125q0.0625 0.46875 0.359375 0.71875q0.3125 0.25 0.84375 0.25q0.546875 0 0.8125 -0.21875q0.265625 -0.21875 0.265625 -0.515625q0 -0.28125 -0.234375 -0.421875q-0.171875 -0.109375 -0.8125 -0.28125q-0.875 -0.21875 -1.21875 -0.375q-0.328125 -0.15625 -0.5 -0.4375q-0.171875 -0.28125 -0.171875 -0.625q0 -0.3125 0.140625 -0.578125q0.140625 -0.265625 0.390625 -0.453125q0.1875 -0.125 0.5 -0.21875q0.328125 -0.09375 0.6875 -0.09375q0.546875 0 0.953125 0.15625q0.421875 0.15625 0.625 0.421875q0.203125 0.265625 0.28125 0.71875l-0.78125 0.109375q-0.046875 -0.359375 -0.3125 -0.5625q-0.25 -0.203125 -0.703125 -0.203125q-0.546875 0 -0.78125 0.1875q-0.234375 0.171875 -0.234375 0.421875q0 0.15625 0.09375 0.265625q0.09375 0.140625 0.3125 0.21875q0.109375 0.046875 0.6875 0.203125q0.84375 0.21875 1.171875 0.359375q0.34375 0.140625 0.53125 0.421875q0.1875 0.265625 0.1875 0.671875q0 0.40625 -0.234375 0.75q-0.234375 0.34375 -0.671875 0.546875q-0.421875 0.1875 -0.984375 0.1875q-0.90625 0 -1.390625 -0.375q-0.46875 -0.390625 -0.59375 -1.125z" fill-rule="nonzero"></path><path fill="#cfe2f3" d="m260.59317 275.87927l88.8819 0l0 24.472443l-88.8819 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="4.0" stroke-linejoin="round" stroke-linecap="butt" d="m260.59317 275.87927l88.8819 0l0 24.472443l-88.8819 0z" fill-rule="nonzero"></path><path fill="#000000" d="m277.4804 285.2605l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm2.6660156 0l0 -6.21875l0.9375 0l0 0.875q0.6875 -1.015625 1.984375 -1.015625q0.5625 0 1.03125 0.203125q0.484375 0.203125 0.71875 0.53125q0.234375 0.328125 0.328125 0.765625q0.046875 0.296875 0.046875 1.03125l0 3.828125l-1.046875 0l0 -3.78125q0 -0.65625 -0.125 -0.96875q-0.125 -0.3125 -0.4375 -0.5q-0.3125 -0.203125 -0.734375 -0.203125q-0.671875 0 -1.171875 0.4375q-0.484375 0.421875 -0.484375 1.609375l0 3.40625l-1.046875 0zm6.251953 -1.859375l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5zm8.71875 0.921875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm5.0996094 0.171875q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm2.6738281 3.125l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm2.6660156 0l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm6.7597656 -0.765625q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm5.001953 2.1875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm1.0371094 -6.4375l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm2.2753906 -3.109375q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.986328 3.109375l0 -6.21875l0.9375 0l0 0.875q0.6875 -1.015625 1.984375 -1.015625q0.5625 0 1.03125 0.203125q0.484375 0.203125 0.71875 0.53125q0.234375 0.328125 0.328125 0.765625q0.046875 0.296875 0.046875 1.03125l0 3.828125l-1.046875 0l0 -3.78125q0 -0.65625 -0.125 -0.96875q-0.125 -0.3125 -0.4375 -0.5q-0.3125 -0.203125 -0.734375 -0.203125q-0.671875 0 -1.171875 0.4375q-0.484375 0.421875 -0.484375 1.609375l0 3.40625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m305.03412 300.3517l0 8.685089l34.771667 0l0 8.700714" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m305.03412 300.3517l0 8.685089l34.771667 0l0 8.700714" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m384.25723 329.958l73.83807 0l0 51.874023l73.846954 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.25723 329.958l73.83807 0l0 51.874023l73.846954 0" fill-rule="nonzero"></path><path fill="#d9ead3" d="m597.98425 414.28348l147.68506 0l0 156.31497l-147.68506 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" stroke-dasharray="8.0,6.0" d="m597.98425 414.28348l147.68506 0l0 156.31497l-147.68506 0z" fill-rule="nonzero"></path><path fill="#000000" d="m607.8124 431.57034l-0.734375 0l0 -1.1875l0.96875 0l0.328125 -1.59375l-1.296875 0l0 -1.171875l1.53125 0l0.421875 -2.109375l1.203125 0l-0.421875 2.109375l1.1875 0l0.421875 -2.109375l1.234375 0l-0.421875 2.109375l0.734375 0l0 1.171875l-0.984375 0l-0.3125 1.59375l1.296875 0l0 1.1875l-1.546875 0l-0.421875 2.09375l-1.203125 0l0.421875 -2.09375l-1.1875 0l-0.4375 2.09375l-1.203125 0l0.421875 -2.09375zm1.765625 -2.78125l-0.3125 1.59375l1.1875 0l0.3125 -1.59375l-1.1875 0zm4.1296387 -1.75l0 -1.390625l1.5 0l0 1.390625l-1.5 0zm0 6.484375l0 -5.703125l1.5 0l0 5.703125l-1.5 0zm8.148132 0l-1.515625 0l0 -2.90625q0 -0.921875 -0.09375 -1.1875q-0.09375 -0.28125 -0.3125 -0.421875q-0.21875 -0.15625 -0.53125 -0.15625q-0.390625 0 -0.703125 0.21875q-0.3125 0.203125 -0.4375 0.5625q-0.109375 0.34375 -0.109375 1.3125l0 2.578125l-1.5 0l0 -5.703125l1.40625 0l0 0.84375q0.734375 -0.96875 1.875 -0.96875q0.5 0 0.90625 0.1875q0.421875 0.171875 0.625 0.453125q0.21875 0.265625 0.296875 0.625q0.09375 0.359375 0.09375 1.015625l0 3.546875zm6.2905273 -4.015625l-1.484375 0.265625q-0.078125 -0.4375 -0.34375 -0.65625q-0.265625 -0.234375 -0.6875 -0.234375q-0.5625 0 -0.90625 0.390625q-0.328125 0.375 -0.328125 1.296875q0 1.015625 0.328125 1.4375q0.34375 0.421875 0.921875 0.421875q0.4375 0 0.703125 -0.25q0.28125 -0.25 0.390625 -0.84375l1.484375 0.25q-0.234375 1.03125 -0.890625 1.546875q-0.65625 0.515625 -1.75 0.515625q-1.265625 0 -2.015625 -0.78125q-0.734375 -0.796875 -0.734375 -2.1875q0 -1.421875 0.75 -2.203125q0.75 -0.78125 2.015625 -0.78125q1.046875 0 1.65625 0.453125q0.625 0.4375 0.890625 1.359375zm0.95770264 4.015625l0 -7.875l1.5 0l0 7.875l-1.5 0zm6.7106934 0l0 -0.859375q-0.3125 0.46875 -0.828125 0.734375q-0.5 0.25 -1.0625 0.25q-0.578125 0 -1.03125 -0.25q-0.453125 -0.25 -0.671875 -0.703125q-0.203125 -0.46875 -0.203125 -1.265625l0 -3.609375l1.515625 0l0 2.625q0 1.203125 0.078125 1.46875q0.09375 0.265625 0.3125 0.4375q0.21875 0.15625 0.5625 0.15625q0.375 0 0.6875 -0.203125q0.3125 -0.21875 0.421875 -0.53125q0.109375 -0.328125 0.109375 -1.546875l0 -2.40625l1.515625 0l0 5.703125l-1.40625 0zm7.9779663 0l-1.390625 0l0 -0.84375q-0.359375 0.5 -0.84375 0.734375q-0.46875 0.234375 -0.9375 0.234375q-0.984375 0 -1.6875 -0.78125q-0.703125 -0.796875 -0.703125 -2.21875q0 -1.453125 0.671875 -2.203125q0.6875 -0.75 1.734375 -0.75q0.953125 0 1.65625 0.796875l0 -2.84375l1.5 0l0 7.875zm-4.015625 -2.96875q0 0.90625 0.25 1.3125q0.359375 0.59375 1.015625 0.59375q0.515625 0 0.875 -0.4375q0.375 -0.453125 0.375 -1.328125q0 -0.984375 -0.359375 -1.40625q-0.34375 -0.4375 -0.90625 -0.4375q-0.53125 0 -0.890625 0.421875q-0.359375 0.421875 -0.359375 1.28125zm8.603027 1.15625l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm6.934021 -1.34375l-1.515625 0l0 -1.078125q0 -0.65625 0.109375 -1.03125q0.125 -0.375 0.4375 -0.671875q0.3125 -0.3125 0.796875 -0.484375l0.296875 0.625q-0.453125 0.15625 -0.65625 0.421875q-0.1875 0.265625 -0.203125 0.71875l0.734375 0l0 1.5zm2.515625 0l-1.5 0l0 -1.078125q0 -0.65625 0.109375 -1.03125q0.109375 -0.375 0.421875 -0.671875q0.328125 -0.3125 0.796875 -0.484375l0.296875 0.625q-0.453125 0.15625 -0.65625 0.421875q-0.1875 0.265625 -0.203125 0.71875l0.734375 0l0 1.5zm6.359375 0.671875l-1.484375 0.265625q-0.078125 -0.4375 -0.34375 -0.65625q-0.265625 -0.234375 -0.6875 -0.234375q-0.5625 0 -0.90625 0.390625q-0.328125 0.375 -0.328125 1.296875q0 1.015625 0.328125 1.4375q0.34375 0.421875 0.921875 0.421875q0.4375 0 0.703125 -0.25q0.28125 -0.25 0.390625 -0.84375l1.484375 0.25q-0.234375 1.03125 -0.890625 1.546875q-0.65625 0.515625 -1.75 0.515625q-1.265625 0 -2.015625 -0.78125q-0.734375 -0.796875 -0.734375 -2.1875q0 -1.421875 0.75 -2.203125q0.75 -0.78125 2.015625 -0.78125q1.046875 0 1.65625 0.453125q0.625 0.4375 0.890625 1.359375zm2.0827637 0.046875l-1.375 -0.25q0.234375 -0.8125 0.796875 -1.203125q0.5625 -0.40625 1.671875 -0.40625q1.015625 0 1.5 0.234375q0.5 0.234375 0.703125 0.609375q0.203125 0.359375 0.203125 1.34375l-0.015625 1.765625q0 0.75 0.0625 1.109375q0.078125 0.359375 0.28125 0.765625l-1.5 0q-0.0625 -0.15625 -0.140625 -0.453125q-0.03125 -0.125 -0.046875 -0.171875q-0.390625 0.375 -0.84375 0.5625q-0.4375 0.1875 -0.9375 0.1875q-0.875 0 -1.390625 -0.46875q-0.5 -0.484375 -0.5 -1.21875q0 -0.484375 0.234375 -0.859375q0.234375 -0.375 0.640625 -0.578125q0.421875 -0.203125 1.203125 -0.359375q1.0625 -0.1875 1.46875 -0.359375l0 -0.15625q0 -0.4375 -0.21875 -0.609375q-0.21875 -0.1875 -0.8125 -0.1875q-0.390625 0 -0.625 0.15625q-0.21875 0.15625 -0.359375 0.546875zm2.015625 1.234375q-0.296875 0.09375 -0.921875 0.234375q-0.625 0.125 -0.8125 0.25q-0.296875 0.21875 -0.296875 0.53125q0 0.328125 0.234375 0.5625q0.234375 0.21875 0.59375 0.21875q0.40625 0 0.78125 -0.265625q0.28125 -0.203125 0.359375 -0.5q0.0625 -0.1875 0.0625 -0.734375l0 -0.296875zm5.3952637 -2.96875l0 1.203125l-1.03125 0l0 2.296875q0 0.703125 0.03125 0.828125q0.03125 0.109375 0.125 0.1875q0.109375 0.0625 0.265625 0.0625q0.203125 0 0.609375 -0.140625l0.125 1.171875q-0.53125 0.21875 -1.1875 0.21875q-0.421875 0 -0.75 -0.125q-0.328125 -0.140625 -0.484375 -0.359375q-0.140625 -0.21875 -0.203125 -0.59375q-0.046875 -0.265625 -0.046875 -1.0625l0 -2.484375l-0.6875 0l0 -1.203125l0.6875 0l0 -1.140625l1.515625 -0.875l0 2.015625l1.03125 0zm5.907959 1.6875l-1.484375 0.265625q-0.078125 -0.4375 -0.34375 -0.65625q-0.265625 -0.234375 -0.6875 -0.234375q-0.5625 0 -0.90625 0.390625q-0.328125 0.375 -0.328125 1.296875q0 1.015625 0.328125 1.4375q0.34375 0.421875 0.921875 0.421875q0.4375 0 0.703125 -0.25q0.28125 -0.25 0.390625 -0.84375l1.484375 0.25q-0.234375 1.03125 -0.890625 1.546875q-0.65625 0.515625 -1.75 0.515625q-1.265625 0 -2.015625 -0.78125q-0.734375 -0.796875 -0.734375 -2.1875q0 -1.421875 0.75 -2.203125q0.75 -0.78125 2.015625 -0.78125q1.046875 0 1.65625 0.453125q0.625 0.4375 0.890625 1.359375zm2.4577637 -3.859375l0 2.890625q0.734375 -0.84375 1.75 -0.84375q0.515625 0 0.921875 0.1875q0.421875 0.1875 0.640625 0.5q0.21875 0.296875 0.296875 0.671875q0.078125 0.359375 0.078125 1.125l0 3.34375l-1.515625 0l0 -3.015625q0 -0.890625 -0.09375 -1.125q-0.078125 -0.25 -0.296875 -0.390625q-0.21875 -0.140625 -0.546875 -0.140625q-0.375 0 -0.671875 0.1875q-0.296875 0.171875 -0.4375 0.546875q-0.125 0.359375 -0.125 1.078125l0 2.859375l-1.515625 0l0 -7.875l1.515625 0zm5.0092773 7.875l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm4.4606323 -7.875l0 2.890625q0.734375 -0.84375 1.75 -0.84375q0.515625 0 0.921875 0.1875q0.421875 0.1875 0.640625 0.5q0.21875 0.296875 0.296875 0.671875q0.078125 0.359375 0.078125 1.125l0 3.34375l-1.515625 0l0 -3.015625q0 -0.890625 -0.09375 -1.125q-0.078125 -0.25 -0.296875 -0.390625q-0.21875 -0.140625 -0.546875 -0.140625q-0.375 0 -0.671875 0.1875q-0.296875 0.171875 -0.4375 0.546875q-0.125 0.359375 -0.125 1.078125l0 2.859375l-1.515625 0l0 -7.875l1.515625 0zm4.9624023 2.171875l1.40625 0l0 0.84375q0.265625 -0.4375 0.734375 -0.703125q0.46875 -0.265625 1.046875 -0.265625q0.984375 0 1.671875 0.78125q0.703125 0.765625 0.703125 2.15625q0 1.4375 -0.703125 2.234375q-0.6875 0.78125 -1.6875 0.78125q-0.46875 0 -0.859375 -0.1875q-0.375 -0.1875 -0.8125 -0.640625l0 2.875l-1.5 0l0 -7.875zm1.484375 2.75q0 0.96875 0.375 1.4375q0.390625 0.453125 0.9375 0.453125q0.53125 0 0.875 -0.421875q0.359375 -0.421875 0.359375 -1.390625q0 -0.890625 -0.359375 -1.328125q-0.359375 -0.4375 -0.890625 -0.4375q-0.5625 0 -0.9375 0.4375q-0.359375 0.421875 -0.359375 1.25zm5.0248413 -2.75l1.40625 0l0 0.84375q0.265625 -0.4375 0.734375 -0.703125q0.46875 -0.265625 1.046875 -0.265625q0.984375 0 1.671875 0.78125q0.703125 0.765625 0.703125 2.15625q0 1.4375 -0.703125 2.234375q-0.6875 0.78125 -1.6875 0.78125q-0.46875 0 -0.859375 -0.1875q-0.375 -0.1875 -0.8125 -0.640625l0 2.875l-1.5 0l0 -7.875zm1.484375 2.75q0 0.96875 0.375 1.4375q0.390625 0.453125 0.9375 0.453125q0.53125 0 0.875 -0.421875q0.359375 -0.421875 0.359375 -1.390625q0 -0.890625 -0.359375 -1.328125q-0.359375 -0.4375 -0.890625 -0.4375q-0.5625 0 -0.9375 0.4375q-0.359375 0.421875 -0.359375 1.25zm7.4936523 -4.921875l1.515625 0l0 1.078125q0 0.65625 -0.125 1.046875q-0.109375 0.375 -0.421875 0.671875q-0.3125 0.296875 -0.796875 0.46875l-0.296875 -0.625q0.453125 -0.140625 0.640625 -0.40625q0.203125 -0.28125 0.21875 -0.71875l-0.734375 0l0 -1.515625zm-2.53125 0l1.515625 0l0 1.078125q0 0.65625 -0.125 1.046875q-0.109375 0.375 -0.421875 0.671875q-0.3125 0.296875 -0.796875 0.46875l-0.296875 -0.625q0.453125 -0.140625 0.640625 -0.40625q0.203125 -0.28125 0.21875 -0.71875l-0.734375 0l0 -1.515625z" fill-rule="nonzero"></path><path fill="#000000" d="m607.3749 443.96097l1.546875 -0.15625q0.140625 0.78125 0.5625 1.15625q0.4375 0.359375 1.15625 0.359375q0.78125 0 1.15625 -0.328125q0.390625 -0.328125 0.390625 -0.75q0 -0.28125 -0.171875 -0.46875q-0.15625 -0.203125 -0.5625 -0.359375q-0.28125 -0.09375 -1.265625 -0.34375q-1.28125 -0.3125 -1.796875 -0.78125q-0.71875 -0.640625 -0.71875 -1.578125q0 -0.59375 0.328125 -1.109375q0.34375 -0.53125 0.984375 -0.8125q0.640625 -0.28125 1.5625 -0.28125q1.46875 0 2.21875 0.65625q0.75 0.65625 0.796875 1.734375l-1.59375 0.078125q-0.109375 -0.609375 -0.4375 -0.875q-0.328125 -0.265625 -1.0 -0.265625q-0.703125 0 -1.09375 0.28125q-0.25 0.1875 -0.25 0.484375q0 0.28125 0.234375 0.484375q0.296875 0.25 1.453125 0.53125q1.171875 0.265625 1.71875 0.5625q0.5625 0.28125 0.875 0.796875q0.3125 0.5 0.3125 1.25q0 0.6875 -0.375 1.28125q-0.375 0.578125 -1.0625 0.875q-0.6875 0.28125 -1.71875 0.28125q-1.484375 0 -2.296875 -0.6875q-0.796875 -0.6875 -0.953125 -2.015625zm12.56073 -0.328125l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm2.6487427 2.890625l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm7.123291 0l0 -7.875l1.546875 0l3.234375 5.265625l0 -5.265625l1.46875 0l0 7.875l-1.59375 0l-3.171875 -5.140625l0 5.140625l-1.484375 0zm14.789307 0l-1.734375 0l-0.6875 -1.78125l-3.140625 0l-0.65625 1.78125l-1.6875 0l3.0625 -7.875l1.6875 0l3.15625 7.875zm-2.9375 -3.109375l-1.078125 -2.921875l-1.0625 2.921875l2.140625 0zm3.5393677 3.109375l0 -7.875l3.34375 0q1.265625 0 1.828125 0.21875q0.578125 0.203125 0.921875 0.75q0.34375 0.546875 0.34375 1.234375q0 0.890625 -0.53125 1.46875q-0.515625 0.578125 -1.546875 0.734375q0.515625 0.296875 0.84375 0.65625q0.34375 0.359375 0.90625 1.28125l0.96875 1.53125l-1.90625 0l-1.15625 -1.71875q-0.609375 -0.90625 -0.84375 -1.140625q-0.21875 -0.25 -0.46875 -0.328125q-0.25 -0.09375 -0.796875 -0.09375l-0.328125 0l0 3.28125l-1.578125 0zm1.578125 -4.546875l1.1875 0q1.140625 0 1.421875 -0.09375q0.28125 -0.09375 0.4375 -0.328125q0.171875 -0.234375 0.171875 -0.59375q0 -0.40625 -0.21875 -0.640625q-0.203125 -0.25 -0.59375 -0.3125q-0.1875 -0.03125 -1.15625 -0.03125l-1.25 0l0 2.0zm6.0549927 4.546875l0 -7.875l1.59375 0l0 7.875l-1.59375 0zm2.6950073 -3.890625q0 -1.203125 0.359375 -2.015625q0.265625 -0.609375 0.71875 -1.078125q0.46875 -0.484375 1.03125 -0.71875q0.734375 -0.3125 1.6875 -0.3125q1.75 0 2.78125 1.09375q1.046875 1.078125 1.046875 3.0q0 1.90625 -1.03125 2.984375q-1.03125 1.078125 -2.765625 1.078125q-1.765625 0 -2.796875 -1.078125q-1.03125 -1.078125 -1.03125 -2.953125zm1.625 -0.046875q0 1.328125 0.625 2.03125q0.625 0.6875 1.5625 0.6875q0.953125 0 1.5625 -0.6875q0.609375 -0.6875 0.609375 -2.0625q0 -1.34375 -0.59375 -2.015625q-0.59375 -0.671875 -1.578125 -0.671875q-0.984375 0 -1.59375 0.6875q-0.59375 0.671875 -0.59375 2.03125zm9.476257 6.25l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm1.048645 -2.3125l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.5388184 2.3125q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0zm6.4623413 -4.515625l0 -1.28125q0.390625 -0.03125 0.578125 -0.125q0.203125 -0.09375 0.34375 -0.3125q0.15625 -0.234375 0.203125 -0.578125q0.046875 -0.265625 0.046875 -0.90625q0 -1.046875 0.09375 -1.453125q0.09375 -0.40625 0.34375 -0.65625q0.265625 -0.25 0.75 -0.390625q0.328125 -0.109375 1.03125 -0.109375l0.28125 0l0 1.296875q-0.59375 0 -0.765625 0.0625q-0.171875 0.0625 -0.25 0.203125q-0.078125 0.140625 -0.078125 0.46875q0 0.34375 -0.046875 1.28125q-0.03125 0.53125 -0.15625 0.875q-0.109375 0.328125 -0.28125 0.546875q-0.171875 0.203125 -0.53125 0.4375q0.3125 0.171875 0.515625 0.421875q0.203125 0.25 0.3125 0.59375q0.109375 0.34375 0.140625 0.9375q0.03125 0.890625 0.03125 1.140625q0 0.359375 0.078125 0.484375q0.09375 0.140625 0.265625 0.21875q0.1875 0.078125 0.765625 0.078125l0 1.28125l-0.28125 0q-0.734375 0 -1.125 -0.125q-0.375 -0.109375 -0.640625 -0.375q-0.265625 -0.265625 -0.359375 -0.671875q-0.09375 -0.390625 -0.09375 -1.234375q0 -0.984375 -0.09375 -1.28125q-0.109375 -0.4375 -0.359375 -0.609375q-0.234375 -0.1875 -0.71875 -0.21875z" fill-rule="nonzero"></path><path fill="#000000" d="m617.37427 456.63284l0 -1.328125l3.421875 0l0 3.140625q-0.5 0.46875 -1.453125 0.84375q-0.9375 0.375 -1.921875 0.375q-1.234375 0 -2.15625 -0.515625q-0.90625 -0.53125 -1.375 -1.484375q-0.453125 -0.96875 -0.453125 -2.109375q0 -1.21875 0.515625 -2.171875q0.515625 -0.96875 1.5 -1.46875q0.765625 -0.40625 1.890625 -0.40625q1.46875 0 2.28125 0.625q0.828125 0.609375 1.0625 1.703125l-1.578125 0.296875q-0.15625 -0.59375 -0.625 -0.921875q-0.453125 -0.34375 -1.140625 -0.34375q-1.046875 0 -1.65625 0.671875q-0.609375 0.65625 -0.609375 1.953125q0 1.40625 0.609375 2.109375q0.625 0.703125 1.640625 0.703125q0.5 0 1.0 -0.1875q0.5 -0.203125 0.859375 -0.484375l0 -1.0l-1.8125 0zm4.5700684 2.890625l0 -7.875l1.59375 0l0 7.875l-1.59375 0zm5.0231323 0l-2.8125 -7.875l1.71875 0l2.0 5.828125l1.921875 -5.828125l1.6875 0l-2.828125 7.875l-1.6875 0zm5.09198 0l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm7.123291 0l0 -7.875l1.546875 0l3.234375 5.265625l0 -5.265625l1.46875 0l0 7.875l-1.59375 0l-3.171875 -5.140625l0 5.140625l-1.484375 0zm10.179993 2.3125l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm1.048584 -2.3125l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606934 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.5387573 2.3125q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0zm6.4624023 -4.515625l0 -1.28125q0.390625 -0.03125 0.578125 -0.125q0.203125 -0.09375 0.34375 -0.3125q0.15625 -0.234375 0.203125 -0.578125q0.046875 -0.265625 0.046875 -0.90625q0 -1.046875 0.09375 -1.453125q0.09375 -0.40625 0.34375 -0.65625q0.265625 -0.25 0.75 -0.390625q0.328125 -0.109375 1.03125 -0.109375l0.28125 0l0 1.296875q-0.59375 0 -0.765625 0.0625q-0.171875 0.0625 -0.25 0.203125q-0.078125 0.140625 -0.078125 0.46875q0 0.34375 -0.046875 1.28125q-0.03125 0.53125 -0.15625 0.875q-0.109375 0.328125 -0.28125 0.546875q-0.171875 0.203125 -0.53125 0.4375q0.3125 0.171875 0.515625 0.421875q0.203125 0.25 0.3125 0.59375q0.109375 0.34375 0.140625 0.9375q0.03125 0.890625 0.03125 1.140625q0 0.359375 0.078125 0.484375q0.09375 0.140625 0.265625 0.21875q0.1875 0.078125 0.765625 0.078125l0 1.28125l-0.28125 0q-0.734375 0 -1.125 -0.125q-0.375 -0.109375 -0.640625 -0.375q-0.265625 -0.265625 -0.359375 -0.671875q-0.09375 -0.390625 -0.09375 -1.234375q0 -0.984375 -0.09375 -1.28125q-0.109375 -0.4375 -0.359375 -0.609375q-0.234375 -0.1875 -0.71875 -0.21875z" fill-rule="nonzero"></path><path fill="#000000" d="m620.7487 472.52347l-1.890625 -7.875l1.640625 0l1.1875 5.40625l1.4375 -5.40625l1.890625 0l1.375 5.5l1.203125 -5.5l1.609375 0l-1.921875 7.875l-1.6875 0l-1.5625 -5.890625l-1.5625 5.890625l-1.71875 0zm8.948486 0l0 -7.875l1.578125 0l0 3.09375l3.125 0l0 -3.09375l1.578125 0l0 7.875l-1.578125 0l0 -3.4375l-3.125 0l0 3.4375l-1.578125 0zm7.6799927 0l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm7.12323 0l0 -7.875l1.546875 0l3.234375 5.265625l0 -5.265625l1.46875 0l0 7.875l-1.59375 0l-3.171875 -5.140625l0 5.140625l-1.484375 0zm10.179993 2.3125l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm1.048645 -2.3125l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.5387573 2.3125q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0zm6.4624023 -4.515625l0 -1.28125q0.390625 -0.03125 0.578125 -0.125q0.203125 -0.09375 0.34375 -0.3125q0.15625 -0.234375 0.203125 -0.578125q0.046875 -0.265625 0.046875 -0.90625q0 -1.046875 0.09375 -1.453125q0.09375 -0.40625 0.34375 -0.65625q0.265625 -0.25 0.75 -0.390625q0.328125 -0.109375 1.03125 -0.109375l0.28125 0l0 1.296875q-0.59375 0 -0.765625 0.0625q-0.171875 0.0625 -0.25 0.203125q-0.078125 0.140625 -0.078125 0.46875q0 0.34375 -0.046875 1.28125q-0.03125 0.53125 -0.15625 0.875q-0.109375 0.328125 -0.28125 0.546875q-0.171875 0.203125 -0.53125 0.4375q0.3125 0.171875 0.515625 0.421875q0.203125 0.25 0.3125 0.59375q0.109375 0.34375 0.140625 0.9375q0.03125 0.890625 0.03125 1.140625q0 0.359375 0.078125 0.484375q0.09375 0.140625 0.265625 0.21875q0.1875 0.078125 0.765625 0.078125l0 1.28125l-0.28125 0q-0.734375 0 -1.125 -0.125q-0.375 -0.109375 -0.640625 -0.375q-0.265625 -0.265625 -0.359375 -0.671875q-0.09375 -0.390625 -0.09375 -1.234375q0 -0.984375 -0.09375 -1.28125q-0.109375 -0.4375 -0.359375 -0.609375q-0.234375 -0.1875 -0.71875 -0.21875z" fill-rule="nonzero"></path><path fill="#000000" d="m627.32623 485.52347l0 -6.546875l-2.34375 0l0 -1.328125l6.265625 0l0 1.328125l-2.34375 0l0 6.546875l-1.578125 0zm4.7436523 0l0 -7.875l1.578125 0l0 3.09375l3.125 0l0 -3.09375l1.578125 0l0 7.875l-1.578125 0l0 -3.4375l-3.125 0l0 3.4375l-1.578125 0zm7.6799927 0l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm7.12323 0l0 -7.875l1.546875 0l3.234375 5.265625l0 -5.265625l1.46875 0l0 7.875l-1.59375 0l-3.171875 -5.140625l0 5.140625l-1.484375 0zm10.179993 2.3125l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm1.048645 -2.3125l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.5387573 2.3125q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0zm6.4624023 -4.515625l0 -1.28125q0.390625 -0.03125 0.578125 -0.125q0.203125 -0.09375 0.34375 -0.3125q0.15625 -0.234375 0.203125 -0.578125q0.046875 -0.265625 0.046875 -0.90625q0 -1.046875 0.09375 -1.453125q0.09375 -0.40625 0.34375 -0.65625q0.265625 -0.25 0.75 -0.390625q0.328125 -0.109375 1.03125 -0.109375l0.28125 0l0 1.296875q-0.59375 0 -0.765625 0.0625q-0.171875 0.0625 -0.25 0.203125q-0.078125 0.140625 -0.078125 0.46875q0 0.34375 -0.046875 1.28125q-0.03125 0.53125 -0.15625 0.875q-0.109375 0.328125 -0.28125 0.546875q-0.171875 0.203125 -0.53125 0.4375q0.3125 0.171875 0.515625 0.421875q0.203125 0.25 0.3125 0.59375q0.109375 0.34375 0.140625 0.9375q0.03125 0.890625 0.03125 1.140625q0 0.359375 0.078125 0.484375q0.09375 0.140625 0.265625 0.21875q0.1875 0.078125 0.765625 0.078125l0 1.28125l-0.28125 0q-0.734375 0 -1.125 -0.125q-0.375 -0.109375 -0.640625 -0.375q-0.265625 -0.265625 -0.359375 -0.671875q-0.09375 -0.390625 -0.09375 -1.234375q0 -0.984375 -0.09375 -1.28125q-0.109375 -0.4375 -0.359375 -0.609375q-0.234375 -0.1875 -0.71875 -0.21875z" fill-rule="nonzero"></path><path fill="#000000" d="m636.5132 495.63284l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm2.6643677 2.890625l0 -7.875l1.578125 0l0 3.09375l3.125 0l0 -3.09375l1.578125 0l0 7.875l-1.578125 0l0 -3.4375l-3.125 0l0 3.4375l-1.578125 0zm7.6799316 0l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm12.154541 -2.890625l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm2.6799927 2.890625l0 -7.875l1.578125 0l0 3.5l3.21875 -3.5l2.140625 0l-2.96875 3.0625l3.125 4.8125l-2.0625 0l-2.15625 -3.703125l-1.296875 1.328125l0 2.375l-1.578125 0zm10.164307 2.3125l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm1.048645 -2.3125l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.5388184 2.3125q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0zm4.204834 -6.5l0 -1.515625l1.515625 0l0 1.515625l-1.515625 0zm0 2.671875l1.515625 0l0 1.078125q0 0.65625 -0.109375 1.03125q-0.109375 0.390625 -0.4375 0.6875q-0.3125 0.296875 -0.796875 0.46875l-0.296875 -0.609375q0.46875 -0.15625 0.65625 -0.421875q0.1875 -0.265625 0.203125 -0.71875l-0.734375 0l0 -1.515625z" fill-rule="nonzero"></path><path fill="#000000" d="m628.67 509.32034q-0.390625 0.015625 -0.59375 0.125q-0.203125 0.09375 -0.34375 0.328125q-0.140625 0.21875 -0.203125 0.5625q-0.046875 0.265625 -0.046875 0.890625q0 1.0468445 -0.09375 1.4530945q-0.09375 0.421875 -0.34375 0.671875q-0.25 0.25 -0.734375 0.390625q-0.328125 0.09375 -1.03125 0.09375l-0.296875 0l0 -1.28125q0.578125 0 0.75 -0.078125q0.1875 -0.078125 0.265625 -0.21875q0.09375 -0.125 0.09375 -0.45309448q0 -0.328125 0.046875 -1.25q0.015625 -0.5625 0.140625 -0.890625q0.125 -0.34375 0.3125 -0.578125q0.203125 -0.234375 0.5 -0.40625q-0.390625 -0.265625 -0.578125 -0.515625q-0.25 -0.34375 -0.34375 -0.890625q-0.0625 -0.375 -0.078125 -1.65625q-0.015625 -0.40625 -0.078125 -0.53125q-0.0625 -0.140625 -0.25 -0.203125q-0.171875 -0.078125 -0.78125 -0.078125l0 -1.296875l0.296875 0q0.71875 0 1.109375 0.125q0.390625 0.109375 0.640625 0.390625q0.265625 0.265625 0.359375 0.65625q0.09375 0.390625 0.09375 1.234375q0 0.984375 0.078125 1.296875q0.125 0.421875 0.359375 0.609375q0.25 0.1875 0.75 0.21875l0 1.28125z" fill-rule="nonzero"></path><path fill="#000000" d="m622.7487 522.3203q-0.390625 0.015625 -0.59375 0.125q-0.203125 0.09375 -0.34375 0.328125q-0.140625 0.21875 -0.203125 0.5625q-0.046875 0.265625 -0.046875 0.890625q0 1.046875 -0.09375 1.453125q-0.09375 0.421875 -0.34375 0.671875q-0.25 0.25 -0.734375 0.390625q-0.328125 0.09375 -1.03125 0.09375l-0.296875 0l0 -1.28125q0.578125 0 0.75 -0.078125q0.1875 -0.078125 0.265625 -0.21875q0.09375 -0.125 0.09375 -0.453125q0 -0.328125 0.046875 -1.25q0.015625 -0.5625 0.140625 -0.890625q0.125 -0.34375 0.3125 -0.578125q0.203125 -0.234375 0.5 -0.40625q-0.390625 -0.265625 -0.578125 -0.515625q-0.25 -0.34375 -0.34375 -0.890625q-0.0625 -0.375 -0.078125 -1.65625q-0.015625 -0.40625 -0.078125 -0.53125q-0.0625 -0.140625 -0.25 -0.203125q-0.171875 -0.078125 -0.78125 -0.078125l0 -1.296875l0.296875 0q0.71875 0 1.109375 0.125q0.390625 0.109375 0.640625 0.390625q0.265625 0.265625 0.359375 0.65625q0.09375 0.390625 0.09375 1.234375q0 0.984375 0.078125 1.296875q0.125 0.421875 0.359375 0.609375q0.25 0.1875 0.75 0.21875l0 1.28125z" fill-rule="nonzero"></path><path fill="#000000" d="m616.8274 535.3203q-0.390625 0.015625 -0.59375 0.125q-0.203125 0.09375 -0.34375 0.328125q-0.140625 0.21875 -0.203125 0.5625q-0.046875 0.265625 -0.046875 0.890625q0 1.046875 -0.09375 1.453125q-0.09375 0.421875 -0.34375 0.671875q-0.25 0.25 -0.734375 0.390625q-0.328125 0.09375 -1.03125 0.09375l-0.296875 0l0 -1.28125q0.578125 0 0.75 -0.078125q0.1875 -0.078125 0.265625 -0.21875q0.09375 -0.125 0.09375 -0.453125q0 -0.328125 0.046875 -1.25q0.015625 -0.5625 0.140625 -0.890625q0.125 -0.34375 0.3125 -0.578125q0.203125 -0.234375 0.5 -0.40625q-0.390625 -0.265625 -0.578125 -0.515625q-0.25 -0.34375 -0.34375 -0.890625q-0.0625 -0.375 -0.078125 -1.65625q-0.015625 -0.40625 -0.078125 -0.53125q-0.0625 -0.140625 -0.25 -0.203125q-0.171875 -0.078125 -0.78125 -0.078125l0 -1.296875l0.296875 0q0.71875 0 1.109375 0.125q0.390625 0.109375 0.640625 0.390625q0.265625 0.265625 0.359375 0.65625q0.09375 0.390625 0.09375 1.234375q0 0.984375 0.078125 1.296875q0.125 0.421875 0.359375 0.609375q0.25 0.1875 0.75 0.21875l0 1.28125z" fill-rule="nonzero"></path><path fill="#000000" d="m610.9061 548.3204q-0.390625 0.015625 -0.59375 0.125q-0.203125 0.09375 -0.34375 0.328125q-0.140625 0.21875 -0.203125 0.5625q-0.046875 0.265625 -0.046875 0.890625q0 1.046875 -0.09375 1.453125q-0.09375 0.421875 -0.34375 0.671875q-0.25 0.25 -0.734375 0.390625q-0.328125 0.09375 -1.03125 0.09375l-0.296875 0l0 -1.28125q0.578125 0 0.75 -0.078125q0.1875 -0.078125 0.265625 -0.21875q0.09375 -0.125 0.09375 -0.453125q0 -0.328125 0.046875 -1.25q0.015625 -0.5625 0.140625 -0.890625q0.125 -0.34375 0.3125 -0.578125q0.203125 -0.234375 0.5 -0.40625q-0.390625 -0.265625 -0.578125 -0.515625q-0.25 -0.34375 -0.34375 -0.890625q-0.0625 -0.375 -0.078125 -1.65625q-0.015625 -0.40625 -0.078125 -0.53125q-0.0625 -0.140625 -0.25 -0.203125q-0.171875 -0.078125 -0.78125 -0.078125l0 -1.296875l0.296875 0q0.71875 0 1.109375 0.125q0.390625 0.109375 0.640625 0.390625q0.265625 0.265625 0.359375 0.65625q0.09375 0.390625 0.09375 1.234375q0 0.984375 0.078125 1.296875q0.125 0.421875 0.359375 0.609375q0.25 0.1875 0.75 0.21875l0 1.28125z" fill-rule="nonzero"></path><path fill="#d5a6bd" d="m310.88715 609.273l0 0c0 -1.8467407 1.4971008 -3.343872 3.3439026 -3.343872l76.20984 0c0.8868408 0 1.7373962 0.35229492 2.3644714 0.97943115c0.6271057 0.6270752 0.97943115 1.4776001 0.97943115 2.364441l0 13.375244c0 1.8467407 -1.4971313 3.343872 -3.3439026 3.343872l-76.20984 0c-1.8468018 0 -3.3439026 -1.4971313 -3.3439026 -3.343872z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m310.88715 609.273l0 0c0 -1.8467407 1.4971008 -3.343872 3.3439026 -3.343872l76.20984 0c0.8868408 0 1.7373962 0.35229492 2.3644714 0.97943115c0.6271057 0.6270752 0.97943115 1.4776001 0.97943115 2.364441l0 13.375244c0 1.8467407 -1.4971313 3.343872 -3.3439026 3.343872l-76.20984 0c-1.8468018 0 -3.3439026 -1.4971313 -3.3439026 -3.343872z" fill-rule="nonzero"></path><path fill="#000000" d="m328.51956 620.48065l0 -8.59375l1.140625 0l0 7.578125l4.234375 0l0 1.015625l-5.375 0zm6.205078 -3.109375q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.783203 3.625l1.03125 0.15625q0.0625 0.46875 0.359375 0.6875q0.390625 0.296875 1.0625 0.296875q0.734375 0 1.125 -0.296875q0.40625 -0.296875 0.546875 -0.8125q0.09375 -0.328125 0.078125 -1.359375q-0.6875 0.8125 -1.71875 0.8125q-1.28125 0 -1.984375 -0.921875q-0.703125 -0.9375 -0.703125 -2.21875q0 -0.890625 0.3125 -1.640625q0.328125 -0.765625 0.9375 -1.171875q0.609375 -0.40625 1.4375 -0.40625q1.109375 0 1.828125 0.890625l0 -0.75l0.96875 0l0 5.375q0 1.453125 -0.296875 2.0625q-0.296875 0.609375 -0.9375 0.953125q-0.640625 0.359375 -1.578125 0.359375q-1.109375 0 -1.796875 -0.5q-0.6875 -0.5 -0.671875 -1.515625zm0.875 -3.734375q0 1.21875 0.484375 1.78125q0.484375 0.5625 1.21875 0.5625q0.734375 0 1.21875 -0.5625q0.5 -0.5625 0.5 -1.75q0 -1.140625 -0.515625 -1.71875q-0.5 -0.578125 -1.21875 -0.578125q-0.703125 0 -1.203125 0.578125q-0.484375 0.5625 -0.484375 1.6875zm5.798828 3.734375l1.03125 0.15625q0.0625 0.46875 0.359375 0.6875q0.390625 0.296875 1.0625 0.296875q0.734375 0 1.125 -0.296875q0.40625 -0.296875 0.546875 -0.8125q0.09375 -0.328125 0.078125 -1.359375q-0.6875 0.8125 -1.71875 0.8125q-1.28125 0 -1.984375 -0.921875q-0.703125 -0.9375 -0.703125 -2.21875q0 -0.890625 0.3125 -1.640625q0.328125 -0.765625 0.9375 -1.171875q0.609375 -0.40625 1.4375 -0.40625q1.109375 0 1.828125 0.890625l0 -0.75l0.96875 0l0 5.375q0 1.453125 -0.296875 2.0625q-0.296875 0.609375 -0.9375 0.953125q-0.640625 0.359375 -1.578125 0.359375q-1.109375 0 -1.796875 -0.5q-0.6875 -0.5 -0.671875 -1.515625zm0.875 -3.734375q0 1.21875 0.484375 1.78125q0.484375 0.5625 1.21875 0.5625q0.734375 0 1.21875 -0.5625q0.5 -0.5625 0.5 -1.75q0 -1.140625 -0.515625 -1.71875q-0.5 -0.578125 -1.21875 -0.578125q-0.703125 0 -1.203125 0.578125q-0.484375 0.5625 -0.484375 1.6875zm10.064453 2.453125q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm3.6738281 3.125l-0.984375 0l0 -8.59375l1.0625 0l0 3.0625q0.671875 -0.828125 1.703125 -0.828125q0.578125 0 1.078125 0.234375q0.515625 0.21875 0.84375 0.640625q0.34375 0.421875 0.53125 1.015625q0.1875 0.59375 0.1875 1.265625q0 1.59375 -0.796875 2.46875q-0.796875 0.875 -1.890625 0.875q-1.109375 0 -1.734375 -0.921875l0 0.78125zm-0.015625 -3.15625q0 1.109375 0.3125 1.609375q0.5 0.8125 1.34375 0.8125q0.6875 0 1.1875 -0.59375q0.515625 -0.59375 0.515625 -1.796875q0 -1.21875 -0.484375 -1.796875q-0.484375 -0.578125 -1.171875 -0.578125q-0.6875 0 -1.203125 0.609375q-0.5 0.59375 -0.5 1.734375zm5.689453 3.15625l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm6.9472656 -2.0l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125z" fill-rule="nonzero"></path><path fill="#b4a7d6" d="m485.05518 517.9685l63.138977 0l3.343933 3.343872l0 16.719116l-69.82681 0l0 -16.719116l0 0c0 -1.8467407 1.4971313 -3.343872 3.3439026 -3.343872z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m485.05518 517.9685l63.138977 0l3.343933 3.343872l0 16.719116l-69.82681 0l0 -16.719116l0 0c0 -1.8467407 1.4971313 -3.343872 3.3439026 -3.343872z" fill-rule="nonzero"></path><path fill="#000000" d="m501.64752 531.1503l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5zm10.671875 -0.140625l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm8.173828 2.765625l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm3.3339844 0l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm5.2871094 -1.0625l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.861328 3.703125l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#b4a7d6" d="m397.12866 517.9685l63.138977 0l3.3439026 3.343872l0 16.719116l-69.82675 0l0 -16.719116l0 0c0 -1.8467407 1.4971008 -3.343872 3.343872 -3.343872z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m397.12866 517.9685l63.138977 0l3.3439026 3.343872l0 16.719116l-69.82675 0l0 -16.719116l0 0c0 -1.8467407 1.4971008 -3.343872 3.343872 -3.343872z" fill-rule="nonzero"></path><path fill="#000000" d="m413.60287 533.5253l1.03125 0.15625q0.0625 0.46875 0.359375 0.6875q0.390625 0.296875 1.0625 0.296875q0.734375 0 1.125 -0.296875q0.40625 -0.296875 0.546875 -0.8125q0.09375 -0.328125 0.078125 -1.359375q-0.6875 0.8125 -1.71875 0.8125q-1.28125 0 -1.984375 -0.921875q-0.703125 -0.9375 -0.703125 -2.21875q0 -0.890625 0.3125 -1.640625q0.328125 -0.765625 0.9375 -1.171875q0.609375 -0.40625 1.4375 -0.40625q1.109375 0 1.828125 0.890625l0 -0.75l0.96875 0l0 5.375q0 1.453125 -0.296875 2.0625q-0.296875 0.609375 -0.9375 0.953125q-0.640625 0.359375 -1.578125 0.359375q-1.109375 0 -1.796875 -0.5q-0.6875 -0.5 -0.671875 -1.515625zm0.875 -3.734375q0 1.21875 0.484375 1.78125q0.484375 0.5625 1.21875 0.5625q0.734375 0 1.21875 -0.5625q0.5 -0.5625 0.5 -1.75q0 -1.140625 -0.515625 -1.71875q-0.5 -0.578125 -1.21875 -0.578125q-0.703125 0 -1.203125 0.578125q-0.484375 0.5625 -0.484375 1.6875zm10.251953 1.21875l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm8.173828 2.765625l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm3.3339844 0l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm5.2871094 -1.0625l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.861328 3.703125l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#b4a7d6" d="m251.6851 424.5433l63.13896 0l3.3439026 3.3439026l0 16.719086l-69.82677 0l0 -16.719086l0 0c0 -1.8467712 1.4971161 -3.3439026 3.3439026 -3.3439026z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m251.6851 424.5433l63.13896 0l3.3439026 3.3439026l0 16.719086l-69.82677 0l0 -16.719086l0 0c0 -1.8467712 1.4971161 -3.3439026 3.3439026 -3.3439026z" fill-rule="nonzero"></path><path fill="#000000" d="m274.76575 437.30325l1.03125 0.140625q-0.171875 1.0625 -0.875 1.671875q-0.703125 0.609375 -1.71875 0.609375q-1.28125 0 -2.0625 -0.828125q-0.765625 -0.84375 -0.765625 -2.40625q0 -1.0 0.328125 -1.75q0.34375 -0.765625 1.015625 -1.140625q0.6875 -0.375 1.5 -0.375q1.0 0 1.640625 0.515625q0.65625 0.5 0.84375 1.453125l-1.03125 0.15625q-0.140625 -0.625 -0.515625 -0.9375q-0.375 -0.328125 -0.90625 -0.328125q-0.796875 0 -1.296875 0.578125q-0.5 0.5625 -0.5 1.796875q0 1.265625 0.484375 1.828125q0.484375 0.5625 1.25 0.5625q0.625 0 1.03125 -0.375q0.421875 -0.375 0.546875 -1.171875zm4.234375 1.34375l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm0.6464844 -2.171875q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.970703 3.109375l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0zm3.5898438 -1.859375l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5z" fill-rule="nonzero"></path><path fill="#b4a7d6" d="m251.6851 456.56693l63.13896 0l3.3439026 3.3439026l0 16.719086l-69.82677 0l0 -16.719086l0 0c0 -1.8467712 1.4971161 -3.3439026 3.3439026 -3.3439026z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m251.6851 456.56693l63.13896 0l3.3439026 3.3439026l0 16.719086l-69.82677 0l0 -16.719086l0 0c0 -1.8467712 1.4971161 -3.3439026 3.3439026 -3.3439026z" fill-rule="nonzero"></path><path fill="#000000" d="m277.39758 471.60812l0 -0.78125q-0.59375 0.921875 -1.734375 0.921875q-0.75 0 -1.375 -0.40625q-0.625 -0.421875 -0.96875 -1.15625q-0.34375 -0.734375 -0.34375 -1.6875q0 -0.921875 0.3125 -1.6875q0.3125 -0.765625 0.9375 -1.15625q0.625 -0.40625 1.390625 -0.40625q0.5625 0 1.0 0.234375q0.4375 0.234375 0.71875 0.609375l0 -3.078125l1.046875 0l0 8.59375l-0.984375 0zm-3.328125 -3.109375q0 1.203125 0.5 1.796875q0.5 0.578125 1.1875 0.578125q0.6875 0 1.171875 -0.5625q0.484375 -0.5625 0.484375 -1.71875q0 -1.28125 -0.5 -1.875q-0.484375 -0.59375 -1.203125 -0.59375q-0.703125 0 -1.171875 0.578125q-0.46875 0.5625 -0.46875 1.796875zm8.267578 2.171875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm0.6464844 -2.171875q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.970703 3.109375l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#b4a7d6" d="m251.6851 488.56693l63.13896 0l3.3439026 3.3439026l0 16.719086l-69.82677 0l0 -16.719086l0 0c0 -1.8467712 1.4971161 -3.3439026 3.3439026 -3.3439026z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m251.6851 488.56693l63.13896 0l3.3439026 3.3439026l0 16.719086l-69.82677 0l0 -16.719086l0 0c0 -1.8467712 1.4971161 -3.3439026 3.3439026 -3.3439026z" fill-rule="nonzero"></path><path fill="#000000" d="m269.63293 500.49875q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.986328 5.5l0 -8.609375l0.953125 0l0 0.796875q0.34375 -0.46875 0.765625 -0.703125q0.4375 -0.234375 1.046875 -0.234375q0.796875 0 1.40625 0.40625q0.609375 0.40625 0.90625 1.15625q0.3125 0.75 0.3125 1.640625q0 0.953125 -0.34375 1.71875q-0.328125 0.765625 -0.984375 1.171875q-0.65625 0.40625 -1.375 0.40625q-0.53125 0 -0.953125 -0.21875q-0.421875 -0.234375 -0.6875 -0.5625l0 3.03125l-1.046875 0zm0.953125 -5.46875q0 1.203125 0.484375 1.78125q0.484375 0.5625 1.171875 0.5625q0.703125 0 1.203125 -0.59375q0.5 -0.59375 0.5 -1.84375q0 -1.1875 -0.484375 -1.765625q-0.484375 -0.59375 -1.171875 -0.59375q-0.671875 0 -1.1875 0.625q-0.515625 0.625 -0.515625 1.828125zm5.580078 -0.671875l0 -0.984375l5.6875 -2.40625l0 1.046875l-4.515625 1.84375l4.515625 1.875l0 1.046875l-5.6875 -2.421875zm7.0078125 0l0 -0.984375l5.6875 -2.40625l0 1.046875l-4.515625 1.84375l4.515625 1.875l0 1.046875l-5.6875 -2.421875z" fill-rule="nonzero"></path><path fill="#b4a7d6" d="m251.6851 520.56696l63.13896 0l3.3439026 3.343872l0 16.719116l-69.82677 0l0 -16.719116l0 0c0 -1.8468018 1.4971161 -3.343872 3.3439026 -3.343872z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m251.6851 520.56696l63.13896 0l3.3439026 3.343872l0 16.719116l-69.82677 0l0 -16.719116l0 0c0 -1.8468018 1.4971161 -3.343872 3.3439026 -3.343872z" fill-rule="nonzero"></path><path fill="#000000" d="m269.63293 532.4987q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.986328 5.5l0 -8.609375l0.953125 0l0 0.796875q0.34375 -0.46875 0.765625 -0.703125q0.4375 -0.234375 1.046875 -0.234375q0.796875 0 1.40625 0.40625q0.609375 0.40625 0.90625 1.15625q0.3125 0.75 0.3125 1.640625q0 0.953125 -0.34375 1.71875q-0.328125 0.765625 -0.984375 1.171875q-0.65625 0.40625 -1.375 0.40625q-0.53125 0 -0.953125 -0.21875q-0.421875 -0.234375 -0.6875 -0.5625l0 3.03125l-1.046875 0zm0.953125 -5.46875q0 1.203125 0.484375 1.78125q0.484375 0.5625 1.171875 0.5625q0.703125 0 1.203125 -0.59375q0.5 -0.59375 0.5 -1.84375q0 -1.1875 -0.484375 -1.765625q-0.484375 -0.59375 -1.171875 -0.59375q-0.671875 0 -1.1875 0.625q-0.515625 0.625 -0.515625 1.828125zm11.267578 -1.96875l-5.671875 0l0 -0.984375l5.671875 0l0 0.984375zm0 2.609375l-5.671875 0l0 -0.984375l5.671875 0l0 0.984375zm7.0078125 -2.609375l-5.671875 0l0 -0.984375l5.671875 0l0 0.984375zm0 2.609375l-5.671875 0l0 -0.984375l5.671875 0l0 0.984375z" fill-rule="nonzero"></path><path fill="#b4a7d6" d="m251.6851 552.56696l63.13896 0l3.3439026 3.343872l0 16.719116l-69.82677 0l0 -16.719116l0 0c0 -1.8468018 1.4971161 -3.343872 3.3439026 -3.343872z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m251.6851 552.56696l63.13896 0l3.3439026 3.343872l0 16.719116l-69.82677 0l0 -16.719116l0 0c0 -1.8468018 1.4971161 -3.343872 3.3439026 -3.343872z" fill-rule="nonzero"></path><path fill="#000000" d="m273.13684 564.4987q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.986328 5.5l0 -8.609375l0.953125 0l0 0.796875q0.34375 -0.46875 0.765625 -0.703125q0.4375 -0.234375 1.046875 -0.234375q0.796875 0 1.40625 0.40625q0.609375 0.40625 0.90625 1.15625q0.3125 0.75 0.3125 1.640625q0 0.953125 -0.34375 1.71875q-0.328125 0.765625 -0.984375 1.171875q-0.65625 0.40625 -1.375 0.40625q-0.53125 0 -0.953125 -0.21875q-0.421875 -0.234375 -0.6875 -0.5625l0 3.03125l-1.046875 0zm0.953125 -5.46875q0 1.203125 0.484375 1.78125q0.484375 0.5625 1.171875 0.5625q0.703125 0 1.203125 -0.59375q0.5 -0.59375 0.5 -1.84375q0 -1.1875 -0.484375 -1.765625q-0.484375 -0.59375 -1.171875 -0.59375q-0.671875 0 -1.1875 0.625q-0.515625 0.625 -0.515625 1.828125zm5.580078 -0.671875l0 -0.984375l5.6875 -2.40625l0 1.046875l-4.515625 1.84375l4.515625 1.875l0 1.046875l-5.6875 -2.421875z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m428.69815 538.0315l0 33.94867l-76.346436 0l0 33.95685" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m428.69815 538.0315l0 33.94867l-76.346436 0l0 33.95685" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m516.6247 538.0315l0 33.94867l-164.28348 0l0 33.95685" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m516.6247 538.0315l0 33.94867l-164.28348 0l0 33.95685" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m318.16797 562.59845l34.17325 0l0 43.338562" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m318.16797 562.59845l34.17325 0l0 43.338562" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m318.16797 530.59845l34.17325 0l0 75.33856" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m318.16797 530.59845l34.17325 0l0 75.33856" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m318.16797 498.59842l34.17325 0l0 107.33859" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m318.16797 498.59842l34.17325 0l0 107.33859" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m318.16797 466.59842l34.17325 0l0 139.3386" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m318.16797 466.59842l34.17325 0l0 139.3386" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m318.16797 434.5748l34.17325 0l0 171.37009" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m318.16797 434.5748l34.17325 0l0 171.37009" fill-rule="nonzero"></path><path fill="#8e7cc3" d="m180.27559 369.5958l93.84245 0l4.0787964 4.078827l0 20.393616l-97.92125 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m180.27559 369.5958l93.84245 0l4.0787964 4.078827l0 20.393616l-97.92125 0z" fill-rule="nonzero"></path><path fill="#000000" d="m196.73898 387.37173l0 -8.59375l1.140625 0l0 7.578125l4.234375 0l0 1.015625l-5.375 0zm6.205078 -3.109375q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.783203 3.625l1.03125 0.15625q0.0625 0.46875 0.359375 0.6875q0.390625 0.296875 1.0625 0.296875q0.734375 0 1.125 -0.296875q0.40625 -0.296875 0.546875 -0.8125q0.09375 -0.328125 0.078125 -1.359375q-0.6875 0.8125 -1.71875 0.8125q-1.28125 0 -1.984375 -0.921875q-0.703125 -0.9375 -0.703125 -2.21875q0 -0.890625 0.3125 -1.640625q0.328125 -0.765625 0.9375 -1.171875q0.609375 -0.40625 1.4375 -0.40625q1.109375 0 1.828125 0.890625l0 -0.75l0.96875 0l0 5.375q0 1.453125 -0.296875 2.0625q-0.296875 0.609375 -0.9375 0.953125q-0.640625 0.359375 -1.578125 0.359375q-1.109375 0 -1.796875 -0.5q-0.6875 -0.5 -0.671875 -1.515625zm0.875 -3.734375q0 1.21875 0.484375 1.78125q0.484375 0.5625 1.21875 0.5625q0.734375 0 1.21875 -0.5625q0.5 -0.5625 0.5 -1.75q0 -1.140625 -0.515625 -1.71875q-0.5 -0.578125 -1.21875 -0.578125q-0.703125 0 -1.203125 0.578125q-0.484375 0.5625 -0.484375 1.6875zm6.126953 3.21875l0 -8.59375l3.25 0q0.84375 0 1.296875 0.078125q0.640625 0.109375 1.0625 0.40625q0.4375 0.296875 0.6875 0.828125q0.265625 0.53125 0.265625 1.171875q0 1.09375 -0.703125 1.859375q-0.6875 0.75 -2.515625 0.75l-2.203125 0l0 3.5l-1.140625 0zm1.140625 -4.5l2.21875 0q1.109375 0 1.5625 -0.40625q0.46875 -0.421875 0.46875 -1.171875q0 -0.53125 -0.28125 -0.90625q-0.265625 -0.390625 -0.703125 -0.515625q-0.296875 -0.078125 -1.0625 -0.078125l-2.203125 0l0 3.078125zm6.7226562 4.5l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0zm3.6210938 -3.109375q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm7.705078 3.109375l-2.359375 -6.21875l1.109375 0l1.34375 3.71875q0.203125 0.609375 0.390625 1.25q0.140625 -0.484375 0.390625 -1.171875l1.390625 -3.796875l1.078125 0l-2.359375 6.21875l-0.984375 0zm4.28125 -7.375l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm6.6972656 0l0 -0.78125q-0.59375 0.921875 -1.734375 0.921875q-0.75 0 -1.375 -0.40625q-0.625 -0.421875 -0.96875 -1.15625q-0.34375 -0.734375 -0.34375 -1.6875q0 -0.921875 0.3125 -1.6875q0.3125 -0.765625 0.9375 -1.15625q0.625 -0.40625 1.390625 -0.40625q0.5625 0 1.0 0.234375q0.4375 0.234375 0.71875 0.609375l0 -3.078125l1.046875 0l0 8.59375l-0.984375 0zm-3.328125 -3.109375q0 1.203125 0.5 1.796875q0.5 0.578125 1.1875 0.578125q0.6875 0 1.171875 -0.5625q0.484375 -0.5625 0.484375 -1.71875q0 -1.28125 -0.5 -1.875q-0.484375 -0.59375 -1.203125 -0.59375q-0.703125 0 -1.171875 0.578125q-0.46875 0.5625 -0.46875 1.796875zm10.220703 1.109375l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.861328 3.703125l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m260.59317 288.11548l-31.370071 0l0 81.480316" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m260.59317 288.11548l-31.370071 0l0 81.480316" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m229.23622 394.06824l0 221.88977l81.6378 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m229.23622 394.06824l0 221.88977l81.6378 0" fill-rule="nonzero"></path><path fill="#6aa84f" d="m431.6509 293.5223l93.84244 0l4.0788574 4.078827l0 20.393616l-97.921295 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m431.6509 293.5223l93.84244 0l4.0788574 4.078827l0 20.393616l-97.921295 0z" fill-rule="nonzero"></path><path fill="#000000" d="m448.6905 311.29825l0 -7.578125l-2.828125 0l0 -1.015625l6.8125 0l0 1.015625l-2.84375 0l0 7.578125l-1.140625 0zm9.267578 -2.0l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.455078 1.84375l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5zm8.71875 0.921875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm1.1621094 0.9375l0 -8.59375l3.25 0q0.84375 0 1.296875 0.078125q0.640625 0.109375 1.0625 0.40625q0.4375 0.296875 0.6875 0.828125q0.265625 0.53125 0.265625 1.171875q0 1.09375 -0.703125 1.859375q-0.6875 0.75 -2.515625 0.75l-2.203125 0l0 3.5l-1.140625 0zm1.140625 -4.5l2.21875 0q1.109375 0 1.5625 -0.40625q0.46875 -0.421875 0.46875 -1.171875q0 -0.53125 -0.28125 -0.90625q-0.265625 -0.390625 -0.703125 -0.515625q-0.296875 -0.078125 -1.0625 -0.078125l-2.203125 0l0 3.078125zm6.7226562 4.5l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0zm3.6210938 -3.109375q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm7.705078 3.109375l-2.359375 -6.21875l1.109375 0l1.34375 3.71875q0.203125 0.609375 0.390625 1.25q0.140625 -0.484375 0.390625 -1.171875l1.390625 -3.796875l1.078125 0l-2.359375 6.21875l-0.984375 0zm4.28125 -7.375l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm6.6972656 0l0 -0.78125q-0.59375 0.921875 -1.734375 0.921875q-0.75 0 -1.375 -0.40625q-0.625 -0.421875 -0.96875 -1.15625q-0.34375 -0.734375 -0.34375 -1.6875q0 -0.921875 0.3125 -1.6875q0.3125 -0.765625 0.9375 -1.15625q0.625 -0.40625 1.390625 -0.40625q0.5625 0 1.0 0.234375q0.4375 0.234375 0.71875 0.609375l0 -3.078125l1.046875 0l0 8.59375l-0.984375 0zm-3.328125 -3.109375q0 1.203125 0.5 1.796875q0.5 0.578125 1.1875 0.578125q0.6875 0 1.171875 -0.5625q0.484375 -0.5625 0.484375 -1.71875q0 -1.28125 -0.5 -1.875q-0.484375 -0.59375 -1.203125 -0.59375q-0.703125 0 -1.171875 0.578125q-0.46875 0.5625 -0.46875 1.796875zm10.220703 1.109375l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.861328 3.703125l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m686.72705 352.28085l0 31.001038l-14.897644 0l0 31.01471" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m686.72705 352.28085l0 31.001038l-14.897644 0l0 31.01471" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m349.47507 288.11548l41.088257 0l0 17.637817l41.08496 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m349.47507 288.11548l41.088257 0l0 17.637817l41.08496 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m529.5722 305.75854l157.16534 0l0 26.456665" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m529.5722 305.75854l157.16534 0l0 26.456665" fill-rule="nonzero"></path><path fill="#e69138" d="m278.19684 213.1916l91.695435 0l6.22583 6.2258453l0 31.128479l-97.921265 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m278.19684 213.1916l91.695435 0l6.22583 6.2258453l0 31.128479l-97.921265 0z" fill-rule="nonzero"></path><path fill="#000000" d="m298.04144 230.94522l0 -8.59375l1.125 0l0 8.59375l-1.125 0zm3.0058594 0l0 -6.21875l0.9375 0l0 0.875q0.6875 -1.015625 1.984375 -1.015625q0.5625 0 1.03125 0.203125q0.484375 0.203125 0.71875 0.53125q0.234375 0.328125 0.328125 0.765625q0.046875 0.296875 0.046875 1.03125l0 3.828125l-1.046875 0l0 -3.78125q0 -0.65625 -0.125 -0.96875q-0.125 -0.3125 -0.4375 -0.5q-0.3125 -0.203125 -0.734375 -0.203125q-0.671875 0 -1.171875 0.4375q-0.484375 0.421875 -0.484375 1.609375l0 3.40625l-1.046875 0zm6.251953 -1.859375l1.03125 -0.15625q0.09375 0.625 0.484375 0.953125q0.40625 0.328125 1.140625 0.328125q0.71875 0 1.0625 -0.28125q0.359375 -0.296875 0.359375 -0.703125q0 -0.359375 -0.3125 -0.5625q-0.21875 -0.140625 -1.078125 -0.359375q-1.15625 -0.296875 -1.609375 -0.5q-0.4375 -0.21875 -0.671875 -0.59375q-0.234375 -0.375 -0.234375 -0.84375q0 -0.40625 0.1875 -0.765625q0.1875 -0.359375 0.515625 -0.59375q0.25 -0.171875 0.671875 -0.296875q0.421875 -0.125 0.921875 -0.125q0.71875 0 1.265625 0.21875q0.5625 0.203125 0.828125 0.5625q0.265625 0.359375 0.359375 0.953125l-1.03125 0.140625q-0.0625 -0.46875 -0.40625 -0.734375q-0.328125 -0.28125 -0.953125 -0.28125q-0.71875 0 -1.03125 0.25q-0.3125 0.234375 -0.3125 0.5625q0 0.203125 0.125 0.359375q0.140625 0.171875 0.40625 0.28125q0.15625 0.0625 0.9375 0.265625q1.125 0.3125 1.5625 0.5q0.4375 0.1875 0.6875 0.546875q0.25 0.359375 0.25 0.90625q0 0.53125 -0.3125 1.0q-0.296875 0.453125 -0.875 0.71875q-0.578125 0.25 -1.3125 0.25q-1.21875 0 -1.859375 -0.5q-0.625 -0.515625 -0.796875 -1.5zm8.71875 0.921875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm5.0996094 0.171875q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm2.6738281 3.125l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm2.6660156 0l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm6.7597656 -0.765625q-0.59375 0.5 -1.140625 0.703125q-0.53125 0.203125 -1.15625 0.203125q-1.03125 0 -1.578125 -0.5q-0.546875 -0.5 -0.546875 -1.28125q0 -0.453125 0.203125 -0.828125q0.203125 -0.390625 0.546875 -0.609375q0.34375 -0.234375 0.765625 -0.34375q0.296875 -0.09375 0.9375 -0.171875q1.265625 -0.140625 1.875 -0.359375q0 -0.21875 0 -0.265625q0 -0.65625 -0.296875 -0.921875q-0.40625 -0.34375 -1.203125 -0.34375q-0.734375 0 -1.09375 0.265625q-0.359375 0.25 -0.53125 0.90625l-1.03125 -0.140625q0.140625 -0.65625 0.46875 -1.0625q0.328125 -0.40625 0.9375 -0.625q0.609375 -0.21875 1.40625 -0.21875q0.796875 0 1.296875 0.1875q0.5 0.1875 0.734375 0.46875q0.234375 0.28125 0.328125 0.71875q0.046875 0.265625 0.046875 0.96875l0 1.40625q0 1.46875 0.0625 1.859375q0.078125 0.390625 0.28125 0.75l-1.109375 0q-0.15625 -0.328125 -0.203125 -0.765625zm-0.09375 -2.359375q-0.578125 0.234375 -1.71875 0.40625q-0.65625 0.09375 -0.921875 0.21875q-0.265625 0.109375 -0.421875 0.328125q-0.140625 0.21875 -0.140625 0.5q0 0.421875 0.3125 0.703125q0.328125 0.28125 0.9375 0.28125q0.609375 0 1.078125 -0.265625q0.484375 -0.265625 0.703125 -0.734375q0.171875 -0.359375 0.171875 -1.046875l0 -0.390625zm5.001953 2.1875l0.15625 0.921875q-0.453125 0.09375 -0.796875 0.09375q-0.578125 0 -0.890625 -0.171875q-0.3125 -0.1875 -0.453125 -0.484375q-0.125 -0.296875 -0.125 -1.25l0 -3.578125l-0.765625 0l0 -0.8125l0.765625 0l0 -1.546875l1.046875 -0.625l0 2.171875l1.0625 0l0 0.8125l-1.0625 0l0 3.640625q0 0.453125 0.046875 0.578125q0.0625 0.125 0.1875 0.203125q0.125 0.078125 0.359375 0.078125q0.1875 0 0.46875 -0.03125zm1.0371094 -6.4375l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm2.2753906 -3.109375q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.986328 3.109375l0 -6.21875l0.9375 0l0 0.875q0.6875 -1.015625 1.984375 -1.015625q0.5625 0 1.03125 0.203125q0.484375 0.203125 0.71875 0.53125q0.234375 0.328125 0.328125 0.765625q0.046875 0.296875 0.046875 1.03125l0 3.828125l-1.046875 0l0 -3.78125q0 -0.65625 -0.125 -0.96875q-0.125 -0.3125 -0.4375 -0.5q-0.3125 -0.203125 -0.734375 -0.203125q-0.671875 0 -1.171875 0.4375q-0.484375 0.421875 -0.484375 1.609375l0 3.40625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#000000" d="m307.79926 244.94524l0 -8.59375l3.234375 0q0.984375 0 1.578125 0.265625q0.59375 0.265625 0.921875 0.8125q0.34375 0.53125 0.34375 1.125q0 0.546875 -0.296875 1.03125q-0.296875 0.484375 -0.90625 0.78125q0.78125 0.234375 1.1875 0.796875q0.421875 0.546875 0.421875 1.296875q0 0.59375 -0.25 1.125q-0.25 0.515625 -0.625 0.796875q-0.375 0.28125 -0.953125 0.421875q-0.5625 0.140625 -1.375 0.140625l-3.28125 0zm1.140625 -4.984375l1.859375 0q0.75 0 1.078125 -0.09375q0.4375 -0.125 0.65625 -0.421875q0.21875 -0.3125 0.21875 -0.765625q0 -0.421875 -0.203125 -0.75q-0.203125 -0.328125 -0.59375 -0.4375q-0.375 -0.125 -1.296875 -0.125l-1.71875 0l0 2.59375zm0 3.96875l2.140625 0q0.546875 0 0.765625 -0.046875q0.40625 -0.0625 0.65625 -0.21875q0.265625 -0.171875 0.4375 -0.484375q0.171875 -0.3125 0.171875 -0.71875q0 -0.484375 -0.25 -0.84375q-0.234375 -0.359375 -0.671875 -0.5q-0.4375 -0.140625 -1.265625 -0.140625l-1.984375 0l0 2.953125zm10.863281 1.015625l0 -0.921875q-0.734375 1.0625 -1.984375 1.0625q-0.546875 0 -1.03125 -0.203125q-0.46875 -0.21875 -0.703125 -0.53125q-0.234375 -0.328125 -0.328125 -0.796875q-0.0625 -0.296875 -0.0625 -0.984375l0 -3.84375l1.0625 0l0 3.453125q0 0.8125 0.0625 1.109375q0.09375 0.40625 0.40625 0.65625q0.328125 0.234375 0.8125 0.234375q0.46875 0 0.875 -0.234375q0.421875 -0.25 0.59375 -0.671875q0.1875 -0.421875 0.1875 -1.21875l0 -3.328125l1.046875 0l0 6.21875l-0.9375 0zm2.5957031 -7.375l0 -1.21875l1.0625 0l0 1.21875l-1.0625 0zm0 7.375l0 -6.21875l1.0625 0l0 6.21875l-1.0625 0zm2.6347656 0l0 -8.59375l1.0625 0l0 8.59375l-1.0625 0zm6.7285156 0l0 -0.78125q-0.59375 0.921875 -1.734375 0.921875q-0.75 0 -1.375 -0.40625q-0.625 -0.421875 -0.96875 -1.15625q-0.34375 -0.734375 -0.34375 -1.6875q0 -0.921875 0.3125 -1.6875q0.3125 -0.765625 0.9375 -1.15625q0.625 -0.40625 1.390625 -0.40625q0.5625 0 1.0 0.234375q0.4375 0.234375 0.71875 0.609375l0 -3.078125l1.046875 0l0 8.59375l-0.984375 0zm-3.328125 -3.109375q0 1.203125 0.5 1.796875q0.5 0.578125 1.1875 0.578125q0.6875 0 1.171875 -0.5625q0.484375 -0.5625 0.484375 -1.71875q0 -1.28125 -0.5 -1.875q-0.484375 -0.59375 -1.203125 -0.59375q-0.703125 0 -1.171875 0.578125q-0.46875 0.5625 -0.46875 1.796875zm10.220703 1.109375l1.09375 0.125q-0.25 0.953125 -0.953125 1.484375q-0.703125 0.53125 -1.78125 0.53125q-1.359375 0 -2.171875 -0.84375q-0.796875 -0.84375 -0.796875 -2.359375q0 -1.5625 0.8125 -2.421875q0.8125 -0.875 2.09375 -0.875q1.25 0 2.03125 0.84375q0.796875 0.84375 0.796875 2.390625q0 0.09375 0 0.28125l-4.640625 0q0.0625 1.03125 0.578125 1.578125q0.515625 0.53125 1.296875 0.53125q0.578125 0 0.984375 -0.296875q0.421875 -0.3125 0.65625 -0.96875zm-3.453125 -1.703125l3.46875 0q-0.0625 -0.796875 -0.390625 -1.1875q-0.515625 -0.609375 -1.3125 -0.609375q-0.734375 0 -1.234375 0.484375q-0.484375 0.484375 -0.53125 1.3125zm5.861328 3.703125l0 -6.21875l0.953125 0l0 0.9375q0.359375 -0.65625 0.65625 -0.859375q0.3125 -0.21875 0.6875 -0.21875q0.53125 0 1.078125 0.328125l-0.359375 0.984375q-0.390625 -0.234375 -0.765625 -0.234375q-0.359375 0 -0.640625 0.21875q-0.265625 0.203125 -0.375 0.578125q-0.1875 0.5625 -0.1875 1.21875l0 3.265625l-1.046875 0z" fill-rule="nonzero"></path><path fill="#f9cb9c" d="m432.19946 163.40158l266.45673 0l0 94.58267l-266.45673 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" stroke-dasharray="8.0,6.0" d="m432.19946 163.40158l266.45673 0l0 94.58267l-266.45673 0z" fill-rule="nonzero"></path><path fill="#000000" d="m446.9651 178.62595l-1.484375 0.265625q-0.078125 -0.4375 -0.34375 -0.65625q-0.265625 -0.234375 -0.6875 -0.234375q-0.5625 0 -0.90625 0.390625q-0.328125 0.375 -0.328125 1.296875q0 1.015625 0.328125 1.4375q0.34375 0.421875 0.921875 0.421875q0.4375 0 0.703125 -0.25q0.28125 -0.25 0.390625 -0.84375l1.484375 0.25q-0.234375 1.03125 -0.890625 1.546875q-0.65625 0.515625 -1.75 0.515625q-1.265625 0 -2.015625 -0.78125q-0.734375 -0.796875 -0.734375 -2.1875q0 -1.421875 0.75 -2.203125q0.75 -0.78125 2.015625 -0.78125q1.046875 0 1.65625 0.453125q0.625 0.4375 0.890625 1.359375zm0.8327637 -1.6875l1.390625 0l0 0.78125q0.75 -0.90625 1.78125 -0.90625q0.546875 0 0.953125 0.234375q0.40625 0.21875 0.65625 0.671875q0.375 -0.453125 0.8125 -0.671875q0.4375 -0.234375 0.9375 -0.234375q0.625 0 1.0625 0.25q0.4375 0.25 0.640625 0.75q0.15625 0.359375 0.15625 1.1875l0 3.640625l-1.5 0l0 -3.265625q0 -0.84375 -0.15625 -1.09375q-0.21875 -0.3125 -0.65625 -0.3125q-0.3125 0 -0.59375 0.1875q-0.28125 0.1875 -0.40625 0.5625q-0.109375 0.375 -0.109375 1.1875l0 2.734375l-1.515625 0l0 -3.125q0 -0.828125 -0.078125 -1.0625q-0.078125 -0.25 -0.25 -0.359375q-0.171875 -0.125 -0.46875 -0.125q-0.34375 0 -0.625 0.1875q-0.28125 0.1875 -0.40625 0.546875q-0.109375 0.34375 -0.109375 1.171875l0 2.765625l-1.515625 0l0 -5.703125zm10.725098 1.734375l-1.375 -0.25q0.234375 -0.8125 0.796875 -1.203125q0.5625 -0.40625 1.671875 -0.40625q1.015625 0 1.5 0.234375q0.5 0.234375 0.703125 0.609375q0.203125 0.359375 0.203125 1.34375l-0.015625 1.765625q0 0.75 0.0625 1.109375q0.078125 0.359375 0.28125 0.765625l-1.5 0q-0.0625 -0.15625 -0.140625 -0.453125q-0.03125 -0.125 -0.046875 -0.171875q-0.390625 0.375 -0.84375 0.5625q-0.4375 0.1875 -0.9375 0.1875q-0.875 0 -1.390625 -0.46875q-0.5 -0.484375 -0.5 -1.21875q0 -0.484375 0.234375 -0.859375q0.234375 -0.375 0.640625 -0.578125q0.421875 -0.203125 1.203125 -0.359375q1.0625 -0.1875 1.46875 -0.359375l0 -0.15625q0 -0.4375 -0.21875 -0.609375q-0.21875 -0.1875 -0.8125 -0.1875q-0.390625 0 -0.625 0.15625q-0.21875 0.15625 -0.359375 0.546875zm2.015625 1.234375q-0.296875 0.09375 -0.921875 0.234375q-0.625 0.125 -0.8125 0.25q-0.296875 0.21875 -0.296875 0.53125q0 0.328125 0.234375 0.5625q0.234375 0.21875 0.59375 0.21875q0.40625 0 0.78125 -0.265625q0.28125 -0.203125 0.359375 -0.5q0.0625 -0.1875 0.0625 -0.734375l0 -0.296875zm2.7233887 2.734375l0 -7.875l1.515625 0l0 4.171875l1.765625 -2.0l1.859375 0l-1.953125 2.078125l2.09375 3.625l-1.625 0l-1.4375 -2.5625l-0.703125 0.734375l0 1.828125l-1.515625 0zm9.285858 -1.8125l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm1.6452637 5.515625l0 -0.96875l6.28125 0l0 0.96875l-6.28125 0zm6.707733 -7.875l1.390625 0l0 0.78125q0.75 -0.90625 1.78125 -0.90625q0.546875 0 0.953125 0.234375q0.40625 0.21875 0.65625 0.671875q0.375 -0.453125 0.8125 -0.671875q0.4375 -0.234375 0.9375 -0.234375q0.625 0 1.0625 0.25q0.4375 0.25 0.640625 0.75q0.15625 0.359375 0.15625 1.1875l0 3.640625l-1.5 0l0 -3.265625q0 -0.84375 -0.15625 -1.09375q-0.21875 -0.3125 -0.65625 -0.3125q-0.3125 0 -0.59375 0.1875q-0.28125 0.1875 -0.40625 0.5625q-0.109375 0.375 -0.109375 1.1875l0 2.734375l-1.515625 0l0 -3.125q0 -0.828125 -0.078125 -1.0625q-0.078125 -0.25 -0.25 -0.359375q-0.171875 -0.125 -0.46875 -0.125q-0.34375 0 -0.625 0.1875q-0.28125 0.1875 -0.40625 0.546875q-0.109375 0.34375 -0.109375 1.171875l0 2.765625l-1.515625 0l0 -5.703125zm9.600128 -0.78125l0 -1.390625l1.5 0l0 1.390625l-1.5 0zm0 6.484375l0 -5.703125l1.5 0l0 5.703125l-1.5 0zm8.148132 0l-1.515625 0l0 -2.90625q0 -0.921875 -0.09375 -1.1875q-0.09375 -0.28125 -0.3125 -0.421875q-0.21875 -0.15625 -0.53125 -0.15625q-0.390625 0 -0.703125 0.21875q-0.3125 0.203125 -0.4375 0.5625q-0.109375 0.34375 -0.109375 1.3125l0 2.578125l-1.5 0l0 -5.703125l1.40625 0l0 0.84375q0.734375 -0.96875 1.875 -0.96875q0.5 0 0.90625 0.1875q0.421875 0.171875 0.625 0.453125q0.21875 0.265625 0.296875 0.625q0.09375 0.359375 0.09375 1.015625l0 3.546875zm1.3217773 -6.484375l0 -1.390625l1.5 0l0 1.390625l-1.5 0zm0 6.484375l0 -5.703125l1.5 0l0 5.703125l-1.5 0zm2.8356323 -5.703125l1.390625 0l0 0.78125q0.75 -0.90625 1.78125 -0.90625q0.546875 0 0.953125 0.234375q0.40625 0.21875 0.65625 0.671875q0.375 -0.453125 0.8125 -0.671875q0.4375 -0.234375 0.9375 -0.234375q0.625 0 1.0625 0.25q0.4375 0.25 0.640625 0.75q0.15625 0.359375 0.15625 1.1875l0 3.640625l-1.5 0l0 -3.265625q0 -0.84375 -0.15625 -1.09375q-0.21875 -0.3125 -0.65625 -0.3125q-0.3125 0 -0.59375 0.1875q-0.28125 0.1875 -0.40625 0.5625q-0.109375 0.375 -0.109375 1.1875l0 2.734375l-1.515625 0l0 -3.125q0 -0.828125 -0.078125 -1.0625q-0.078125 -0.25 -0.25 -0.359375q-0.171875 -0.125 -0.46875 -0.125q-0.34375 0 -0.625 0.1875q-0.28125 0.1875 -0.40625 0.546875q-0.109375 0.34375 -0.109375 1.171875l0 2.765625l-1.515625 0l0 -5.703125zm13.350128 5.703125l0 -0.859375q-0.3125 0.46875 -0.828125 0.734375q-0.5 0.25 -1.0625 0.25q-0.578125 0 -1.03125 -0.25q-0.453125 -0.25 -0.671875 -0.703125q-0.203125 -0.46875 -0.203125 -1.265625l0 -3.609375l1.515625 0l0 2.625q0 1.203125 0.078125 1.46875q0.09375 0.265625 0.3125 0.4375q0.21875 0.15625 0.5625 0.15625q0.375 0 0.6875 -0.203125q0.3125 -0.21875 0.421875 -0.53125q0.109375 -0.328125 0.109375 -1.546875l0 -2.40625l1.515625 0l0 5.703125l-1.40625 0zm2.6342773 -5.703125l1.390625 0l0 0.78125q0.75 -0.90625 1.78125 -0.90625q0.546875 0 0.953125 0.234375q0.40625 0.21875 0.65625 0.671875q0.375 -0.453125 0.8125 -0.671875q0.4375 -0.234375 0.9375 -0.234375q0.625 0 1.0625 0.25q0.4375 0.25 0.640625 0.75q0.15625 0.359375 0.15625 1.1875l0 3.640625l-1.5 0l0 -3.265625q0 -0.84375 -0.15625 -1.09375q-0.21875 -0.3125 -0.65625 -0.3125q-0.3125 0 -0.59375 0.1875q-0.28125 0.1875 -0.40625 0.5625q-0.109375 0.375 -0.109375 1.1875l0 2.734375l-1.515625 0l0 -3.125q0 -0.828125 -0.078125 -1.0625q-0.078125 -0.25 -0.25 -0.359375q-0.171875 -0.125 -0.46875 -0.125q-0.34375 0 -0.625 0.1875q-0.28125 0.1875 -0.40625 0.546875q-0.109375 0.34375 -0.109375 1.171875l0 2.765625l-1.515625 0l0 -5.703125zm8.693848 7.875l0 -0.96875l6.28125 0l0 0.96875l-6.28125 0zm8.270264 -2.171875l-1.515625 0l0 -5.703125l1.40625 0l0 0.8125q0.359375 -0.578125 0.640625 -0.75q0.296875 -0.1875 0.65625 -0.1875q0.515625 0 1.0 0.28125l-0.46875 1.3125q-0.375 -0.25 -0.703125 -0.25q-0.328125 0 -0.546875 0.1875q-0.21875 0.171875 -0.34375 0.625q-0.125 0.453125 -0.125 1.90625l0 1.765625zm6.0063477 -1.8125l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm6.2702637 5.515625l0 -2.875q-0.296875 0.390625 -0.734375 0.609375q-0.4375 0.21875 -0.953125 0.21875q-0.96875 0 -1.59375 -0.734375q-0.75 -0.84375 -0.75 -2.3125q0 -1.375 0.6875 -2.140625q0.703125 -0.765625 1.734375 -0.765625q0.578125 0 0.984375 0.25q0.421875 0.234375 0.75 0.71875l0 -0.84375l1.390625 0l0 7.875l-1.515625 0zm0.046875 -5.09375q0 -0.875 -0.359375 -1.296875q-0.359375 -0.4375 -0.890625 -0.4375q-0.546875 0 -0.921875 0.4375q-0.359375 0.4375 -0.359375 1.375q0 0.953125 0.34375 1.375q0.359375 0.40625 0.890625 0.40625q0.515625 0 0.90625 -0.46875q0.390625 -0.46875 0.390625 -1.390625zm6.4935913 2.921875l0 -0.859375q-0.3125 0.46875 -0.828125 0.734375q-0.5 0.25 -1.0625 0.25q-0.578125 0 -1.03125 -0.25q-0.453125 -0.25 -0.671875 -0.703125q-0.203125 -0.46875 -0.203125 -1.265625l0 -3.609375l1.515625 0l0 2.625q0 1.203125 0.078125 1.46875q0.09375 0.265625 0.3125 0.4375q0.21875 0.15625 0.5625 0.15625q0.375 0 0.6875 -0.203125q0.3125 -0.21875 0.421875 -0.53125q0.109375 -0.328125 0.109375 -1.546875l0 -2.40625l1.515625 0l0 5.703125l-1.40625 0zm2.7592773 -6.484375l0 -1.390625l1.5 0l0 1.390625l-1.5 0zm0 6.484375l0 -5.703125l1.5 0l0 5.703125l-1.5 0zm4.3981323 0l-1.515625 0l0 -5.703125l1.40625 0l0 0.8125q0.359375 -0.578125 0.640625 -0.75q0.296875 -0.1875 0.65625 -0.1875q0.515625 0 1.0 0.28125l-0.46875 1.3125q-0.375 -0.25 -0.703125 -0.25q-0.328125 0 -0.546875 0.1875q-0.21875 0.171875 -0.34375 0.625q-0.125 0.453125 -0.125 1.90625l0 1.765625zm6.0064087 -1.8125l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm7.7702026 3.34375l-1.390625 0l0 -0.84375q-0.359375 0.5 -0.84375 0.734375q-0.46875 0.234375 -0.9375 0.234375q-0.984375 0 -1.6875 -0.78125q-0.703125 -0.796875 -0.703125 -2.21875q0 -1.453125 0.671875 -2.203125q0.6875 -0.75 1.734375 -0.75q0.953125 0 1.65625 0.796875l0 -2.84375l1.5 0l0 7.875zm-4.015625 -2.96875q0 0.90625 0.25 1.3125q0.359375 0.59375 1.015625 0.59375q0.515625 0 0.875 -0.4375q0.375 -0.453125 0.375 -1.328125q0 -0.984375 -0.359375 -1.40625q-0.34375 -0.4375 -0.90625 -0.4375q-0.53125 0 -0.890625 0.421875q-0.359375 0.421875 -0.359375 1.28125zm10.766785 5.28125l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm3.06427 -2.3125l-2.8125 -7.875l1.71875 0l2.0 5.828125l1.921875 -5.828125l1.6875 0l-2.828125 7.875l-1.6875 0zm5.09198 0l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm7.123291 0l0 -7.875l3.34375 0q1.265625 0 1.828125 0.21875q0.578125 0.203125 0.921875 0.75q0.34375 0.546875 0.34375 1.234375q0 0.890625 -0.53125 1.46875q-0.515625 0.578125 -1.546875 0.734375q0.515625 0.296875 0.84375 0.65625q0.34375 0.359375 0.90625 1.28125l0.96875 1.53125l-1.90625 0l-1.15625 -1.71875q-0.609375 -0.90625 -0.84375 -1.140625q-0.21875 -0.25 -0.46875 -0.328125q-0.25 -0.09375 -0.796875 -0.09375l-0.328125 0l0 3.28125l-1.578125 0zm1.578125 -4.546875l1.1875 0q1.140625 0 1.421875 -0.09375q0.28125 -0.09375 0.4375 -0.328125q0.171875 -0.234375 0.171875 -0.59375q0 -0.40625 -0.21875 -0.640625q-0.203125 -0.25 -0.59375 -0.3125q-0.1875 -0.03125 -1.15625 -0.03125l-1.25 0l0 2.0zm5.6956177 1.984375l1.546875 -0.15625q0.140625 0.78125 0.5625 1.15625q0.4375 0.359375 1.15625 0.359375q0.78125 0 1.15625 -0.328125q0.390625 -0.328125 0.390625 -0.75q0 -0.28125 -0.171875 -0.46875q-0.15625 -0.203125 -0.5625 -0.359375q-0.28125 -0.09375 -1.265625 -0.34375q-1.28125 -0.3125 -1.796875 -0.78125q-0.71875 -0.640625 -0.71875 -1.578125q0 -0.59375 0.328125 -1.109375q0.34375 -0.53125 0.984375 -0.8125q0.640625 -0.28125 1.5625 -0.28125q1.46875 0 2.21875 0.65625q0.75 0.65625 0.796875 1.734375l-1.59375 0.078125q-0.109375 -0.609375 -0.4375 -0.875q-0.328125 -0.265625 -1.0 -0.265625q-0.703125 0 -1.09375 0.28125q-0.25 0.1875 -0.25 0.484375q0 0.28125 0.234375 0.484375q0.296875 0.25 1.453125 0.53125q1.171875 0.265625 1.71875 0.5625q0.5625 0.28125 0.875 0.796875q0.3125 0.5 0.3125 1.25q0 0.6875 -0.375 1.28125q-0.375 0.578125 -1.0625 0.875q-0.6875 0.28125 -1.71875 0.28125q-1.484375 0 -2.296875 -0.6875q-0.796875 -0.6875 -0.953125 -2.015625zm7.46698 2.5625l0 -7.875l1.59375 0l0 7.875l-1.59375 0zm2.6950073 -3.890625q0 -1.203125 0.359375 -2.015625q0.265625 -0.609375 0.71875 -1.078125q0.46875 -0.484375 1.03125 -0.71875q0.734375 -0.3125 1.6875 -0.3125q1.75 0 2.78125 1.09375q1.046875 1.078125 1.046875 3.0q0 1.90625 -1.03125 2.984375q-1.03125 1.078125 -2.765625 1.078125q-1.765625 0 -2.796875 -1.078125q-1.03125 -1.078125 -1.03125 -2.953125zm1.625 -0.046875q0 1.328125 0.625 2.03125q0.625 0.6875 1.5625 0.6875q0.953125 0 1.5625 -0.6875q0.609375 -0.6875 0.609375 -2.0625q0 -1.34375 -0.59375 -2.015625q-0.59375 -0.671875 -1.578125 -0.671875q-0.984375 0 -1.59375 0.6875q-0.59375 0.671875 -0.59375 2.03125zm6.9919434 3.9375l0 -7.875l1.546875 0l3.234375 5.265625l0 -5.265625l1.46875 0l0 7.875l-1.59375 0l-3.171875 -5.140625l0 5.140625l-1.484375 0zm15.40625 -1.40625l0 1.40625l-5.28125 0q0.078125 -0.796875 0.5 -1.5q0.4375 -0.71875 1.703125 -1.890625q1.015625 -0.953125 1.25 -1.296875q0.3125 -0.46875 0.3125 -0.921875q0 -0.5 -0.28125 -0.765625q-0.265625 -0.28125 -0.734375 -0.28125q-0.46875 0 -0.75 0.296875q-0.28125 0.28125 -0.328125 0.9375l-1.5 -0.15625q0.125 -1.25 0.828125 -1.78125q0.71875 -0.546875 1.78125 -0.546875q1.171875 0 1.828125 0.625q0.671875 0.625 0.671875 1.5625q0 0.53125 -0.1875 1.015625q-0.1875 0.484375 -0.609375 1.0q-0.265625 0.359375 -0.984375 1.015625q-0.71875 0.65625 -0.90625 0.875q-0.1875 0.203125 -0.3125 0.40625l3.0 0zm1.1608276 1.40625l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm3.9294434 -4.25q-0.59375 -0.25 -0.859375 -0.671875q-0.265625 -0.4375 -0.265625 -0.953125q0 -0.875 0.609375 -1.453125q0.625 -0.578125 1.765625 -0.578125q1.109375 0 1.734375 0.578125q0.625 0.578125 0.625 1.453125q0 0.546875 -0.296875 0.984375q-0.28125 0.421875 -0.796875 0.640625q0.65625 0.265625 1.0 0.78125q0.34375 0.5 0.34375 1.15625q0 1.09375 -0.703125 1.78125q-0.6875 0.671875 -1.84375 0.671875q-1.078125 0 -1.78125 -0.5625q-0.84375 -0.671875 -0.84375 -1.828125q0 -0.640625 0.3125 -1.171875q0.3125 -0.53125 1.0 -0.828125zm0.3125 -1.515625q0 0.453125 0.25 0.703125q0.25 0.25 0.6875 0.25q0.421875 0 0.671875 -0.25q0.265625 -0.265625 0.265625 -0.71875q0 -0.421875 -0.265625 -0.671875q-0.25 -0.25 -0.65625 -0.25q-0.4375 0 -0.703125 0.265625q-0.25 0.25 -0.25 0.671875zm-0.140625 3.375q0 0.625 0.3125 0.96875q0.328125 0.34375 0.796875 0.34375q0.46875 0 0.765625 -0.328125q0.3125 -0.34375 0.3125 -0.96875q0 -0.5625 -0.3125 -0.890625q-0.3125 -0.34375 -0.78125 -0.34375q-0.5625 0 -0.828125 0.390625q-0.265625 0.375 -0.265625 0.828125zm4.3639526 4.703125q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0z" fill-rule="nonzero"></path><path fill="#000000" d="m441.99634 202.1572l0 -1.390625l1.5 0l0 1.390625l-1.5 0zm0 6.484375l0 -5.703125l1.5 0l0 5.703125l-1.5 0zm8.148163 0l-1.515625 0l0 -2.90625q0 -0.921875 -0.09375 -1.1875q-0.09375 -0.28125 -0.3125 -0.421875q-0.21875 -0.15625 -0.53125 -0.15625q-0.390625 0 -0.703125 0.21875q-0.3125 0.203125 -0.4375 0.5625q-0.109375 0.34375 -0.109375 1.3125l0 2.578125l-1.5 0l0 -5.703125l1.40625 0l0 0.84375q0.734375 -0.96875 1.875 -0.96875q0.5 0 0.90625 0.1875q0.421875 0.171875 0.625 0.453125q0.21875 0.265625 0.296875 0.625q0.09375 0.359375 0.09375 1.015625l0 3.546875zm6.290497 -4.015625l-1.484375 0.265625q-0.078125 -0.4375 -0.34375 -0.65625q-0.265625 -0.234375 -0.6875 -0.234375q-0.5625 0 -0.90625 0.390625q-0.328125 0.375 -0.328125 1.296875q0 1.015625 0.328125 1.4375q0.34375 0.421875 0.921875 0.421875q0.4375 0 0.703125 -0.25q0.28125 -0.25 0.390625 -0.84375l1.484375 0.25q-0.234375 1.03125 -0.890625 1.546875q-0.65625 0.515625 -1.75 0.515625q-1.265625 0 -2.015625 -0.78125q-0.734375 -0.796875 -0.734375 -2.1875q0 -1.421875 0.75 -2.203125q0.75 -0.78125 2.015625 -0.78125q1.046875 0 1.65625 0.453125q0.625 0.4375 0.890625 1.359375zm0.9577637 4.015625l0 -7.875l1.5 0l0 7.875l-1.5 0zm6.7106323 0l0 -0.859375q-0.3125 0.46875 -0.828125 0.734375q-0.5 0.25 -1.0625 0.25q-0.578125 0 -1.03125 -0.25q-0.453125 -0.25 -0.671875 -0.703125q-0.203125 -0.46875 -0.203125 -1.265625l0 -3.609375l1.515625 0l0 2.625q0 1.203125 0.078125 1.46875q0.09375 0.265625 0.3125 0.4375q0.21875 0.15625 0.5625 0.15625q0.375 0 0.6875 -0.203125q0.3125 -0.21875 0.421875 -0.53125q0.109375 -0.328125 0.109375 -1.546875l0 -2.40625l1.515625 0l0 5.703125l-1.40625 0zm7.9780273 0l-1.390625 0l0 -0.84375q-0.359375 0.5 -0.84375 0.734375q-0.46875 0.234375 -0.9375 0.234375q-0.984375 0 -1.6875 -0.78125q-0.703125 -0.796875 -0.703125 -2.21875q0 -1.453125 0.671875 -2.203125q0.6875 -0.75 1.734375 -0.75q0.953125 0 1.65625 0.796875l0 -2.84375l1.5 0l0 7.875zm-4.015625 -2.96875q0 0.90625 0.25 1.3125q0.359375 0.59375 1.015625 0.59375q0.515625 0 0.875 -0.4375q0.375 -0.453125 0.375 -1.328125q0 -0.984375 -0.359375 -1.40625q-0.34375 -0.4375 -0.90625 -0.4375q-0.53125 0 -0.890625 0.421875q-0.359375 0.421875 -0.359375 1.28125zm8.602997 1.15625l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm5.0515137 5.65625l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm6.0954895 -5.203125l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm4.148712 -4.984375l0 2.890625q0.734375 -0.84375 1.75 -0.84375q0.515625 0 0.921875 0.1875q0.421875 0.1875 0.640625 0.5q0.21875 0.296875 0.296875 0.671875q0.078125 0.359375 0.078125 1.125l0 3.34375l-1.515625 0l0 -3.015625q0 -0.890625 -0.09375 -1.125q-0.078125 -0.25 -0.296875 -0.390625q-0.21875 -0.140625 -0.546875 -0.140625q-0.375 0 -0.671875 0.1875q-0.296875 0.171875 -0.4375 0.546875q-0.125 0.359375 -0.125 1.078125l0 2.859375l-1.515625 0l0 -7.875l1.515625 0zm8.306152 6.0625l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm7.520233 -0.671875l-1.484375 0.265625q-0.078125 -0.4375 -0.34375 -0.65625q-0.265625 -0.234375 -0.6875 -0.234375q-0.5625 0 -0.90625 0.390625q-0.328125 0.375 -0.328125 1.296875q0 1.015625 0.328125 1.4375q0.34375 0.421875 0.921875 0.421875q0.4375 0 0.703125 -0.25q0.28125 -0.25 0.390625 -0.84375l1.484375 0.25q-0.234375 1.03125 -0.890625 1.546875q-0.65625 0.515625 -1.75 0.515625q-1.265625 0 -2.015625 -0.78125q-0.734375 -0.796875 -0.734375 -2.1875q0 -1.421875 0.75 -2.203125q0.75 -0.78125 2.015625 -0.78125q1.046875 0 1.65625 0.453125q0.625 0.4375 0.890625 1.359375zm0.8952637 4.015625l0 -7.875l1.515625 0l0 4.171875l1.765625 -2.0l1.859375 0l-1.953125 2.078125l2.09375 3.625l-1.625 0l-1.4375 -2.5625l-0.703125 0.734375l0 1.828125l-1.515625 0zm11.035889 -2.890625l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm1.8518066 2.890625l2.6875 -4.109375l-2.4375 -3.765625l1.859375 0l1.578125 2.53125l1.546875 -2.53125l1.84375 0l-2.453125 3.828125l2.703125 4.046875l-1.921875 0l-1.75 -2.71875l-1.75 2.71875l-1.90625 0zm7.107666 0l2.6875 -4.109375l-2.4375 -3.765625l1.859375 0l1.578125 2.53125l1.546875 -2.53125l1.84375 0l-2.453125 3.828125l2.703125 4.046875l-1.921875 0l-1.75 -2.71875l-1.75 2.71875l-1.90625 0zm12.951355 -2.890625l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm2.2893677 -0.046875q0 -0.75 0.375 -1.453125q0.375 -0.703125 1.046875 -1.0625q0.6875 -0.375 1.515625 -0.375q1.296875 0 2.125 0.84375q0.828125 0.828125 0.828125 2.109375q0 1.296875 -0.84375 2.15625q-0.828125 0.84375 -2.09375 0.84375q-0.78125 0 -1.5 -0.34375q-0.703125 -0.359375 -1.078125 -1.046875q-0.375 -0.6875 -0.375 -1.671875zm1.546875 0.078125q0 0.859375 0.40625 1.3125q0.40625 0.453125 1.0 0.453125q0.578125 0 0.984375 -0.453125q0.40625 -0.453125 0.40625 -1.3125q0 -0.84375 -0.40625 -1.296875q-0.40625 -0.453125 -0.984375 -0.453125q-0.59375 0 -1.0 0.453125q-0.40625 0.453125 -0.40625 1.296875zm5.1967773 -2.84375l1.390625 0l0 0.78125q0.75 -0.90625 1.78125 -0.90625q0.546875 0 0.953125 0.234375q0.40625 0.21875 0.65625 0.671875q0.375 -0.453125 0.8125 -0.671875q0.4375 -0.234375 0.9375 -0.234375q0.625 0 1.0625 0.25q0.4375 0.25 0.640625 0.75q0.15625 0.359375 0.15625 1.1875l0 3.640625l-1.5 0l0 -3.265625q0 -0.84375 -0.15625 -1.09375q-0.21875 -0.3125 -0.65625 -0.3125q-0.3125 0 -0.59375 0.1875q-0.28125 0.1875 -0.40625 0.5625q-0.109375 0.375 -0.109375 1.1875l0 2.734375l-1.515625 0l0 -3.125q0 -0.828125 -0.078125 -1.0625q-0.078125 -0.25 -0.25 -0.359375q-0.171875 -0.125 -0.46875 -0.125q-0.34375 0 -0.625 0.1875q-0.28125 0.1875 -0.40625 0.546875q-0.109375 0.34375 -0.109375 1.171875l0 2.765625l-1.515625 0l0 -5.703125zm9.553223 0l1.40625 0l0 0.84375q0.265625 -0.4375 0.734375 -0.703125q0.46875 -0.265625 1.046875 -0.265625q0.984375 0 1.671875 0.78125q0.703125 0.765625 0.703125 2.15625q0 1.4375 -0.703125 2.234375q-0.6875 0.78125 -1.6875 0.78125q-0.46875 0 -0.859375 -0.1875q-0.375 -0.1875 -0.8125 -0.640625l0 2.875l-1.5 0l0 -7.875zm1.484375 2.75q0 0.96875 0.375 1.4375q0.390625 0.453125 0.9375 0.453125q0.53125 0 0.875 -0.421875q0.359375 -0.421875 0.359375 -1.390625q0 -0.890625 -0.359375 -1.328125q-0.359375 -0.4375 -0.890625 -0.4375q-0.5625 0 -0.9375 0.4375q-0.359375 0.421875 -0.359375 1.25zm5.0717773 -3.53125l0 -1.390625l1.5 0l0 1.390625l-1.5 0zm0 6.484375l0 -5.703125l1.5 0l0 5.703125l-1.5 0zm2.9606323 0l0 -7.875l1.5 0l0 7.875l-1.5 0zm6.2575073 -1.8125l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm3.9890137 3.34375l-1.515625 0l0 -5.703125l1.40625 0l0 0.8125q0.359375 -0.578125 0.640625 -0.75q0.296875 -0.1875 0.65625 -0.1875q0.515625 0 1.0 0.28125l-0.46875 1.3125q-0.375 -0.25 -0.703125 -0.25q-0.328125 0 -0.546875 0.1875q-0.21875 0.171875 -0.34375 0.625q-0.125 0.453125 -0.125 1.90625l0 1.765625zm2.7250977 0l0 -7.875l5.390625 0l0 1.328125l-3.796875 0l0 1.875l3.28125 0l0 1.328125l-3.28125 0l0 3.34375l-1.59375 0zm6.4936523 0l0 -7.875l1.5 0l0 7.875l-1.5 0zm4.0856323 -3.96875l-1.375 -0.25q0.234375 -0.8125 0.796875 -1.203125q0.5625 -0.40625 1.671875 -0.40625q1.015625 0 1.5 0.234375q0.5 0.234375 0.703125 0.609375q0.203125 0.359375 0.203125 1.34375l-0.015625 1.765625q0 0.75 0.0625 1.109375q0.078125 0.359375 0.28125 0.765625l-1.5 0q-0.0625 -0.15625 -0.140625 -0.453125q-0.03125 -0.125 -0.046875 -0.171875q-0.390625 0.375 -0.84375 0.5625q-0.4375 0.1875 -0.9375 0.1875q-0.875 0 -1.390625 -0.46875q-0.5 -0.484375 -0.5 -1.21875q0 -0.484375 0.234375 -0.859375q0.234375 -0.375 0.640625 -0.578125q0.421875 -0.203125 1.203125 -0.359375q1.0625 -0.1875 1.46875 -0.359375l0 -0.15625q0 -0.4375 -0.21875 -0.609375q-0.21875 -0.1875 -0.8125 -0.1875q-0.390625 0 -0.625 0.15625q-0.21875 0.15625 -0.359375 0.546875zm2.015625 1.234375q-0.296875 0.09375 -0.921875 0.234375q-0.625 0.125 -0.8125 0.25q-0.296875 0.21875 -0.296875 0.53125q0 0.328125 0.234375 0.5625q0.234375 0.21875 0.59375 0.21875q0.40625 0 0.78125 -0.265625q0.28125 -0.203125 0.359375 -0.5q0.0625 -0.1875 0.0625 -0.734375l0 -0.296875zm2.6452637 3.109375l1.71875 0.203125q0.046875 0.3125 0.203125 0.421875q0.203125 0.15625 0.671875 0.15625q0.59375 0 0.890625 -0.171875q0.1875 -0.125 0.296875 -0.375q0.0625 -0.203125 0.0625 -0.703125l0 -0.828125q-0.671875 0.921875 -1.703125 0.921875q-1.140625 0 -1.8125 -0.96875q-0.53125 -0.765625 -0.53125 -1.921875q0 -1.421875 0.6875 -2.171875q0.6875 -0.765625 1.71875 -0.765625q1.0625 0 1.75 0.921875l0 -0.796875l1.40625 0l0 5.125q0 1.0 -0.171875 1.5q-0.15625 0.5 -0.46875 0.78125q-0.296875 0.296875 -0.796875 0.453125q-0.5 0.15625 -1.265625 0.15625q-1.453125 0 -2.0625 -0.5q-0.609375 -0.484375 -0.609375 -1.25q0 -0.078125 0.015625 -0.1875zm1.34375 -3.34375q0 0.90625 0.34375 1.328125q0.359375 0.421875 0.875 0.421875q0.546875 0 0.921875 -0.421875q0.390625 -0.4375 0.390625 -1.296875q0 -0.875 -0.375 -1.296875q-0.359375 -0.4375 -0.90625 -0.4375q-0.546875 0 -0.90625 0.421875q-0.34375 0.421875 -0.34375 1.28125zm4.8842773 5.28125q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0z" fill-rule="nonzero"></path><path fill="#000000" d="m441.4651 220.01657l1.5 -0.234375q0.109375 0.4375 0.390625 0.671875q0.296875 0.21875 0.828125 0.21875q0.59375 0 0.890625 -0.203125q0.203125 -0.15625 0.203125 -0.40625q0 -0.171875 -0.109375 -0.28125q-0.125 -0.109375 -0.515625 -0.203125q-1.8125 -0.40625 -2.3125 -0.734375q-0.671875 -0.46875 -0.671875 -1.28125q0 -0.75 0.578125 -1.25q0.59375 -0.5 1.828125 -0.5q1.171875 0 1.734375 0.375q0.5625 0.375 0.78125 1.125l-1.421875 0.265625q-0.09375 -0.328125 -0.34375 -0.5q-0.25 -0.1875 -0.734375 -0.1875q-0.59375 0 -0.84375 0.171875q-0.171875 0.109375 -0.171875 0.296875q0 0.171875 0.140625 0.28125q0.203125 0.15625 1.40625 0.4375q1.21875 0.265625 1.6875 0.65625q0.484375 0.40625 0.484375 1.125q0 0.78125 -0.65625 1.34375q-0.65625 0.5625 -1.953125 0.5625q-1.15625 0 -1.84375 -0.46875q-0.671875 -0.46875 -0.875 -1.28125zm9.754639 -0.1875l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm5.160858 -2.359375l0 1.203125l-1.03125 0l0 2.296875q0 0.703125 0.03125 0.828125q0.03125 0.109375 0.125 0.1875q0.109375 0.0625 0.265625 0.0625q0.203125 0 0.609375 -0.140625l0.125 1.171875q-0.53125 0.21875 -1.1875 0.21875q-0.421875 0 -0.75 -0.125q-0.328125 -0.140625 -0.484375 -0.359375q-0.140625 -0.21875 -0.203125 -0.59375q-0.046875 -0.265625 -0.046875 -1.0625l0 -2.484375l-0.6875 0l0 -1.203125l0.6875 0l0 -1.140625l1.515625 -0.875l0 2.015625l1.03125 0zm3.4392395 8.015625l-1.046875 0q-0.8125 -1.234375 -1.25 -2.5625q-0.421875 -1.34375 -0.421875 -2.609375q0 -1.546875 0.53125 -2.9375q0.453125 -1.203125 1.171875 -2.21875l1.03125 0q-0.734375 1.640625 -1.015625 2.78125q-0.28125 1.140625 -0.28125 2.421875q0 0.875 0.15625 1.796875q0.171875 0.921875 0.453125 1.75q0.1875 0.5625 0.671875 1.578125zm6.0954895 -5.203125l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm1.8518677 2.890625l2.6875 -4.109375l-2.4375 -3.765625l1.859375 0l1.578125 2.53125l1.546875 -2.53125l1.84375 0l-2.453125 3.828125l2.703125 4.046875l-1.921875 0l-1.75 -2.71875l-1.75 2.71875l-1.90625 0zm7.1076355 0l2.6875 -4.109375l-2.4375 -3.765625l1.859375 0l1.578125 2.53125l1.546875 -2.53125l1.84375 0l-2.453125 3.828125l2.703125 4.046875l-1.921875 0l-1.75 -2.71875l-1.75 2.71875l-1.90625 0zm6.9982605 2.171875l0 -0.96875l6.28125 0l0 0.96875l-6.28125 0zm6.848358 -2.171875l0 -7.875l5.390625 0l0 1.328125l-3.796875 0l0 1.875l3.28125 0l0 1.328125l-3.28125 0l0 3.34375l-1.59375 0zm6.5405273 0l0 -7.8125l1.59375 0l0 6.484375l3.953125 0l0 1.328125l-5.546875 0zm13.571747 0l-1.734375 0l-0.6875 -1.78125l-3.140625 0l-0.65625 1.78125l-1.6875 0l3.0625 -7.875l1.6875 0l3.15625 7.875zm-2.9375 -3.109375l-1.078125 -2.921875l-1.0625 2.921875l2.140625 0zm7.195587 0.21875l0 -1.328125l3.421875 0l0 3.140625q-0.5 0.46875 -1.453125 0.84375q-0.9375 0.375 -1.921875 0.375q-1.2343445 0 -2.1562195 -0.515625q-0.90625 -0.53125 -1.375 -1.484375q-0.453125 -0.96875 -0.453125 -2.109375q0 -1.21875 0.515625 -2.171875q0.515625 -0.96875 1.5 -1.46875q0.7655945 -0.40625 1.8905945 -0.40625q1.46875 0 2.28125 0.625q0.828125 0.609375 1.0625 1.703125l-1.578125 0.296875q-0.15625 -0.59375 -0.625 -0.921875q-0.453125 -0.34375 -1.140625 -0.34375q-1.046875 0 -1.6562195 0.671875q-0.609375 0.65625 -0.609375 1.953125q0 1.40625 0.609375 2.109375q0.6249695 0.703125 1.6405945 0.703125q0.5 0 1.0 -0.1875q0.5 -0.203125 0.859375 -0.484375l0 -1.0l-1.8125 0zm4.2106934 0.328125l1.546875 -0.15625q0.140625 0.78125 0.5625 1.15625q0.4375 0.359375 1.15625 0.359375q0.78125 0 1.15625 -0.328125q0.390625 -0.328125 0.390625 -0.75q0 -0.28125 -0.171875 -0.46875q-0.15625 -0.203125 -0.5625 -0.359375q-0.28125 -0.09375 -1.265625 -0.34375q-1.28125 -0.3125 -1.796875 -0.78125q-0.71875 -0.640625 -0.71875 -1.578125q0 -0.59375 0.328125 -1.109375q0.34375 -0.53125 0.984375 -0.8125q0.640625 -0.28125 1.5625 -0.28125q1.46875 0 2.21875 0.65625q0.75 0.65625 0.796875 1.734375l-1.59375 0.078125q-0.109375 -0.609375 -0.4375 -0.875q-0.328125 -0.265625 -1.0 -0.265625q-0.703125 0 -1.09375 0.28125q-0.25 0.1875 -0.25 0.484375q0 0.28125 0.234375 0.484375q0.296875 0.25 1.453125 0.53125q1.171875 0.265625 1.71875 0.5625q0.5625 0.28125 0.875 0.796875q0.3125 0.5 0.3125 1.25q0 0.6875 -0.375 1.28125q-0.375 0.578125 -1.0625 0.875q-0.6875 0.28125 -1.71875 0.28125q-1.484375 0 -2.296875 -0.6875q-0.796875 -0.6875 -0.953125 -2.015625zm10.568237 -2.515625l-0.28125 -1.46875l0 -1.328125l1.640625 0l0 1.328125l-0.25 1.46875l-1.109375 0zm2.421875 0l-0.28125 -1.46875l0 -1.328125l1.640625 0l0 1.328125l-0.25 1.46875l-1.109375 0zm4.364868 3.90625l0 -2.25q-1.046875 -0.296875 -1.546875 -0.90625q-0.484375 -0.609375 -0.484375 -1.484375q0 -0.875 0.546875 -1.46875q0.5625 -0.609375 1.484375 -0.6875l0 -0.546875l0.78125 0l0 0.546875q0.859375 0.09375 1.359375 0.578125q0.5 0.484375 0.640625 1.28125l-1.359375 0.171875q-0.125 -0.625 -0.640625 -0.859375l0 2.109375q1.28125 0.34375 1.75 0.90625q0.46875 0.546875 0.46875 1.421875q0 0.96875 -0.59375 1.640625q-0.578125 0.671875 -1.625 0.8125l0 1.0l-0.78125 0l0 -0.96875q-0.921875 -0.109375 -1.515625 -0.6875q-0.578125 -0.59375 -0.734375 -1.640625l1.40625 -0.15625q0.078125 0.4375 0.3125 0.75q0.25 0.296875 0.53125 0.4375zm0 -5.640625q-0.3125 0.109375 -0.5 0.375q-0.1875 0.25 -0.1875 0.5625q0 0.28125 0.171875 0.53125q0.171875 0.25 0.515625 0.390625l0 -1.859375zm0.78125 5.71875q0.40625 -0.078125 0.65625 -0.375q0.25 -0.296875 0.25 -0.703125q0 -0.359375 -0.21875 -0.609375q-0.203125 -0.265625 -0.6875 -0.40625l0 2.09375zm2.8483887 -1.109375l0 -1.28125q0.390625 -0.03125 0.578125 -0.125q0.203125 -0.09375 0.34375 -0.3125q0.15625 -0.234375 0.203125 -0.578125q0.046875 -0.265625 0.046875 -0.90625q0 -1.046875 0.09375 -1.453125q0.09375 -0.40625 0.34375 -0.65625q0.265625 -0.25 0.75 -0.390625q0.328125 -0.109375 1.03125 -0.109375l0.28125 0l0 1.296875q-0.59375 0 -0.765625 0.0625q-0.171875 0.0625 -0.25 0.203125q-0.078125 0.140625 -0.078125 0.46875q0 0.34375 -0.046875 1.28125q-0.03125 0.53125 -0.15625 0.875q-0.109375 0.328125 -0.28125 0.546875q-0.171875 0.203125 -0.53125 0.4375q0.3125 0.171875 0.515625 0.421875q0.203125 0.25 0.3125 0.59375q0.109375 0.34375 0.140625 0.9375q0.03125 0.890625 0.03125 1.140625q0 0.359375 0.078125 0.484375q0.09375 0.140625 0.265625 0.21875q0.1875 0.078125 0.765625 0.078125l0 1.28125l-0.28125 0q-0.734375 0 -1.125 -0.125q-0.375 -0.109375 -0.640625 -0.375q-0.265625 -0.265625 -0.359375 -0.671875q-0.09375 -0.390625 -0.09375 -1.234375q0 -0.984375 -0.09375 -1.28125q-0.109375 -0.4375 -0.359375 -0.609375q-0.234375 -0.1875 -0.71875 -0.21875zm9.662598 -0.6875l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm1.8518677 2.890625l2.6875 -4.109375l-2.4375 -3.765625l1.859375 0l1.578125 2.53125l1.546875 -2.53125l1.84375 0l-2.453125 3.828125l2.703125 4.046875l-1.921875 0l-1.75 -2.71875l-1.75 2.71875l-1.90625 0zm7.107605 0l2.6875 -4.109375l-2.4375 -3.765625l1.859375 0l1.578125 2.53125l1.546875 -2.53125l1.84375 0l-2.453125 3.828125l2.703125 4.046875l-1.921875 0l-1.75 -2.71875l-1.75 2.71875l-1.90625 0zm6.998291 2.171875l0 -0.96875l6.28125 0l0 0.96875l-6.28125 0zm6.8483887 -2.171875l0 -7.875l5.390625 0l0 1.328125l-3.796875 0l0 1.875l3.28125 0l0 1.328125l-3.28125 0l0 3.34375l-1.59375 0zm6.5404663 0l0 -7.8125l1.59375 0l0 6.484375l3.953125 0l0 1.328125l-5.546875 0zm13.571777 0l-1.734375 0l-0.6875 -1.78125l-3.140625 0l-0.65625 1.78125l-1.6875 0l3.0625 -7.875l1.6875 0l3.15625 7.875zm-2.9375 -3.109375l-1.078125 -2.921875l-1.0625 2.921875l2.140625 0zm7.1956177 0.21875l0 -1.328125l3.421875 0l0 3.140625q-0.5 0.46875 -1.453125 0.84375q-0.9375 0.375 -1.921875 0.375q-1.234375 0 -2.15625 -0.515625q-0.90625 -0.53125 -1.375 -1.484375q-0.453125 -0.96875 -0.453125 -2.109375q0 -1.21875 0.515625 -2.171875q0.515625 -0.96875 1.5 -1.46875q0.765625 -0.40625 1.890625 -0.40625q1.46875 0 2.28125 0.625q0.828125 0.609375 1.0625 1.703125l-1.578125 0.296875q-0.15625 -0.59375 -0.625 -0.921875q-0.453125 -0.34375 -1.140625 -0.34375q-1.046875 0 -1.65625 0.671875q-0.609375 0.65625 -0.609375 1.953125q0 1.40625 0.609375 2.109375q0.625 0.703125 1.640625 0.703125q0.5 0 1.0 -0.1875q0.5 -0.203125 0.859375 -0.484375l0 -1.0l-1.8125 0zm4.2106323 0.328125l1.546875 -0.15625q0.140625 0.78125 0.5625 1.15625q0.4375 0.359375 1.15625 0.359375q0.78125 0 1.15625 -0.328125q0.390625 -0.328125 0.390625 -0.75q0 -0.28125 -0.171875 -0.46875q-0.15625 -0.203125 -0.5625 -0.359375q-0.28125 -0.09375 -1.265625 -0.34375q-1.28125 -0.3125 -1.796875 -0.78125q-0.71875 -0.640625 -0.71875 -1.578125q0 -0.59375 0.328125 -1.109375q0.34375 -0.53125 0.984375 -0.8125q0.640625 -0.28125 1.5625 -0.28125q1.46875 0 2.21875 0.65625q0.75 0.65625 0.796875 1.734375l-1.59375 0.078125q-0.109375 -0.609375 -0.4375 -0.875q-0.328125 -0.265625 -1.0 -0.265625q-0.703125 0 -1.09375 0.28125q-0.25 0.1875 -0.25 0.484375q0 0.28125 0.234375 0.484375q0.296875 0.25 1.453125 0.53125q1.171875 0.265625 1.71875 0.5625q0.5625 0.28125 0.875 0.796875q0.3125 0.5 0.3125 1.25q0 0.6875 -0.375 1.28125q-0.375 0.578125 -1.0625 0.875q-0.6875 0.28125 -1.71875 0.28125q-1.484375 0 -2.296875 -0.6875q-0.796875 -0.6875 -0.953125 -2.015625zm10.638916 0.359375q-0.390625 0.015625 -0.59375 0.125q-0.203125 0.09375 -0.34375 0.328125q-0.140625 0.21875 -0.203125 0.5625q-0.046875 0.265625 -0.046875 0.890625q0 1.046875 -0.09375 1.453125q-0.09375 0.421875 -0.34375 0.671875q-0.25 0.25 -0.734375 0.390625q-0.328125 0.09375 -1.03125 0.09375l-0.296875 0l0 -1.28125q0.578125 0 0.75 -0.078125q0.1875 -0.078125 0.265625 -0.21875q0.09375 -0.125 0.09375 -0.453125q0 -0.328125 0.046875 -1.25q0.015625 -0.5625 0.140625 -0.890625q0.125 -0.34375 0.3125 -0.578125q0.203125 -0.234375 0.5 -0.40625q-0.390625 -0.265625 -0.578125 -0.515625q-0.25 -0.34375 -0.34375 -0.890625q-0.0625 -0.375 -0.078125 -1.65625q-0.015625 -0.40625 -0.078125 -0.53125q-0.0625 -0.140625 -0.25 -0.203125q-0.171875 -0.078125 -0.78125 -0.078125l0 -1.296875l0.296875 0q0.71875 0 1.109375 0.125q0.390625 0.109375 0.640625 0.390625q0.265625 0.265625 0.359375 0.65625q0.09375 0.390625 0.09375 1.234375q0 0.984375 0.078125 1.296875q0.125 0.421875 0.359375 0.609375q0.25 0.1875 0.75 0.21875l0 1.28125zm3.795105 0.109375l0 -1.515625l2.96875 0l0 1.515625l-2.96875 0zm3.204895 0.46875l1.5 -0.234375q0.109375 0.4375 0.390625 0.671875q0.296875 0.21875 0.828125 0.21875q0.59375 0 0.890625 -0.203125q0.203125 -0.15625 0.203125 -0.40625q0 -0.171875 -0.109375 -0.28125q-0.125 -0.109375 -0.515625 -0.203125q-1.8125 -0.40625 -2.3125 -0.734375q-0.671875 -0.46875 -0.671875 -1.28125q0 -0.75 0.578125 -1.25q0.59375 -0.5 1.828125 -0.5q1.171875 0 1.734375 0.375q0.5625 0.375 0.78125 1.125l-1.421875 0.265625q-0.09375 -0.328125 -0.34375 -0.5q-0.25 -0.1875 -0.734375 -0.1875q-0.59375 0 -0.84375 0.171875q-0.171875 0.109375 -0.171875 0.296875q0 0.171875 0.140625 0.28125q0.203125 0.15625 1.40625 0.4375q1.21875 0.265625 1.6875 0.65625q0.484375 0.40625 0.484375 1.125q0 0.78125 -0.65625 1.34375q-0.65625 0.5625 -1.953125 0.5625q-1.15625 0 -1.84375 -0.46875q-0.671875 -0.46875 -0.875 -1.28125zm9.067078 -4.078125l0 1.203125l-1.03125 0l0 2.296875q0 0.703125 0.03125 0.828125q0.03125 0.109375 0.125 0.1875q0.109375 0.0625 0.265625 0.0625q0.203125 0 0.609375 -0.140625l0.125 1.171875q-0.53125 0.21875 -1.1875 0.21875q-0.421875 0 -0.75 -0.125q-0.328125 -0.140625 -0.484375 -0.359375q-0.140625 -0.21875 -0.203125 -0.59375q-0.046875 -0.265625 -0.046875 -1.0625l0 -2.484375l-0.6875 0l0 -1.203125l0.6875 0l0 -1.140625l1.515625 -0.875l0 2.015625l1.03125 0zm6.15802 5.703125l-1.390625 0l0 -0.84375q-0.359375 0.5 -0.84375 0.734375q-0.46875 0.234375 -0.9375 0.234375q-0.984375 0 -1.6875 -0.78125q-0.703125 -0.796875 -0.703125 -2.21875q0 -1.453125 0.671875 -2.203125q0.6875 -0.75 1.734375 -0.75q0.953125 0 1.65625 0.796875l0 -2.84375l1.5 0l0 7.875zm-4.015625 -2.96875q0 0.90625 0.25 1.3125q0.359375 0.59375 1.015625 0.59375q0.515625 0 0.875 -0.4375q0.375 -0.453125 0.375 -1.328125q0 -0.984375 -0.359375 -1.40625q-0.34375 -0.4375 -0.90625 -0.4375q-0.53125 0 -0.890625 0.421875q-0.359375 0.421875 -0.359375 1.28125zm4.9624023 -1.421875l0 -1.375l5.515625 0l0 1.375l-5.515625 0zm0 2.390625l0 -1.390625l5.515625 0l0 1.390625l-5.515625 0zm11.5355835 -2.015625l-1.484375 0.265625q-0.078125 -0.4375 -0.34375 -0.65625q-0.265625 -0.234375 -0.6875 -0.234375q-0.5625 0 -0.90625 0.390625q-0.328125 0.375 -0.328125 1.296875q0 1.015625 0.328125 1.4375q0.34375 0.421875 0.921875 0.421875q0.4375 0 0.703125 -0.25q0.28125 -0.25 0.390625 -0.84375l1.484375 0.25q-0.234375 1.03125 -0.890625 1.546875q-0.65625 0.515625 -1.75 0.515625q-1.265625 0 -2.015625 -0.78125q-0.734375 -0.796875 -0.734375 -2.1875q0 -1.421875 0.75 -2.203125q0.75 -0.78125 2.015625 -0.78125q1.046875 0 1.65625 0.453125q0.625 0.4375 0.890625 1.359375zm2.6764526 2.875l0 -2.046875l-2.0625 0l0 -1.40625l2.0625 0l0 -2.046875l1.375 0l0 2.046875l2.078125 0l0 1.40625l-2.078125 0l0 2.046875l-1.375 0zm6.2230835 0l0 -2.046875l-2.0625 0l0 -1.40625l2.0625 0l0 -2.046875l1.375 0l0 2.046875l2.078125 0l0 1.40625l-2.078125 0l0 2.046875l-1.375 0zm8.0355835 1.140625l-1.515625 0l0 -5.6875q-0.8125 0.765625 -1.9375 1.140625l0 -1.359375q0.59375 -0.203125 1.28125 -0.734375q0.6875 -0.546875 0.953125 -1.265625l1.21875 0l0 7.90625zm5.9265137 0l-1.515625 0l0 -5.6875q-0.8125 0.765625 -1.9375 1.140625l0 -1.359375q0.59375 -0.203125 1.28125 -0.734375q0.6875 -0.546875 0.953125 -1.265625l1.21875 0l0 7.90625zm2.4890137 -5.078125l-0.28125 -1.46875l0 -1.328125l1.640625 0l0 1.328125l-0.25 1.46875l-1.109375 0zm2.421875 0l-0.28125 -1.46875l0 -1.328125l1.640625 0l0 1.328125l-0.25 1.46875l-1.109375 0zm2.1148682 7.390625q0.4375 -0.953125 0.625 -1.46875q0.1875 -0.5 0.34375 -1.171875q0.15625 -0.671875 0.21875 -1.265625q0.078125 -0.59375 0.078125 -1.21875q0 -1.28125 -0.28125 -2.421875q-0.265625 -1.140625 -1.0 -2.78125l1.03125 0q0.8125 1.15625 1.25 2.453125q0.453125 1.296875 0.453125 2.625q0 1.125 -0.359375 2.40625q-0.390625 1.4375 -1.328125 2.84375l-1.03125 0z" fill-rule="nonzero"></path><path fill="#000000" d="m441.99634 234.64157l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606628 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606323 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m305.03412 275.87927l0 -12.666748l22.11023 0l0 -12.656097" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m305.03412 275.87927l0 -12.666748l22.11023 0l0 -12.656097" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m376.1181 231.86876l28.040527 0l0 -21.165344l28.053955 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m376.1181 231.86876l28.040527 0l0 -21.165344l28.053955 0" fill-rule="nonzero"></path><path fill="#dd7e6b" d="m162.67192 239.17061l93.84242 0l4.078827 4.0788116l0 20.393616l-97.92125 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m162.67192 239.17061l93.84242 0l4.078827 4.0788116l0 20.393616l-97.92125 0z" fill-rule="nonzero"></path><path fill="#000000" d="m202.76714 256.94653l0 -0.78125q-0.59375 0.921875 -1.734375 0.921875q-0.75 0 -1.375 -0.40625q-0.625 -0.421875 -0.96875 -1.15625q-0.34375 -0.734375 -0.34375 -1.6875q0 -0.921875 0.3125 -1.6875q0.3125 -0.765625 0.9375 -1.15625q0.625 -0.40625 1.390625 -0.40625q0.5625 0 1.0 0.234375q0.4375 0.234375 0.71875 0.609375l0 -3.078125l1.046875 0l0 8.59375l-0.984375 0zm-3.328125 -3.109375q0 1.203125 0.5 1.796875q0.5 0.578125 1.1875 0.578125q0.6875 0 1.171875 -0.5625q0.484375 -0.5625 0.484375 -1.71875q0 -1.28125 -0.5 -1.875q-0.484375 -0.59375 -1.203125 -0.59375q-0.703125 0 -1.171875 0.578125q-0.46875 0.5625 -0.46875 1.796875zm5.580078 0q0 -1.734375 0.953125 -2.5625q0.796875 -0.6875 1.953125 -0.6875q1.28125 0 2.09375 0.84375q0.828125 0.828125 0.828125 2.3125q0 1.203125 -0.359375 1.890625q-0.359375 0.6875 -1.0625 1.078125q-0.6875 0.375 -1.5 0.375q-1.296875 0 -2.109375 -0.828125q-0.796875 -0.84375 -0.796875 -2.421875zm1.078125 0q0 1.1875 0.515625 1.78125q0.53125 0.59375 1.3125 0.59375q0.796875 0 1.3125 -0.59375q0.515625 -0.59375 0.515625 -1.8125q0 -1.15625 -0.53125 -1.75q-0.515625 -0.59375 -1.296875 -0.59375q-0.78125 0 -1.3125 0.59375q-0.515625 0.578125 -0.515625 1.78125zm5.283203 3.109375l2.265625 -3.234375l-2.09375 -2.984375l1.3125 0l0.953125 1.453125q0.265625 0.421875 0.4375 0.703125q0.25 -0.390625 0.46875 -0.6875l1.046875 -1.46875l1.265625 0l-2.15625 2.921875l2.3125 3.296875l-1.28125 0l-1.28125 -1.9375l-0.34375 -0.515625l-1.640625 2.453125l-1.265625 0zm6.65625 2.390625l-0.125 -0.984375q0.34375 0.09375 0.609375 0.09375q0.34375 0 0.546875 -0.125q0.21875 -0.109375 0.359375 -0.3125q0.09375 -0.171875 0.328125 -0.796875q0.015625 -0.078125 0.09375 -0.25l-2.375 -6.234375l1.140625 0l1.296875 3.59375q0.25 0.6875 0.453125 1.453125q0.1875 -0.734375 0.4375 -1.421875l1.328125 -3.625l1.046875 0l-2.359375 6.328125q-0.390625 1.015625 -0.59375 1.40625q-0.28125 0.53125 -0.65625 0.765625q-0.359375 0.25 -0.859375 0.25q-0.296875 0 -0.671875 -0.140625z" fill-rule="nonzero"></path><path fill="#e6b8af" d="m162.67192 135.78215l172.85039 0l0 67.90552l-172.85039 0z" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" stroke-dasharray="8.0,6.0" d="m162.67192 135.78215l172.85039 0l0 67.90552l-172.85039 0z" fill-rule="nonzero"></path><path fill="#000000" d="m172.50005 153.06903l-0.734375 0l0 -1.1875l0.96875 0l0.328125 -1.59375l-1.296875 0l0 -1.171875l1.53125 0l0.421875 -2.109375l1.203125 0l-0.421875 2.109375l1.1875 0l0.421875 -2.109375l1.234375 0l-0.421875 2.109375l0.734375 0l0 1.171875l-0.984375 0l-0.3125 1.59375l1.296875 0l0 1.1875l-1.546875 0l-0.421875 2.09375l-1.203125 0l0.421875 -2.09375l-1.1875 0l-0.4375 2.09375l-1.203125 0l0.421875 -2.09375zm1.765625 -2.78125l-0.3125 1.59375l1.1875 0l0.3125 -1.59375l-1.1875 0zm7.090271 -3.140625l2.90625 0q0.984375 0 1.5 0.15625q0.6875 0.203125 1.171875 0.71875q0.5 0.515625 0.75 1.28125q0.265625 0.75 0.265625 1.859375q0 0.96875 -0.234375 1.671875q-0.296875 0.859375 -0.84375 1.390625q-0.421875 0.40625 -1.125 0.625q-0.515625 0.171875 -1.40625 0.171875l-2.984375 0l0 -7.875zm1.59375 1.328125l0 5.21875l1.1875 0q0.65625 0 0.953125 -0.078125q0.390625 -0.09375 0.640625 -0.3125q0.265625 -0.234375 0.421875 -0.765625q0.15625 -0.53125 0.15625 -1.453125q0 -0.90625 -0.15625 -1.390625q-0.15625 -0.5 -0.453125 -0.765625q-0.296875 -0.28125 -0.734375 -0.375q-0.328125 -0.078125 -1.3125 -0.078125l-0.703125 0zm5.7424774 3.609375q0 -0.75 0.375 -1.453125q0.375 -0.703125 1.046875 -1.0625q0.6875 -0.375 1.515625 -0.375q1.296875 0 2.125 0.84375q0.828125 0.828125 0.828125 2.109375q0 1.296875 -0.84375 2.15625q-0.828125 0.84375 -2.09375 0.84375q-0.78125 0 -1.5 -0.34375q-0.703125 -0.359375 -1.078125 -1.046875q-0.375 -0.6875 -0.375 -1.671875zm1.546875 0.078125q0 0.859375 0.40625 1.3125q0.40625 0.453125 1.0 0.453125q0.578125 0 0.984375 -0.453125q0.40625 -0.453125 0.40625 -1.3125q0 -0.84375 -0.40625 -1.296875q-0.40625 -0.453125 -0.984375 -0.453125q-0.59375 0 -1.0 0.453125q-0.40625 0.453125 -0.40625 1.296875zm4.587387 2.859375l2.0625 -2.9375l-1.96875 -2.765625l1.84375 0l1.0 1.5625l1.0625 -1.5625l1.78125 0l-1.9375 2.703125l2.109375 3.0l-1.859375 0l-1.15625 -1.765625l-1.171875 1.765625l-1.765625 0zm5.9421234 -5.703125l1.609375 0l1.359375 4.046875l1.328125 -4.046875l1.5625 0l-2.015625 5.484375l-0.359375 1.0q-0.1875 0.5 -0.375 0.765625q-0.171875 0.265625 -0.40625 0.421875q-0.234375 0.15625 -0.578125 0.25q-0.34375 0.09375 -0.78125 0.09375q-0.421875 0 -0.84375 -0.09375l-0.140625 -1.171875q0.359375 0.0625 0.640625 0.0625q0.53125 0 0.78125 -0.3125q0.25 -0.296875 0.390625 -0.78125l-2.171875 -5.71875zm5.9733734 0l0.84375 0l0 -0.4375q0 -0.71875 0.140625 -1.0625q0.15625 -0.359375 0.5625 -0.578125q0.421875 -0.234375 1.046875 -0.234375q0.65625 0 1.265625 0.203125l-0.203125 1.046875q-0.359375 -0.078125 -0.6875 -0.078125q-0.328125 0 -0.484375 0.15625q-0.140625 0.140625 -0.140625 0.578125l0 0.40625l1.140625 0l0 1.1875l-1.140625 0l0 4.515625l-1.5 0l0 -4.515625l-0.84375 0l0 -1.1875zm4.2204895 -0.78125l0 -1.390625l1.5 0l0 1.390625l-1.5 0zm0 6.484375l0 -5.703125l1.5 0l0 5.703125l-1.5 0zm2.9606476 0l0 -7.875l1.5 0l0 7.875l-1.5 0zm6.2575226 -1.8125l1.5 0.25q-0.28125 0.828125 -0.90625 1.265625q-0.625 0.421875 -1.578125 0.421875q-1.484375 0 -2.203125 -0.96875q-0.5625 -0.78125 -0.5625 -1.96875q0 -1.421875 0.734375 -2.21875q0.75 -0.796875 1.890625 -0.796875q1.265625 0 2.0 0.84375q0.734375 0.828125 0.703125 2.5625l-3.78125 0q0.015625 0.671875 0.359375 1.046875q0.359375 0.375 0.875 0.375q0.359375 0 0.59375 -0.1875q0.25 -0.203125 0.375 -0.625zm0.078125 -1.53125q-0.015625 -0.65625 -0.34375 -0.984375q-0.3125 -0.34375 -0.765625 -0.34375q-0.5 0 -0.828125 0.359375q-0.3125 0.359375 -0.3125 0.96875l2.25 0zm9.043396 3.34375l-1.515625 0l0 -5.6875q-0.8125 0.765625 -1.9375 1.140625l0 -1.359375q0.59375 -0.203125 1.28125 -0.734375q0.6875 -0.546875 0.953125 -1.265625l1.21875 0l0 7.90625zm2.3952484 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm3.9293976 -4.25q-0.59375 -0.25 -0.859375 -0.671875q-0.265625 -0.4375 -0.265625 -0.953125q0 -0.875 0.609375 -1.453125q0.625 -0.578125 1.765625 -0.578125q1.109375 0 1.734375 0.578125q0.625 0.578125 0.625 1.453125q0 0.546875 -0.296875 0.984375q-0.28125 0.421875 -0.796875 0.640625q0.65625 0.265625 1.0 0.78125q0.34375 0.5 0.34375 1.15625q0 1.09375 -0.703125 1.78125q-0.6875 0.671875 -1.84375 0.671875q-1.078125 0 -1.78125 -0.5625q-0.84375 -0.671875 -0.84375 -1.828125q0 -0.640625 0.3125 -1.171875q0.3125 -0.53125 1.0 -0.828125zm0.3125 -1.515625q0 0.453125 0.25 0.703125q0.25 0.25 0.6875 0.25q0.421875 0 0.671875 -0.25q0.265625 -0.265625 0.265625 -0.71875q0 -0.421875 -0.265625 -0.671875q-0.25 -0.25 -0.65625 -0.25q-0.4375 0 -0.703125 0.265625q-0.25 0.25 -0.25 0.671875zm-0.140625 3.375q0 0.625 0.3125 0.96875q0.328125 0.34375 0.796875 0.34375q0.46875 0 0.765625 -0.328125q0.3125 -0.34375 0.3125 -0.96875q0 -0.5625 -0.3125 -0.890625q-0.3125 -0.34375 -0.78125 -0.34375q-0.5625 0 -0.828125 0.390625q-0.265625 0.375 -0.265625 0.828125zm4.7858734 2.390625l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm7.7418976 -5.953125l-1.453125 0.171875q-0.0625 -0.453125 -0.296875 -0.671875q-0.21875 -0.21875 -0.578125 -0.21875q-0.46875 0 -0.8125 0.4375q-0.328125 0.421875 -0.40625 1.78125q0.5625 -0.65625 1.40625 -0.65625q0.9375 0 1.609375 0.71875q0.671875 0.71875 0.671875 1.859375q0 1.203125 -0.703125 1.9375q-0.703125 0.734375 -1.8125 0.734375q-1.203125 0 -1.96875 -0.921875q-0.765625 -0.9375 -0.765625 -3.046875q0 -2.171875 0.796875 -3.125q0.796875 -0.953125 2.078125 -0.953125q0.90625 0 1.484375 0.5q0.59375 0.5 0.75 1.453125zm-3.421875 3.296875q0 0.734375 0.34375 1.140625q0.34375 0.390625 0.765625 0.390625q0.421875 0 0.703125 -0.3125q0.28125 -0.328125 0.28125 -1.078125q0 -0.765625 -0.3125 -1.125q-0.296875 -0.359375 -0.734375 -0.359375q-0.4375 0 -0.75 0.34375q-0.296875 0.34375 -0.296875 1.0z" fill-rule="nonzero"></path><path fill="#000000" d="m177.51567 178.13153l1.53125 0.484375q-0.34375 1.28125 -1.171875 1.921875q-0.828125 0.625 -2.09375 0.625q-1.5625 0 -2.578125 -1.078125q-1.015625 -1.078125 -1.015625 -2.9375q0 -1.953125 1.015625 -3.046875q1.015625 -1.09375 2.671875 -1.09375q1.453125 0 2.34375 0.859375q0.546875 0.5 0.8125 1.453125l-1.578125 0.375q-0.140625 -0.609375 -0.578125 -0.96875q-0.4375 -0.359375 -1.078125 -0.359375q-0.875 0 -1.421875 0.640625q-0.546875 0.625 -0.546875 2.03125q0 1.484375 0.53125 2.125q0.546875 0.640625 1.40625 0.640625q0.640625 0 1.09375 -0.40625q0.453125 -0.40625 0.65625 -1.265625zm2.6643524 2.890625l0 -7.875l3.34375 0q1.265625 0 1.828125 0.21875q0.578125 0.203125 0.921875 0.75q0.34375 0.546875 0.34375 1.234375q0 0.890625 -0.53125 1.46875q-0.515625 0.578125 -1.546875 0.734375q0.515625 0.296875 0.84375 0.65625q0.34375 0.359375 0.90625 1.28125l0.96875 1.53125l-1.90625 0l-1.15625 -1.71875q-0.609375 -0.90625 -0.84375 -1.140625q-0.21875 -0.25 -0.46875 -0.328125q-0.25 -0.09375 -0.796875 -0.09375l-0.328125 0l0 3.28125l-1.578125 0zm1.578125 -4.546875l1.1875 0q1.140625 0 1.421875 -0.09375q0.28125 -0.09375 0.4375 -0.328125q0.171875 -0.234375 0.171875 -0.59375q0 -0.40625 -0.21875 -0.640625q-0.203125 -0.25 -0.59375 -0.3125q-0.1875 -0.03125 -1.15625 -0.03125l-1.25 0l0 2.0zm6.1018524 4.546875l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm14.2170105 0l-1.734375 0l-0.6875 -1.78125l-3.140625 0l-0.65625 1.78125l-1.6875 0l3.0625 -7.875l1.6875 0l3.15625 7.875zm-2.9375 -3.109375l-1.078125 -2.921875l-1.0625 2.921875l2.140625 0zm5.3049774 3.109375l0 -6.546875l-2.34375 0l0 -1.328125l6.265625 0l0 1.328125l-2.34375 0l0 6.546875l-1.578125 0zm4.728012 0l0 -7.875l5.84375 0l0 1.328125l-4.25 0l0 1.75l3.953125 0l0 1.328125l-3.953125 0l0 2.140625l4.40625 0l0 1.328125l-6.0 0zm6.2013855 2.171875l0 -0.96875l6.28125 0l0 0.96875l-6.28125 0zm6.4264984 -4.734375l1.546875 -0.15625q0.140625 0.78125 0.5625 1.15625q0.4375 0.359375 1.15625 0.359375q0.78125 0 1.15625 -0.328125q0.390625 -0.328125 0.390625 -0.75q0 -0.28125 -0.171875 -0.46875q-0.15625 -0.203125 -0.5625 -0.359375q-0.28125 -0.09375 -1.265625 -0.34375q-1.28125 -0.3125 -1.796875 -0.78125q-0.71875 -0.640625 -0.71875 -1.578125q0 -0.59375 0.328125 -1.109375q0.34375 -0.53125 0.984375 -0.8125q0.640625 -0.28125 1.5625 -0.28125q1.46875 0 2.21875 0.65625q0.75 0.65625 0.796875 1.734375l-1.59375 0.078125q-0.109375 -0.609375 -0.4375 -0.875q-0.328125 -0.265625 -1.0 -0.265625q-0.703125 0 -1.09375 0.28125q-0.25 0.1875 -0.25 0.484375q0 0.28125 0.234375 0.484375q0.296875 0.25 1.453125 0.53125q1.171875 0.265625 1.71875 0.5625q0.5625 0.28125 0.875 0.796875q0.3125 0.5 0.3125 1.25q0 0.6875 -0.375 1.28125q-0.375 0.578125 -1.0625 0.875q-0.6875 0.28125 -1.71875 0.28125q-1.484375 0 -2.296875 -0.6875q-0.796875 -0.6875 -0.953125 -2.015625zm7.5138855 -5.3125l1.578125 0l0 4.265625q0 1.015625 0.0625 1.3125q0.109375 0.484375 0.484375 0.78125q0.390625 0.296875 1.046875 0.296875q0.6875 0 1.03125 -0.28125q0.34375 -0.28125 0.40625 -0.671875q0.078125 -0.40625 0.078125 -1.34375l0 -4.359375l1.578125 0l0 4.140625q0 1.40625 -0.125 2.0q-0.125 0.578125 -0.46875 0.984375q-0.34375 0.40625 -0.9375 0.65625q-0.578125 0.234375 -1.515625 0.234375q-1.125 0 -1.703125 -0.265625q-0.578125 -0.265625 -0.921875 -0.671875q-0.34375 -0.421875 -0.4375 -0.875q-0.15625 -0.6875 -0.15625 -2.0l0 -4.203125zm7.7112274 0l3.140625 0q0.9375 0 1.390625 0.078125q0.46875 0.078125 0.828125 0.328125q0.359375 0.25 0.59375 0.65625q0.25 0.40625 0.25 0.921875q0 0.546875 -0.296875 1.015625q-0.296875 0.453125 -0.8125 0.6875q0.71875 0.21875 1.109375 0.71875q0.390625 0.5 0.390625 1.1875q0 0.53125 -0.25 1.046875q-0.25 0.5 -0.6875 0.8125q-0.4375 0.296875 -1.0625 0.375q-0.40625 0.03125 -1.921875 0.046875l-2.671875 0l0 -7.875zm1.578125 1.3125l0 1.8125l1.046875 0q0.921875 0 1.15625 -0.015625q0.40625 -0.046875 0.640625 -0.28125q0.234375 -0.234375 0.234375 -0.625q0 -0.359375 -0.203125 -0.578125q-0.203125 -0.234375 -0.59375 -0.28125q-0.234375 -0.03125 -1.359375 -0.03125l-0.921875 0zm0 3.125l0 2.109375l1.484375 0q0.859375 0 1.078125 -0.046875q0.359375 -0.0625 0.578125 -0.3125q0.234375 -0.25 0.234375 -0.671875q0 -0.359375 -0.171875 -0.59375q-0.171875 -0.25 -0.5 -0.359375q-0.328125 -0.125 -1.421875 -0.125l-1.28125 0zm6.1018524 -4.4375l2.90625 0q0.984375 0 1.5 0.15625q0.6875 0.203125 1.171875 0.71875q0.5 0.515625 0.75 1.28125q0.265625 0.75 0.265625 1.859375q0 0.96875 -0.234375 1.671875q-0.296875 0.859375 -0.84375 1.390625q-0.421875 0.40625 -1.125 0.625q-0.515625 0.171875 -1.40625 0.171875l-2.984375 0l0 -7.875zm1.59375 1.328125l0 5.21875l1.1875 0q0.65625 0 0.953125 -0.078125q0.390625 -0.09375 0.640625 -0.3125q0.265625 -0.234375 0.421875 -0.765625q0.15625 -0.53125 0.15625 -1.453125q0 -0.90625 -0.15625 -1.390625q-0.15625 -0.5 -0.453125 -0.765625q-0.296875 -0.28125 -0.734375 -0.375q-0.328125 -0.078125 -1.3125 -0.078125l-0.703125 0zm6.0549774 6.546875l0 -7.875l1.59375 0l0 7.875l-1.59375 0zm3.0231476 0l0 -7.875l3.34375 0q1.265625 0 1.828125 0.21875q0.578125 0.203125 0.921875 0.75q0.34375 0.546875 0.34375 1.234375q0 0.890625 -0.53125 1.46875q-0.515625 0.578125 -1.546875 0.734375q0.515625 0.296875 0.84375 0.65625q0.34375 0.359375 0.90625 1.28125l0.96875 1.53125l-1.90625 0l-1.15625 -1.71875q-0.609375 -0.90625 -0.84375 -1.140625q-0.21875 -0.25 -0.46875 -0.328125q-0.25 -0.09375 -0.796875 -0.09375l-0.328125 0l0 3.28125l-1.578125 0zm1.578125 -4.546875l1.1875 0q1.140625 0 1.421875 -0.09375q0.28125 -0.09375 0.4375 -0.328125q0.171875 -0.234375 0.171875 -0.59375q0 -0.40625 -0.21875 -0.640625q-0.203125 -0.25 -0.59375 -0.3125q-0.1875 -0.03125 -1.15625 -0.03125l-1.25 0l0 2.0zm5.695587 1.984375l1.546875 -0.15625q0.140625 0.78125 0.5625 1.15625q0.4375 0.359375 1.15625 0.359375q0.78125 0 1.15625 -0.328125q0.390625 -0.328125 0.390625 -0.75q0 -0.28125 -0.171875 -0.46875q-0.15625 -0.203125 -0.5625 -0.359375q-0.28125 -0.09375 -1.265625 -0.34375q-1.28125 -0.3125 -1.796875 -0.78125q-0.71875 -0.640625 -0.71875 -1.578125q0 -0.59375 0.328125 -1.109375q0.34375 -0.53125 0.984375 -0.8125q0.640625 -0.28125 1.5625 -0.28125q1.46875 0 2.21875 0.65625q0.75 0.65625 0.796875 1.734375l-1.59375 0.078125q-0.109375 -0.609375 -0.4375 -0.875q-0.328125 -0.265625 -1.0 -0.265625q-0.703125 0 -1.09375 0.28125q-0.25 0.1875 -0.25 0.484375q0 0.28125 0.234375 0.484375q0.296875 0.25 1.453125 0.53125q1.171875 0.265625 1.71875 0.5625q0.5625 0.28125 0.875 0.796875q0.3125 0.5 0.3125 1.25q0 0.6875 -0.375 1.28125q-0.375 0.578125 -1.0625 0.875q-0.6875 0.28125 -1.71875 0.28125q-1.484375 0 -2.296875 -0.6875q-0.796875 -0.6875 -0.953125 -2.015625zm16.052094 -1.828125l0 -1.375l5.515625 0l0 1.375l-5.515625 0zm0 2.390625l0 -1.390625l5.515625 0l0 1.390625l-5.515625 0zm9.543091 2.0l0 -7.875l1.546875 0l3.234375 5.265625l0 -5.265625l1.46875 0l0 7.875l-1.59375 0l-3.171875 -5.140625l0 5.140625l-1.484375 0zm7.3674927 -3.890625q0 -1.203125 0.359375 -2.015625q0.265625 -0.609375 0.71875 -1.078125q0.46875 -0.484375 1.03125 -0.71875q0.734375 -0.3125 1.6875 -0.3125q1.75 0 2.78125 1.09375q1.046875 1.078125 1.046875 3.0q0 1.90625 -1.03125 2.984375q-1.03125 1.078125 -2.765625 1.078125q-1.765625 0 -2.796875 -1.078125q-1.03125 -1.078125 -1.03125 -2.953125zm1.625 -0.046875q0 1.328125 0.625 2.03125q0.625 0.6875 1.5625 0.6875q0.953125 0 1.5625 -0.6875q0.609375 -0.6875 0.609375 -2.0625q0 -1.34375 -0.59375 -2.015625q-0.59375 -0.671875 -1.578125 -0.671875q-0.984375 0 -1.59375 0.6875q-0.59375 0.671875 -0.59375 2.03125z" fill-rule="nonzero"></path><path fill="#000000" d="m172.4688 194.02216l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606476 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0zm2.9606476 0l0 -1.515625l1.5 0l0 1.515625l-1.5 0z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m260.59317 288.11548l-48.94487 0l0 -24.472443" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m260.59317 288.11548l-48.94487 0l0 -24.472443" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m211.63255 239.17061l0 -17.741653l37.4803 0l0 -17.75441" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m211.63255 239.17061l0 -17.741653l37.4803 0l0 -17.75441" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m384.25723 413.64566l-33.044678 0l0 20.944885l-33.03409 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.25723 413.64566l-33.044678 0l0 20.944885l-33.03409 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m384.25723 413.64566l-33.044678 0l0 52.944885l-33.03409 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.25723 413.64566l-33.044678 0l0 52.944885l-33.03409 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m384.25723 413.64566l-33.044678 0l0 84.944885l-33.03409 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.25723 413.64566l-33.044678 0l0 84.944885l-33.03409 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m384.25723 413.64566l-33.044678 0l0 116.944916l-33.03409 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.25723 413.64566l-33.044678 0l0 116.944916l-33.03409 0" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m384.25723 413.64566l-33.044678 0l0 148.94492l-33.03409 0" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m384.25723 413.64566l-33.044678 0l0 148.94492l-33.03409 0" fill-rule="nonzero"></path><path fill="#fff2cc" d="m275.57086 83.26903l398.27557 0l0 26.622047l-398.27557 0z" fill-rule="nonzero"></path><path fill="#fff4d6" d="m271.13385 78.83202l407.1496 0l-4.4370117 4.437004l-398.27557 0z" fill-rule="nonzero"></path><path fill="#cbc1a3" d="m271.13385 114.32809l4.4370117 -4.4370117l398.27557 0l4.4370117 4.4370117z" fill-rule="nonzero"></path><path fill="#fff7e0" d="m271.13385 78.83202l4.4370117 4.437004l0 26.622047l-4.4370117 4.4370117z" fill-rule="nonzero"></path><path fill="#98917a" d="m678.28345 78.83202l0 35.496063l-4.4370117 -4.4370117l0 -26.622047z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m271.13385 78.83202l407.1496 0l0 35.496063l-407.1496 0zm4.4370117 4.437004l398.27557 0l0 26.622047l-398.27557 0zm-4.4370117 -4.437004l4.4370117 4.437004m-4.4370117 31.05906l4.4370117 -4.4370117m402.7126 -31.059052l-4.4370117 4.437004m4.4370117 31.05906l-4.4370117 -4.4370117" fill-rule="nonzero"></path><path stroke="#000000" stroke-width="2.0" stroke-linejoin="round" stroke-linecap="butt" d="m271.13385 78.83202l407.1496 0l0 35.496063l-407.1496 0zm4.4370117 4.437004l398.27557 0l0 26.622047l-398.27557 0zm-4.4370117 -4.437004l4.4370117 4.437004m-4.4370117 31.05906l4.4370117 -4.4370117m402.7126 -31.059052l-4.4370117 4.437004m4.4370117 31.05906l-4.4370117 -4.4370117" fill-rule="nonzero"></path><path fill="#000000" d="m362.937 100.32818l1.71875 0.21875q-0.40625 1.5 -1.515625 2.34375q-1.09375 0.828125 -2.8125 0.828125q-2.15625 0 -3.421875 -1.328125q-1.265625 -1.328125 -1.265625 -3.734375q0 -2.484375 1.265625 -3.859375q1.28125 -1.375 3.328125 -1.375q1.984375 0 3.234375 1.34375q1.25 1.34375 1.25 3.796875q0 0.140625 -0.015625 0.4375l-7.34375 0q0.09375 1.625 0.921875 2.484375q0.828125 0.859375 2.0625 0.859375q0.90625 0 1.546875 -0.46875q0.65625 -0.484375 1.046875 -1.546875zm-5.484375 -2.703125l5.5 0q-0.109375 -1.234375 -0.625 -1.859375q-0.796875 -0.96875 -2.078125 -0.96875q-1.140625 0 -1.9375 0.78125q-0.78125 0.765625 -0.859375 2.046875zm10.656952 5.875l-1.546875 0l0 -13.59375l1.65625 0l0 4.84375q1.0625 -1.328125 2.703125 -1.328125q0.90625 0 1.71875 0.375q0.8125 0.359375 1.328125 1.03125q0.53125 0.65625 0.828125 1.59375q0.296875 0.9375 0.296875 2.0q0 2.53125 -1.25 3.921875q-1.25 1.375 -3.0 1.375q-1.75 0 -2.734375 -1.453125l0 1.234375zm-0.015625 -5.0q0 1.765625 0.46875 2.5625q0.796875 1.28125 2.140625 1.28125q1.09375 0 1.890625 -0.9375q0.796875 -0.953125 0.796875 -2.84375q0 -1.921875 -0.765625 -2.84375q-0.765625 -0.921875 -1.84375 -0.921875q-1.09375 0 -1.890625 0.953125q-0.796875 0.953125 -0.796875 2.75zm8.860107 -6.6875l0 -1.90625l1.671875 0l0 1.90625l-1.671875 0zm0 11.6875l0 -9.859375l1.671875 0l0 9.859375l-1.671875 0zm3.457306 -2.9375l1.65625 -0.265625q0.140625 1.0 0.765625 1.53125q0.640625 0.515625 1.78125 0.515625q1.15625 0 1.703125 -0.46875q0.5625 -0.46875 0.5625 -1.09375q0 -0.5625 -0.484375 -0.890625q-0.34375 -0.21875 -1.703125 -0.5625q-1.84375 -0.46875 -2.5625 -0.796875q-0.703125 -0.34375 -1.078125 -0.9375q-0.359375 -0.609375 -0.359375 -1.328125q0 -0.65625 0.296875 -1.21875q0.3125 -0.5625 0.828125 -0.9375q0.390625 -0.28125 1.0625 -0.484375q0.671875 -0.203125 1.4375 -0.203125q1.171875 0 2.046875 0.34375q0.875 0.328125 1.28125 0.90625q0.421875 0.5625 0.578125 1.515625l-1.625 0.21875q-0.109375 -0.75 -0.65625 -1.171875q-0.53125 -0.4375 -1.5 -0.4375q-1.15625 0 -1.640625 0.390625q-0.484375 0.375 -0.484375 0.875q0 0.328125 0.203125 0.59375q0.203125 0.265625 0.640625 0.4375q0.25 0.09375 1.46875 0.4375q1.765625 0.46875 2.46875 0.765625q0.703125 0.296875 1.09375 0.875q0.40625 0.578125 0.40625 1.4375q0 0.828125 -0.484375 1.578125q-0.484375 0.734375 -1.40625 1.140625q-0.921875 0.390625 -2.078125 0.390625q-1.921875 0 -2.9375 -0.796875q-1.0 -0.796875 -1.28125 -2.359375zm16.453125 2.9375l0 -1.453125q-1.140625 1.671875 -3.125 1.671875q-0.859375 0 -1.625 -0.328125q-0.75 -0.34375 -1.125 -0.84375q-0.359375 -0.5 -0.515625 -1.234375q-0.09375 -0.5 -0.09375 -1.5625l0 -6.109375l1.671875 0l0 5.46875q0 1.3125 0.09375 1.765625q0.15625 0.65625 0.671875 1.03125q0.515625 0.375 1.265625 0.375q0.75 0 1.40625 -0.375q0.65625 -0.390625 0.921875 -1.046875q0.28125 -0.671875 0.28125 -1.9375l0 -5.28125l1.671875 0l0 9.859375l-1.5 0zm2.3913574 3.78125l0 -1.21875l11.0625 0l0 1.21875l-11.0625 0zm18.344452 -7.390625l1.640625 0.21875q-0.265625 1.6875 -1.375 2.65625q-1.109375 0.953125 -2.734375 0.953125q-2.015625 0 -3.25 -1.3125q-1.21875 -1.328125 -1.21875 -3.796875q0 -1.59375 0.515625 -2.78125q0.53125 -1.203125 1.609375 -1.796875q1.09375 -0.609375 2.359375 -0.609375q1.609375 0 2.625 0.8125q1.015625 0.8125 1.3125 2.3125l-1.625 0.25q-0.234375 -1.0 -0.828125 -1.5q-0.59375 -0.5 -1.421875 -0.5q-1.265625 0 -2.0625 0.90625q-0.78125 0.90625 -0.78125 2.859375q0 1.984375 0.765625 2.890625q0.765625 0.890625 1.984375 0.890625q0.984375 0 1.640625 -0.59375q0.65625 -0.609375 0.84375 -1.859375zm2.890625 7.390625l0 -13.640625l1.53125 0l0 1.28125q0.53125 -0.75 1.203125 -1.125q0.6875 -0.375 1.640625 -0.375q1.265625 0 2.234375 0.65625q0.96875 0.640625 1.453125 1.828125q0.5 1.1875 0.5 2.59375q0 1.515625 -0.546875 2.734375q-0.546875 1.203125 -1.578125 1.84375q-1.03125 0.640625 -2.171875 0.640625q-0.84375 0 -1.515625 -0.34375q-0.65625 -0.359375 -1.078125 -0.890625l0 4.796875l-1.671875 0zm1.515625 -8.65625q0 1.90625 0.765625 2.8125q0.78125 0.90625 1.875 0.90625q1.109375 0 1.890625 -0.9375q0.796875 -0.9375 0.796875 -2.921875q0 -1.875 -0.78125 -2.8125q-0.765625 -0.9375 -1.84375 -0.9375q-1.0625 0 -1.890625 1.0q-0.8125 1.0 -0.8125 2.890625zm8.860107 8.65625l0 -13.640625l1.53125 0l0 1.28125q0.53125 -0.75 1.203125 -1.125q0.6875 -0.375 1.640625 -0.375q1.265625 0 2.234375 0.65625q0.96875 0.640625 1.453125 1.828125q0.5 1.1875 0.5 2.59375q0 1.515625 -0.546875 2.734375q-0.546875 1.203125 -1.578125 1.84375q-1.03125 0.640625 -2.171875 0.640625q-0.84375 0 -1.515625 -0.34375q-0.65625 -0.359375 -1.078125 -0.890625l0 4.796875l-1.671875 0zm1.515625 -8.65625q0 1.90625 0.765625 2.8125q0.78125 0.90625 1.875 0.90625q1.109375 0 1.890625 -0.9375q0.796875 -0.9375 0.796875 -2.921875q0 -1.875 -0.78125 -2.8125q-0.765625 -0.9375 -1.84375 -0.9375q-1.0625 0 -1.890625 1.0q-0.8125 1.0 -0.8125 2.890625zm13.371521 1.9375l1.65625 -0.265625q0.140625 1.0 0.765625 1.53125q0.640625 0.515625 1.78125 0.515625q1.15625 0 1.703125 -0.46875q0.5625 -0.46875 0.5625 -1.09375q0 -0.5625 -0.484375 -0.890625q-0.34375 -0.21875 -1.703125 -0.5625q-1.84375 -0.46875 -2.5625 -0.796875q-0.703125 -0.34375 -1.078125 -0.9375q-0.359375 -0.609375 -0.359375 -1.328125q0 -0.65625 0.296875 -1.21875q0.3125 -0.5625 0.828125 -0.9375q0.390625 -0.28125 1.0625 -0.484375q0.671875 -0.203125 1.4375 -0.203125q1.171875 0 2.046875 0.34375q0.875 0.328125 1.28125 0.90625q0.421875 0.5625 0.578125 1.515625l-1.625 0.21875q-0.109375 -0.75 -0.65625 -1.171875q-0.53125 -0.4375 -1.5 -0.4375q-1.15625 0 -1.640625 0.390625q-0.484375 0.375 -0.484375 0.875q0 0.328125 0.203125 0.59375q0.203125 0.265625 0.640625 0.4375q0.25 0.09375 1.46875 0.4375q1.765625 0.46875 2.46875 0.765625q0.703125 0.296875 1.09375 0.875q0.40625 0.578125 0.40625 1.4375q0 0.828125 -0.484375 1.578125q-0.484375 0.734375 -1.40625 1.140625q-0.921875 0.390625 -2.078125 0.390625q-1.921875 0 -2.9375 -0.796875q-1.0 -0.796875 -1.28125 -2.359375zm13.65625 1.4375l0.234375 1.484375q-0.703125 0.140625 -1.265625 0.140625q-0.90625 0 -1.40625 -0.28125q-0.5 -0.296875 -0.703125 -0.75q-0.203125 -0.46875 -0.203125 -1.984375l0 -5.65625l-1.234375 0l0 -1.3125l1.234375 0l0 -2.4375l1.65625 -1.0l0 3.4375l1.6875 0l0 1.3125l-1.6875 0l0 5.75q0 0.71875 0.078125 0.921875q0.09375 0.203125 0.296875 0.328125q0.203125 0.125 0.578125 0.125q0.265625 0 0.734375 -0.078125zm1.5114136 1.5l0 -9.859375l1.5 0l0 1.5q0.578125 -1.046875 1.0625 -1.375q0.484375 -0.34375 1.078125 -0.34375q0.84375 0 1.71875 0.546875l-0.578125 1.546875q-0.609375 -0.359375 -1.234375 -0.359375q-0.546875 0 -0.984375 0.328125q-0.421875 0.328125 -0.609375 0.90625q-0.28125 0.890625 -0.28125 1.953125l0 5.15625l-1.671875 0zm12.681427 0l0 -1.453125q-1.140625 1.671875 -3.125 1.671875q-0.859375 0 -1.625 -0.328125q-0.75 -0.34375 -1.125 -0.84375q-0.359375 -0.5 -0.515625 -1.234375q-0.09375 -0.5 -0.09375 -1.5625l0 -6.109375l1.671875 0l0 5.46875q0 1.3125 0.09375 1.765625q0.15625 0.65625 0.671875 1.03125q0.515625 0.375 1.265625 0.375q0.75 0 1.40625 -0.375q0.65625 -0.390625 0.921875 -1.046875q0.28125 -0.671875 0.28125 -1.9375l0 -5.28125l1.671875 0l0 9.859375l-1.5 0zm10.360107 -3.609375l1.640625 0.21875q-0.265625 1.6875 -1.375 2.65625q-1.109375 0.953125 -2.734375 0.953125q-2.015625 0 -3.25 -1.3125q-1.21875 -1.328125 -1.21875 -3.796875q0 -1.59375 0.515625 -2.78125q0.53125 -1.203125 1.609375 -1.796875q1.09375 -0.609375 2.359375 -0.609375q1.609375 0 2.625 0.8125q1.015625 0.8125 1.3125 2.3125l-1.625 0.25q-0.234375 -1.0 -0.828125 -1.5q-0.59375 -0.5 -1.421875 -0.5q-1.265625 0 -2.0625 0.90625q-0.78125 0.90625 -0.78125 2.859375q0 1.984375 0.765625 2.890625q0.765625 0.890625 1.984375 0.890625q0.984375 0 1.640625 -0.59375q0.65625 -0.609375 0.84375 -1.859375zm6.546875 2.109375l0.234375 1.484375q-0.703125 0.140625 -1.265625 0.140625q-0.90625 0 -1.40625 -0.28125q-0.5 -0.296875 -0.703125 -0.75q-0.203125 -0.46875 -0.203125 -1.984375l0 -5.65625l-1.234375 0l0 -1.3125l1.234375 0l0 -2.4375l1.65625 -1.0l0 3.4375l1.6875 0l0 1.3125l-1.6875 0l0 5.75q0 0.71875 0.078125 0.921875q0.09375 0.203125 0.296875 0.328125q0.203125 0.125 0.578125 0.125q0.265625 0 0.734375 -0.078125zm7.9801636 1.5l0 -1.453125q-1.140625 1.671875 -3.125 1.671875q-0.859375 0 -1.625 -0.328125q-0.75 -0.34375 -1.125 -0.84375q-0.359375 -0.5 -0.515625 -1.234375q-0.09375 -0.5 -0.09375 -1.5625l0 -6.109375l1.671875 0l0 5.46875q0 1.3125 0.09375 1.765625q0.15625 0.65625 0.671875 1.03125q0.515625 0.375 1.265625 0.375q0.75 0 1.40625 -0.375q0.65625 -0.390625 0.921875 -1.046875q0.28125 -0.671875 0.28125 -1.9375l0 -5.28125l1.671875 0l0 9.859375l-1.5 0zm3.9069824 0l0 -9.859375l1.5 0l0 1.5q0.578125 -1.046875 1.0625 -1.375q0.484375 -0.34375 1.078125 -0.34375q0.84375 0 1.71875 0.546875l-0.578125 1.546875q-0.609375 -0.359375 -1.234375 -0.359375q-0.546875 0 -0.984375 0.328125q-0.421875 0.328125 -0.609375 0.90625q-0.28125 0.890625 -0.28125 1.953125l0 5.15625l-1.671875 0zm12.6657715 -1.21875q-0.9375 0.796875 -1.796875 1.125q-0.859375 0.3125 -1.8437195 0.3125q-1.609375 0 -2.484375 -0.78125q-0.875 -0.796875 -0.875 -2.03125q0 -0.734375 0.328125 -1.328125q0.328125 -0.59375 0.859375 -0.953125q0.53125 -0.359375 1.203125 -0.546875q0.5 -0.140625 1.484375 -0.25q2.0312195 -0.25 2.9843445 -0.578125q0 -0.34375 0 -0.4375q0 -1.015625 -0.46875 -1.4375q-0.640625 -0.5625 -1.90625 -0.5625q-1.1718445 0 -1.7343445 0.40625q-0.5625 0.40625 -0.828125 1.46875l-1.640625 -0.234375q0.234375 -1.046875 0.734375 -1.6875q0.515625 -0.640625 1.46875 -0.984375q0.96875 -0.359375 2.2499695 -0.359375q1.265625 0 2.046875 0.296875q0.78125 0.296875 1.15625 0.75q0.375 0.453125 0.515625 1.140625q0.09375 0.421875 0.09375 1.53125l0 2.234375q0 2.328125 0.09375 2.953125q0.109375 0.609375 0.4375 1.171875l-1.75 0q-0.265625 -0.515625 -0.328125 -1.21875zm-0.140625 -3.71875q-0.90625 0.359375 -2.734375 0.625q-1.0312195 0.140625 -1.4530945 0.328125q-0.421875 0.1875 -0.65625 0.546875q-0.234375 0.359375 -0.234375 0.796875q0 0.671875 0.5 1.125q0.515625 0.4375 1.484375 0.4375q0.9687195 0 1.7187195 -0.421875q0.75 -0.4375 1.109375 -1.15625q0.265625 -0.578125 0.265625 -1.671875l0 -0.609375zm4.0476074 4.9375l0 -13.59375l1.671875 0l0 13.59375l-1.671875 0zm15.75 0l0 -1.25q-0.9375 1.46875 -2.75 1.46875q-1.171875 0 -2.171875 -0.640625q-0.984375 -0.65625 -1.53125 -1.8125q-0.53125 -1.171875 -0.53125 -2.6875q0 -1.46875 0.484375 -2.671875q0.5 -1.203125 1.46875 -1.84375q0.984375 -0.640625 2.203125 -0.640625q0.890625 0 1.578125 0.375q0.703125 0.375 1.140625 0.984375l0 -4.875l1.65625 0l0 13.59375l-1.546875 0zm-5.28125 -4.921875q0 1.890625 0.796875 2.828125q0.8125 0.9375 1.890625 0.9375q1.09375 0 1.859375 -0.890625q0.765625 -0.890625 0.765625 -2.734375q0 -2.015625 -0.78125 -2.953125q-0.78125 -0.953125 -1.921875 -0.953125q-1.109375 0 -1.859375 0.90625q-0.75 0.90625 -0.75 2.859375zm9.281982 -6.765625l0 -1.90625l1.671875 0l0 1.90625l-1.671875 0zm0 11.6875l0 -9.859375l1.671875 0l0 9.859375l-1.671875 0zm10.566711 -1.21875q-0.9375 0.796875 -1.796875 1.125q-0.859375 0.3125 -1.84375 0.3125q-1.609375 0 -2.484375 -0.78125q-0.875 -0.796875 -0.875 -2.03125q0 -0.734375 0.328125 -1.328125q0.328125 -0.59375 0.859375 -0.953125q0.53125 -0.359375 1.203125 -0.546875q0.5 -0.140625 1.484375 -0.25q2.03125 -0.25 2.984375 -0.578125q0 -0.34375 0 -0.4375q0 -1.015625 -0.46875 -1.4375q-0.640625 -0.5625 -1.90625 -0.5625q-1.171875 0 -1.734375 0.40625q-0.5625 0.40625 -0.828125 1.46875l-1.640625 -0.234375q0.234375 -1.046875 0.734375 -1.6875q0.515625 -0.640625 1.46875 -0.984375q0.96875 -0.359375 2.25 -0.359375q1.265625 0 2.046875 0.296875q0.78125 0.296875 1.15625 0.75q0.375 0.453125 0.515625 1.140625q0.09375 0.421875 0.09375 1.53125l0 2.234375q0 2.328125 0.09375 2.953125q0.109375 0.609375 0.4375 1.171875l-1.75 0q-0.265625 -0.515625 -0.328125 -1.21875zm-0.140625 -3.71875q-0.90625 0.359375 -2.734375 0.625q-1.03125 0.140625 -1.453125 0.328125q-0.421875 0.1875 -0.65625 0.546875q-0.234375 0.359375 -0.234375 0.796875q0 0.671875 0.5 1.125q0.515625 0.4375 1.484375 0.4375q0.96875 0 1.71875 -0.421875q0.75 -0.4375 1.109375 -1.15625q0.265625 -0.578125 0.265625 -1.671875l0 -0.609375zm3.7819214 5.75l1.609375 0.25q0.109375 0.75 0.578125 1.09375q0.609375 0.453125 1.6875 0.453125q1.171875 0 1.796875 -0.46875q0.625 -0.453125 0.859375 -1.28125q0.125 -0.515625 0.109375 -2.15625q-1.09375 1.296875 -2.71875 1.296875q-2.03125 0 -3.15625 -1.46875q-1.109375 -1.46875 -1.109375 -3.515625q0 -1.40625 0.515625 -2.59375q0.515625 -1.203125 1.484375 -1.84375q0.96875 -0.65625 2.265625 -0.65625q1.75 0 2.875 1.40625l0 -1.1875l1.546875 0l0 8.515625q0 2.3125 -0.46875 3.265625q-0.46875 0.96875 -1.484375 1.515625q-1.015625 0.5625 -2.5 0.5625q-1.765625 0 -2.859375 -0.796875q-1.078125 -0.796875 -1.03125 -2.390625zm1.375 -5.921875q0 1.953125 0.765625 2.84375q0.78125 0.890625 1.9375 0.890625q1.140625 0 1.921875 -0.890625q0.78125 -0.890625 0.78125 -2.78125q0 -1.8125 -0.8125 -2.71875q-0.796875 -0.921875 -1.921875 -0.921875q-1.109375 0 -1.890625 0.90625q-0.78125 0.890625 -0.78125 2.671875zm9.281982 5.109375l0 -9.859375l1.5 0l0 1.5q0.578125 -1.046875 1.0625 -1.375q0.484375 -0.34375 1.078125 -0.34375q0.84375 0 1.71875 0.546875l-0.578125 1.546875q-0.609375 -0.359375 -1.234375 -0.359375q-0.546875 0 -0.984375 0.328125q-0.421875 0.328125 -0.609375 0.90625q-0.28125 0.890625 -0.28125 1.953125l0 5.15625l-1.671875 0zm12.6658325 -1.21875q-0.9375 0.796875 -1.796875 1.125q-0.859375 0.3125 -1.84375 0.3125q-1.609375 0 -2.484375 -0.78125q-0.875 -0.796875 -0.875 -2.03125q0 -0.734375 0.328125 -1.328125q0.328125 -0.59375 0.859375 -0.953125q0.53125 -0.359375 1.203125 -0.546875q0.5 -0.140625 1.484375 -0.25q2.03125 -0.25 2.984375 -0.578125q0 -0.34375 0 -0.4375q0 -1.015625 -0.46875 -1.4375q-0.640625 -0.5625 -1.90625 -0.5625q-1.171875 0 -1.734375 0.40625q-0.5625 0.40625 -0.828125 1.46875l-1.640625 -0.234375q0.234375 -1.046875 0.734375 -1.6875q0.515625 -0.640625 1.46875 -0.984375q0.96875 -0.359375 2.25 -0.359375q1.265625 0 2.046875 0.296875q0.78125 0.296875 1.15625 0.75q0.375 0.453125 0.515625 1.140625q0.09375 0.421875 0.09375 1.53125l0 2.234375q0 2.328125 0.09375 2.953125q0.109375 0.609375 0.4375 1.171875l-1.75 0q-0.265625 -0.515625 -0.328125 -1.21875zm-0.140625 -3.71875q-0.90625 0.359375 -2.734375 0.625q-1.03125 0.140625 -1.453125 0.328125q-0.421875 0.1875 -0.65625 0.546875q-0.234375 0.359375 -0.234375 0.796875q0 0.671875 0.5 1.125q0.515625 0.4375 1.484375 0.4375q0.96875 0 1.71875 -0.421875q0.75 -0.4375 1.109375 -1.15625q0.265625 -0.578125 0.265625 -1.671875l0 -0.609375zm4.0787964 4.9375l0 -9.859375l1.5 0l0 1.390625q0.453125 -0.71875 1.21875 -1.15625q0.78125 -0.453125 1.765625 -0.453125q1.09375 0 1.796875 0.453125q0.703125 0.453125 0.984375 1.28125q1.171875 -1.734375 3.046875 -1.734375q1.46875 0 2.25 0.8125q0.796875 0.8125 0.796875 2.5l0 6.765625l-1.671875 0l0 -6.203125q0 -1.0 -0.15625 -1.4375q-0.15625 -0.453125 -0.59375 -0.71875q-0.421875 -0.265625 -1.0 -0.265625q-1.03125 0 -1.71875 0.6875q-0.6875 0.6875 -0.6875 2.21875l0 5.71875l-1.671875 0l0 -6.40625q0 -1.109375 -0.40625 -1.65625q-0.40625 -0.5625 -1.34375 -0.5625q-0.703125 0 -1.3125 0.375q-0.59375 0.359375 -0.859375 1.078125q-0.265625 0.71875 -0.265625 2.0625l0 5.109375l-1.671875 0z" fill-rule="nonzero"></path><path fill="#000000" fill-opacity="0.0" d="m191.15749 316.61942l45.007874 0l0 24.472443l-45.007874 0z" fill-rule="nonzero"></path><path fill="#000000" d="m207.46574 334.5794l0 -0.6875q-0.546875 0.796875 -1.484375 0.796875q-0.40625 0 -0.765625 -0.15625q-0.359375 -0.171875 -0.546875 -0.40625q-0.171875 -0.234375 -0.234375 -0.578125q-0.046875 -0.234375 -0.046875 -0.75l0 -2.890625l0.78125 0l0 2.59375q0 0.625 0.0625 0.828125q0.0625 0.3125 0.296875 0.5q0.25 0.171875 0.609375 0.171875q0.359375 0 0.671875 -0.171875q0.3125 -0.1875 0.4375 -0.5q0.125 -0.3125 0.125 -0.921875l0 -2.5l0.796875 0l0 4.671875l-0.703125 0zm1.8128662 -1.390625l0.78125 -0.125q0.0625 0.46875 0.359375 0.71875q0.3125 0.25 0.84375 0.25q0.546875 0 0.8125 -0.21875q0.265625 -0.21875 0.265625 -0.515625q0 -0.28125 -0.234375 -0.421875q-0.171875 -0.109375 -0.8125 -0.28125q-0.875 -0.21875 -1.21875 -0.375q-0.328125 -0.15625 -0.5 -0.4375q-0.171875 -0.28125 -0.171875 -0.625q0 -0.3125 0.140625 -0.578125q0.140625 -0.265625 0.390625 -0.453125q0.1875 -0.125 0.5 -0.21875q0.328125 -0.09375 0.6875 -0.09375q0.546875 0 0.953125 0.15625q0.421875 0.15625 0.625 0.421875q0.203125 0.265625 0.28125 0.71875l-0.78125 0.109375q-0.046875 -0.359375 -0.3125 -0.5625q-0.25 -0.203125 -0.703125 -0.203125q-0.546875 0 -0.78125 0.1875q-0.234375 0.171875 -0.234375 0.421875q0 0.15625 0.09375 0.265625q0.09375 0.140625 0.3125 0.21875q0.109375 0.046875 0.6875 0.203125q0.84375 0.21875 1.171875 0.359375q0.34375 0.140625 0.53125 0.421875q0.1875 0.265625 0.1875 0.671875q0 0.40625 -0.234375 0.75q-0.234375 0.34375 -0.671875 0.546875q-0.421875 0.1875 -0.984375 0.1875q-0.90625 0 -1.390625 -0.375q-0.46875 -0.390625 -0.59375 -1.125zm8.1640625 -0.109375l0.828125 0.09375q-0.203125 0.71875 -0.734375 1.125q-0.515625 0.390625 -1.328125 0.390625q-1.015625 0 -1.625 -0.625q-0.59375 -0.640625 -0.59375 -1.78125q0 -1.171875 0.609375 -1.8125q0.609375 -0.65625 1.578125 -0.65625q0.9375 0 1.515625 0.640625q0.59375 0.625 0.59375 1.78125q0 0.078125 0 0.21875l-3.484375 0q0.046875 0.765625 0.4375 1.171875q0.390625 0.40625 0.984375 0.40625q0.4375 0 0.734375 -0.21875q0.3125 -0.234375 0.484375 -0.734375zm-2.59375 -1.28125l2.609375 0q-0.046875 -0.59375 -0.296875 -0.890625q-0.375 -0.453125 -0.984375 -0.453125q-0.546875 0 -0.921875 0.375q-0.359375 0.359375 -0.40625 0.96875zm4.281616 1.390625l0.78125 -0.125q0.0625 0.46875 0.359375 0.71875q0.3125 0.25 0.84375 0.25q0.546875 0 0.8125 -0.21875q0.265625 -0.21875 0.265625 -0.515625q0 -0.28125 -0.234375 -0.421875q-0.171875 -0.109375 -0.8125 -0.28125q-0.875 -0.21875 -1.21875 -0.375q-0.328125 -0.15625 -0.5 -0.4375q-0.171875 -0.28125 -0.171875 -0.625q0 -0.3125 0.140625 -0.578125q0.140625 -0.265625 0.390625 -0.453125q0.1875 -0.125 0.5 -0.21875q0.328125 -0.09375 0.6875 -0.09375q0.546875 0 0.953125 0.15625q0.421875 0.15625 0.625 0.421875q0.203125 0.265625 0.28125 0.71875l-0.78125 0.109375q-0.046875 -0.359375 -0.3125 -0.5625q-0.25 -0.203125 -0.703125 -0.203125q-0.546875 0 -0.78125 0.1875q-0.234375 0.171875 -0.234375 0.421875q0 0.15625 0.09375 0.265625q0.09375 0.140625 0.3125 0.21875q0.109375 0.046875 0.6875 0.203125q0.84375 0.21875 1.171875 0.359375q0.34375 0.140625 0.53125 0.421875q0.1875 0.265625 0.1875 0.671875q0 0.40625 -0.234375 0.75q-0.234375 0.34375 -0.671875 0.546875q-0.421875 0.1875 -0.984375 0.1875q-0.90625 0 -1.390625 -0.375q-0.46875 -0.390625 -0.59375 -1.125z" fill-rule="nonzero"></path></g></svg>


## Why Dart

There are many code generation systems out there and this sort of
thing can be done in any number of languages. Dart has a very nice set
of facilities that make it perfect for this type of work.

### Compiled Languages vs Scripting Languages

Scripting languages obviously have an edge in code generation due to
the ability to rapidly prototype and lack of a build cycle. Compiled
languages have an advantage that in general due to typing and all else
equal, if the code compiles it has a better chance of running as
intended as more errors are taken care of at compile time. With Dart
you get the best of both approaches. Among the many awsome features of
Dart, those that most help with code generation include:

* Quick turnaround time as with other scripting langauges

* String interpolation. This is very important and it makes the job
  much simpler. Without it you would probably want to use a template
  engine. However, that has a cost since they usually make debugging
  painful.

* Errors caught by the type system during analysis

* Excellent set of support libraries. Need to read json input, to
  generate code no problem.

* For those that like to use an IDE, Dart Editor certainly adds value
  in trying to make use of the libraries to declaritively generate
  your code. It is nice to be able to use the intellisense to
  determine *properties* and *methods*.

### Templates vs Code Functions

The choice here is tougher. Template engines can provide a nice set of
features. However the trouble with them is often they are a pain to
debug, let alone test. No template engine is used for this code
generation, so finding and fixing errors involves reading and
understading Dart as opposed to scanning C++ looking template text for
the issues. Both can work.

## Examples

Here are some samples illustrating C++ scripts and corresponding
generated code:

* Dart Source [utils.dart](https://github.com/patefacio/fcs/blob/master/codegen/bin/libs/utils.dart) => C++ Generated [utils C++](https://github.com/patefacio/fcs/tree/master/cpp/fcs/utils)
* Dart Source [raii.dart](https://github.com/patefacio/fcs/blob/master/codegen/bin/libs/raii.dart) => C++ Generated  [raii C++](https://github.com/patefacio/fcs/tree/master/cpp/fcs/raii)

